/*
   Copyright (C) 2007-2010 contributors

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

*/

public class Base64
{

    /** Specify encoding (value is <tt>true</tt>). */
   public final static boolean ENCODE = true;


   /** Specify decoding (value is <tt>false</tt>). */
   public final static boolean DECODE = false;

   /** Specify that data should be compressed (value is <tt>true</tt>). */
   public final static boolean COMPRESS = true;


   /** Specify that data should not be compressed (value is <tt>false</tt>). */
   public final static boolean DONT_COMPRESS = false;


   /** Specify that data was compressed (value is <tt>true</tt>). */
   public final static boolean COMPRESSED = true;


   /** Specify that data was not compressed (value is <tt>false</tt>). */
   public final static boolean NOT_COMPRESSED = false;



   /** Break lines when encoding at 76-ish characters. */
   public final static boolean BREAK_LINES = true;


   /** Don't break lines when encoding (violates strict Base64 specification) */
   public final static boolean DONT_BREAK_LINES = false;


   /** The equals sign (=) as a byte. */
   private final static byte EQUALS_SIGN = (byte)'=';


   /**
   * Translates a Base64 value to either its 6-bit reconstruction value
   * or a negative number indicating some other meaning.
   **/
   private final static byte[] DECODABET =
   {
      -9,-9,-9,-9,-9,-9,-9,-9,-9,                 // Decimal  0 -  8
      -5,-5,                                      // Whitespace: Tab and Linefeed
      -9,-9,                                      // Decimal 11 - 12
      -5,                                         // Whitespace: Carriage Return
      -9,-9,-9,-9,-9,-9,-9,-9,-9,-9,-9,-9,-9,     // Decimal 14 - 26
      -9,-9,-9,-9,-9,                             // Decimal 27 - 31
      -5,                                         // Whitespace: Space
      -9,-9,-9,-9,-9,-9,-9,-9,-9,-9,              // Decimal 33 - 42
      62,                                         // Plus sign at decimal 43
      -9,-9,-9,                                   // Decimal 44 - 46
      63,                                         // Slash at decimal 47
      52,53,54,55,56,57,58,59,60,61,              // Numbers zero through nine
      -9,-9,-9,                                   // Decimal 58 - 60
      -1,                                         // Equals sign at decimal 61
      -9,-9,-9,                                      // Decimal 62 - 64
      0,1,2,3,4,5,6,7,8,9,10,11,12,13,            // Letters 'A' through 'N'
      14,15,16,17,18,19,20,21,22,23,24,25,        // Letters 'O' through 'Z'
      -9,-9,-9,-9,-9,-9,                          // Decimal 91 - 96
      26,27,28,29,30,31,32,33,34,35,36,37,38,     // Letters 'a' through 'm'
      39,40,41,42,43,44,45,46,47,48,49,50,51,     // Letters 'n' through 'z'
      -9,-9,-9,-9                                 // Decimal 123 - 126
      /*,-9,-9,-9,-9,-9,-9,-9,-9,-9,-9,-9,-9,-9,     // Decimal 127 - 139
      -9,-9,-9,-9,-9,-9,-9,-9,-9,-9,-9,-9,-9,     // Decimal 140 - 152
      -9,-9,-9,-9,-9,-9,-9,-9,-9,-9,-9,-9,-9,     // Decimal 153 - 165
      -9,-9,-9,-9,-9,-9,-9,-9,-9,-9,-9,-9,-9,     // Decimal 166 - 178
      -9,-9,-9,-9,-9,-9,-9,-9,-9,-9,-9,-9,-9,     // Decimal 179 - 191
      -9,-9,-9,-9,-9,-9,-9,-9,-9,-9,-9,-9,-9,     // Decimal 192 - 204
      -9,-9,-9,-9,-9,-9,-9,-9,-9,-9,-9,-9,-9,     // Decimal 205 - 217
      -9,-9,-9,-9,-9,-9,-9,-9,-9,-9,-9,-9,-9,     // Decimal 218 - 230
      -9,-9,-9,-9,-9,-9,-9,-9,-9,-9,-9,-9,-9,     // Decimal 231 - 243
      -9,-9,-9,-9,-9,-9,-9,-9,-9,-9,-9,-9         // Decimal 244 - 255 */
   };

   private final static byte WHITE_SPACE_ENC = -5; // Indicates white space in encoding
   private final static byte EQUALS_SIGN_ENC = -1; // Indicates equals sign in encoding


   /** Defeats instantiation. */
   private Base64(){}

   public static byte[] decode( String s )
   {
      byte[] bytes = s.getBytes();
      return decode(bytes, 0, bytes.length);
   }

   public static String decodeToString(String s)
   {
      return new String(decode(s));
   }

   private static int decode4to3( byte[] source, int srcOffset, byte[] destination, int destOffset )
    {
        // Example: Dk==
        if( source[ srcOffset + 2] == EQUALS_SIGN )
        {
            int outBuff =   ( ( DECODABET[ source[ srcOffset    ] ] & 0xFF ) << 18 )
                          | ( ( DECODABET[ source[ srcOffset + 1] ] & 0xFF ) << 12 );

            destination[ destOffset ] = (byte)( outBuff >>> 16 );
            return 1;
        }

        // Example: DkL=
        else if( source[ srcOffset + 3 ] == EQUALS_SIGN )
        {
            int outBuff =   ( ( DECODABET[ source[ srcOffset     ] ] & 0xFF ) << 18 )
                          | ( ( DECODABET[ source[ srcOffset + 1 ] ] & 0xFF ) << 12 )
                          | ( ( DECODABET[ source[ srcOffset + 2 ] ] & 0xFF ) <<  6 );

            destination[ destOffset     ] = (byte)( outBuff >>> 16 );
            destination[ destOffset + 1 ] = (byte)( outBuff >>>  8 );
            return 2;
        }

        // Example: DkLE
        else
        {
            try
            {
               int outBuff =   ( ( DECODABET[ source[ srcOffset     ] ] & 0xFF ) << 18 )
                             | ( ( DECODABET[ source[ srcOffset + 1 ] ] & 0xFF ) << 12 )
                             | ( ( DECODABET[ source[ srcOffset + 2 ] ] & 0xFF ) <<  6)
                             | ( ( DECODABET[ source[ srcOffset + 3 ] ] & 0xFF )      );


               destination[ destOffset     ] = (byte)( outBuff >> 16 );
               destination[ destOffset + 1 ] = (byte)( outBuff >>  8 );
               destination[ destOffset + 2 ] = (byte)( outBuff       );

               return 3;
            }
            catch( Exception e)
            {
                DVScheduler.log(""+source[srcOffset]+ ": " + ( DECODABET[ source[ srcOffset     ] ]  ) );
                DVScheduler.log(""+source[srcOffset+1]+  ": " + ( DECODABET[ source[ srcOffset + 1 ] ]  ) );
                DVScheduler.log(""+source[srcOffset+2]+  ": " + ( DECODABET[ source[ srcOffset + 2 ] ]  ) );
                DVScheduler.log(""+source[srcOffset+3]+  ": " + ( DECODABET[ source[ srcOffset + 3 ] ]  ) );
                return -1;
            }   //e nd catch
        }
    }   // end decodeToBytes

   public static byte[] decode( byte[] source, int off, int len )
   {
      int    len34   = len * 3 / 4;
      byte[] outBuff = new byte[ len34 ]; // Upper limit on size of output
      int    outBuffPosn = 0;

      byte[] b4        = new byte[4];
      int    b4Posn    = 0;
      int    i         = 0;
      byte   sbiCrop   = 0;
      byte   sbiDecode = 0;
      for( i = off; i < off+len; i++ )
      {
         sbiCrop = (byte)(source[i] & 0x7f); // Only the low seven bits
         sbiDecode = DECODABET[ sbiCrop ];

         if( sbiDecode >= WHITE_SPACE_ENC ) // White space, Equals sign or better
         {
            if( sbiDecode >= EQUALS_SIGN_ENC )
            {
               b4[ b4Posn++ ] = sbiCrop;
               if( b4Posn > 3 )
               {
                  outBuffPosn += decode4to3( b4, 0, outBuff, outBuffPosn );
                  b4Posn = 0;

                  // If that was the equals sign, break out of 'for' loop
                  if( sbiCrop == EQUALS_SIGN )
                  break;
               }   // end if: quartet built

            }   // end if: equals sign or better

         }   // end if: white space, equals sign or better
         else
         {
            System.err.println( "Bad Base64 input character at " + i + ": " + source[i] + "(decimal)" );
            return null;
         }   // end else:
      }   // each input character

      byte[] out = new byte[ outBuffPosn ];
      System.arraycopy( outBuff, 0, out, 0, outBuffPosn );
      return out;
   }

}
