/*
   Copyright (C) 2007-2010 contributors

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

 */

import java.util.*;

import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.NamedNodeMap;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;
import org.w3c.dom.Text;

public class CaptureDevice
{
	public static final int TUNING_NORMAL = 1;
	public static final int TUNING_FAST = 2;

	private String m_name = "";
	private int m_number = -1;
	private String m_uniqueName = "";
	private String m_setting = "";
	private String m_settingName = "";
	private String m_hardSetting = "";
	private CaptureProcess capture = null;
	private NetworkType m_networkType = new NetworkType();
	private String m_userData = "";
	private int m_tuningMethod = TUNING_NORMAL;

	public CaptureDevice(String name, int number, String uniqueName, String networkType)
	{
		m_name = name;
		m_number = number;
		m_uniqueName = uniqueName;

		m_networkType.set(networkType);
	}

	public CaptureDevice(String name, int number, String uniqueName, SortedSet<String> networkTypes)
	{
		m_name = name;
		m_number = number;
		m_uniqueName = uniqueName;

		m_networkType.set(networkTypes);
	}

	public CaptureDevice(Node item)
	{
		m_uniqueName = null;	// Previous version doesn't have unique name.

		NamedNodeMap itemAttribs = item.getAttributes();
		m_name = itemAttribs.getNamedItem("name").getTextContent();

		NodeList itemNodes = item.getChildNodes();
		Node itemNode = null;  
		for(int y = 0; y < itemNodes.getLength(); y++)
		{
			itemNode = itemNodes.item(y);

			if(itemNode.getNodeName().equals("id"))
			{
				m_number = Integer.parseInt(itemNode.getTextContent());
			}
			else if(itemNode.getNodeName().equals("uniquename"))
			{
				m_uniqueName = itemNode.getTextContent();
			}
			else if(itemNode.getNodeName().equals("networktype"))
			{
				m_networkType.add(itemNode.getTextContent());
			}
			else if(itemNode.getNodeName().equals("tuning"))
			{
				setTuningMethod( Integer.parseInt(itemNode.getTextContent()) );
			}
			else if(itemNode.getNodeName().equals("userdata"))
			{
				m_userData = DVScheduler.checkBreaks(itemNode.getTextContent());
			}
		}
	}

	public void setName(String name)
	{
		m_name = name;
	}

	public String getName()
	{
		return m_name;
	}

	public void setUniqueName(String uniqueName)
	{
		m_uniqueName = uniqueName;
	}

	public String getUniqueName()
	{
		return m_uniqueName;
	}

	public void setNumber(int number)
	{
		m_number = number;
	}

	public int getNumber()
	{
		return m_number;
	}

	// Return value
	// 0 not running.
	// 1 running idle.
	// 2 running with captures.
	public int getStatus()
	{
		return capture == null ? 0 : capture.getStatus();
	}

	public void setInUse(String hardSetting, String setting, String settingName, CaptureProcess cap)
	{
		m_hardSetting = hardSetting;
		m_setting = setting;
		m_settingName = settingName;
		capture = cap;
	}

	public String getSetting()
	{
		return m_setting;
	}

	public String getSettingName()
	{
		return (m_settingName == null || m_settingName.isEmpty()) ? m_setting : m_settingName;
	}

	public void setSetting(String setting, String settingName)
	{
		m_setting = setting;
		m_settingName = settingName;
	}

	public CaptureProcess getCaptureProcess()
	{
		return capture;
	}

	public boolean isNetworkTypeCompatible(String networkType)
	{
		return m_networkType.isCompatible(networkType);
	}

	public boolean isHardSettingEqual(String hardSetting)
	{
		return m_hardSetting.equals(hardSetting);
	}

	public boolean isHardSettingEqual(Channel channel)
	{
		return isHardSettingEqual(channel.getHardSetting());
	}

	public boolean isHardSettingCompatible(String hardSetting)
	{
		return m_networkType.isCompatible(hardSetting);
	}

	public boolean isHardSettingCompatible(Channel channel)
	{
		return isHardSettingCompatible(channel.getHardSetting());
	}

	public boolean isSettingCompatible(String setting)
	{
		return m_setting.equals(setting);
	}

	public String getNetworkType()
	{
		return m_networkType.toString();
	}

	public void setNetworkType(String networkType)
	{
		m_networkType.set(networkType);
	}

	public void setUserData(String userData)
	{
		m_userData = DVScheduler.checkBreaks(userData);
	}

	public String getUserData()
	{
		return m_userData;
	}

	public void setTuningMethod(int tuningMethod)
	{
		m_tuningMethod = tuningMethod;
	}

	public int getTuningMethod()
	{
		return m_tuningMethod;
	}

	public Element getXML(Document doc)
	{
		Element elm = null;
		Text text = null;

		Element item = doc.createElement("device");
		item.setAttribute("name", m_name);

		elm = doc.createElement("id");
		text = doc.createTextNode(new Integer(m_number).toString());
		elm.appendChild(text);
		item.appendChild(elm);  

		elm = doc.createElement("uniquename");
		text = doc.createTextNode(m_uniqueName);
		elm.appendChild(text);
		item.appendChild(elm);  

		m_networkType.getXML(doc, item);

		elm = doc.createElement("tuning");
		text = doc.createTextNode(Integer.toString(getTuningMethod()));
		elm.appendChild(text);
		item.appendChild(elm);  

		elm = doc.createElement("userdata");
		text = doc.createTextNode(m_userData);
		elm.appendChild(text);
		item.appendChild(elm);  

		return item;
	}

	public String replaceKeywords(String text)
	{
		text = text.replace("$deviceID0", Integer.toString(getNumber()));
		text = text.replace("$deviceID1", Integer.toString(getNumber() + 1));
		text = text.replace("$deviceID", Integer.toString(getNumber()));
		text = text.replace("$methodtuning", Integer.toString(getTuningMethod()));

		if ( (m_uniqueName == null) || m_uniqueName.isEmpty() )
			text = text.replace("$deviceuniquename", getName());
		else
			text = text.replace("$deviceuniquename", m_uniqueName);

		String[] userDatas = DVScheduler.getArray(getUserData());
		for(int x = 0; x < userDatas.length; x++)
		{
			text = text.replace("$devicedata" + (x < 9 ? "0" : "") + (x+1), userDatas[x]);
		}

		CaptureDeviceList devList = CaptureDeviceList.getInstance();
		userDatas = DVScheduler.getArray( devList.getDefaultData() );
		for(int x = 0; x < userDatas.length; x++)
		{
			text = text.replace("$devicedata" + (x < 9 ? "0" : "") + (x+1), userDatas[x]);
		}

		return text.replace("$devicename", getName());
	}
}