/*
   Copyright (C) 2007-2010 contributors

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

 */

import java.io.*;
import java.util.*;

import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.transform.OutputKeys;
import javax.xml.transform.Result;
import javax.xml.transform.Source;
import javax.xml.transform.Transformer;
import javax.xml.transform.TransformerFactory;
import javax.xml.transform.dom.DOMSource;
import javax.xml.transform.stream.StreamResult;

import org.w3c.dom.DOMImplementation;
import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;
import org.w3c.dom.Text;

public class CaptureDeviceList
{
	private static CaptureDeviceList instance = null;
	private Vector<CaptureDevice> captureDevices = new Vector<CaptureDevice>();
	private String m_defaultData = "";

	private CaptureDeviceList()
	{
		loadDeviceList();
	}

	public String toString()
	{
		return "Currently have " + captureDevices.size() + " devices in available device list";
	}

	public static CaptureDeviceList getInstance()
	{
		synchronized (DataStore.class)
		{
			if (instance == null)
			{
				instance = new CaptureDeviceList();
				return instance;
			}
			else
			{
				return instance;
			}
		}
	}   

	private void loadDeviceList()
	{
		try
		{
			DocumentBuilderFactory docBuilderFactory = DocumentBuilderFactory.newInstance();
			DocumentBuilder docBuilder = docBuilderFactory.newDocumentBuilder();
			DataStore store = DataStore.getInstance();
			Document doc = docBuilder.parse(DataStore.m_saveFolderData + "devices.xml");

			int tuningMethod = -2;
			try
			{
				if (doc.getDocumentElement().getAttributes().getNamedItem("version").getTextContent().equals("2"))
					tuningMethod = -1;
			}
			catch(Exception e){}

			NodeList items = doc.getElementsByTagName("device");

			if (tuningMethod == -2)
			{
				String userKeywords = store.getProperty("user.keywords", "$$vlcexe->C:\\Program Files\\VideoLAN\\VLC\\vlc.exe");
				String[] userKeywords2 = DVScheduler.getArray(userKeywords);
				String newUserKeywords = "";
				for(int x = 0; x < userKeywords2.length; x++)
				{
					int pos = userKeywords2[x].indexOf("->");
					if (pos >= 0)
					{
						String key = userKeywords2[x].substring(0, pos);
						String value = userKeywords2[x].substring(pos + 2);
						if (key.equals("$$tuningmethod"))
						{
							try
							{
								tuningMethod = Integer.parseInt(value);
							}
							catch (Exception e)
							{
							}
						}
						else
						{
							if (x > 0)
								newUserKeywords += "\n";
							newUserKeywords += userKeywords2[x];
						}
					}
				}
				try
				{
					store.setProperty("user.keywords", newUserKeywords);
				}
				catch (Exception e)
				{
					e.printStackTrace();
				}
			}

			captureDevices = new Vector<CaptureDevice>();

			for(int x = 0; x < items.getLength(); x++)
			{
				Node item = items.item(x);
				CaptureDevice cap = new CaptureDevice(item);
				if (tuningMethod > -1)
					cap.setTuningMethod(tuningMethod);
				if (cap.getUniqueName() != null)	// Previous version doesn't have unique name.
					captureDevices.add(cap);
			}

			items = doc.getElementsByTagName("defaultData");
			if (items.getLength() == 1)
				m_defaultData = items.item(0).getTextContent();

			DVScheduler.log("devices.xml found and loaded (" + items.getLength() + ")");
			if (tuningMethod > -1)
				saveDeviceList();
		}
		catch (Exception e)
		{
			captureDevices = new Vector<CaptureDevice>();
			DVScheduler.log("Error loading devices.xml, starting with blank device list.");
			//e.printStackTrace();
		}
	}

	public void saveDeviceList() throws Exception
	{
		// Creation of an XML document
		DocumentBuilderFactory dbf = DocumentBuilderFactory.newInstance();
		DocumentBuilder db = dbf.newDocumentBuilder();
		DOMImplementation di = db.getDOMImplementation();

		Document doc = di.createDocument("", "devices", null);
		Element root = doc.getDocumentElement();
		root.setAttribute("version", "2");

		for (int x = 0; x < captureDevices.size(); x++)
		{
			CaptureDevice cap = captureDevices.get(x);
			root.appendChild(cap.getXML(doc));
		}

		Element item = doc.createElement("defaultData");
		Text text = doc.createTextNode(m_defaultData);
		item.appendChild(text);
		root.appendChild(item);

		ByteArrayOutputStream buff = new ByteArrayOutputStream();
		TransformerFactory factory = TransformerFactory.newInstance();

		Transformer transformer = factory.newTransformer();
		transformer.setOutputProperty(OutputKeys.METHOD, "xml");
		transformer.setOutputProperty(OutputKeys.INDENT, "yes");      

		Source source = new DOMSource(doc);
		Result result = new StreamResult(buff);
		transformer.transform(source, result);

		FileOutputStream fos = new FileOutputStream(DataStore.m_saveFolderData + "devices.xml");
		Writer out = new BufferedWriter(new OutputStreamWriter(fos, "UTF-8"));
		out.write(buff.toString("UTF-8"));
		out.close();
	}

	public void addDevice(CaptureDevice dev)
	{
		captureDevices.add(dev);
	}

	public void addDeviceAt(int index, CaptureDevice dev)
	{
		captureDevices.add(index, dev);
	}   

	public CaptureDevice removeDevice(int index)
	{
		return (CaptureDevice)captureDevices.remove(index);
	}    

	public void removeAllDevices()
	{
		captureDevices.removeAllElements();
	}    

	public int getDeviceCount()
	{
		return captureDevices.size();
	}

	public int getDeviceCount(String networkType)
	{
		int count = 0;
		for(int x = 0; x < captureDevices.size(); x++)
		{
			CaptureDevice cap = (CaptureDevice)captureDevices.get(x);
			if ( cap.isNetworkTypeCompatible(networkType) )
			{
				count++;
			}
		}
		return count;
	}

	public int getDeviceCountByChannel(String channelID)
	{
		DataStore store = DataStore.getInstance();

		Channel ch = store.getChannel(channelID);

		return ch == null ? 0 : getDeviceCount(ch.getNetworkType());
	}

	public CaptureDevice getDevice(int x)
	{
		return (CaptureDevice)captureDevices.get(x);
	}

	public CaptureDevice GetFirstDevice(String networkType)
	{
		for(int x = 0; x < captureDevices.size(); x++)
		{
			CaptureDevice cap = (CaptureDevice)captureDevices.get(x);
			if (cap.isNetworkTypeCompatible(networkType))
			{
				return cap;
			}
		}

		return null;
	}

	public CaptureDevice GetFirstFreeDevice(String networkType)
	{
		for(int x = 0; x < captureDevices.size(); x++)
		{
			CaptureDevice cap = (CaptureDevice)captureDevices.get(x);
			if ((cap.getStatus() == 0) && cap.isNetworkTypeCompatible(networkType))
			{
				return cap;
			}
		}

		return null;
	}

	public int getRunningDeviceCount()
	{
		int active = 0;
		for(int x = 0; x < captureDevices.size(); x++)
		{
			CaptureDevice cap = (CaptureDevice)captureDevices.get(x);
			if (cap.getStatus() > 0)
			{
				active++;
			}
		}
		return active;
	}   

	public int getActiveDeviceCount(String networkType)
	{
		int active = 0;
		for(int x = 0; x < captureDevices.size(); x++)
		{
			CaptureDevice cap = (CaptureDevice)captureDevices.get(x);
			if ( (cap.getStatus() == 2) && (cap.isNetworkTypeCompatible(networkType)) )
			{
				active++;
			}
		}
		return active;
	}   

	public boolean isAvailableDevice(String networkType, String hardSetting, String setting)
	{
		for(int x = 0; x < captureDevices.size(); x++)
		{
			CaptureDevice cap = (CaptureDevice)captureDevices.get(x);
			if ( ((cap.getStatus() == 0) && cap.isNetworkTypeCompatible(networkType)) ||
					((cap.getStatus() == 1) && cap.isHardSettingCompatible(hardSetting)) || 
					((cap.getStatus() == 2) && cap.isSettingCompatible(setting)) )
			{
				return true;
			}
		}
		return false;
	}

	public boolean isAvailableRunningDevice(String hardSetting, String setting)
	{
		for(int x = 0; x < captureDevices.size(); x++)
		{
			CaptureDevice cap = (CaptureDevice)captureDevices.get(x);
			if ( ((cap.getStatus() == 1) && cap.isHardSettingCompatible(hardSetting)) || 
					((cap.getStatus() == 2) && cap.isSettingCompatible(setting)) )
			{
				return true;
			}
		}
		return false;
	}

	public void stopIdleRunningDevices()
	{
		for(int x = 0; x < captureDevices.size(); x++)
		{
			CaptureDevice captureDevice = captureDevices.get(x);
			if ( (captureDevice != null) && (captureDevice.getStatus() == 1) )
			{
				captureDevice.getCaptureProcess().stopDevice();
			}
		}
	}

	public void stopIdleRunningDevice(int index)
	{
		if ( (index < 0) || (index >= captureDevices.size()) )
			return;
		
		CaptureDevice captureDevice = captureDevices.get(index);
		if ( (captureDevice != null) && (captureDevice.getStatus() == 1) )
		{
			captureDevice.getCaptureProcess().stopDevice();
		}
	}

	public void stopIdleRunningDevice(String uniqueName)
	{
		for(int x = 0; x < captureDevices.size(); x++)
		{
			CaptureDevice captureDevice = captureDevices.get(x);
			if ( (captureDevice != null) && (captureDevice.getStatus() == 1) && captureDevice.getUniqueName().equals(uniqueName) )
			{
				captureDevice.getCaptureProcess().stopDevice();
				break;
			}
		}
	}
	
	public boolean isSurf()
	{
		for(int x = 0; x < captureDevices.size(); x++)
		{
			CaptureDevice cap = (CaptureDevice)captureDevices.get(x);
			CaptureProcess proc = cap.getCaptureProcess();
			if ( proc != null )
			{
				if ( !proc.isProcessRunning( ))
					proc.stopDevice();
				else if ( proc.isSurf() )
					return true;
			}
		}
		return false;
	}

	public boolean isPreviewOrSurfOrEpgScan()
	{
		for(int x = 0; x < captureDevices.size(); x++)
		{
			CaptureDevice cap = (CaptureDevice)captureDevices.get(x);
			CaptureProcess proc = cap.getCaptureProcess();
			if ( proc != null )
			{
				if ( !proc.isProcessRunning( ))
					proc.stopDevice();
				else if ( proc.isPreviewOrSurfOrEpgScan() )
					return true;
			}
		}
		return false;
	}

	void setDefaultData(String data)
	{
		m_defaultData = DVScheduler.checkBreaks(data);
	}

	String getDefaultData()
	{
		return m_defaultData;
	}
}
