/*
   Copyright (C) 2007-2010 contributors

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

 */

import java.io.File;
import java.io.OutputStream;
import java.util.*;

public class CaptureProcess
{
	Process m_captureProcess = null;
	OutputStream m_out = null;
	boolean m_bCaptureRunning = false;
	boolean m_bEpgScanRunning = false;

	// Timeouts in 1/10 second. E.g. 650 = 65 seconds.
	int m_timeoutStartDevice = 65 * 10;	// Device is to be initialized, and capture started.
	int m_timeoutStopDevice = 65 * 10;	// Device to be stopped.
	int m_timeoutStartCapture = 65 * 10;	// Add capture. Device already initialized.
	int m_timeoutStopCapture = 65 * 10;	// Stop capture. Device already initialized.
	int m_timeoutChangeChannel = 20 * 10;	// Change channel (but same frequency, etc). Device already initialized.
	int m_timeoutChangeSetting = 20 * 10;	// Change frequency, etc. Device already initialized.
	int m_timeoutChangeInitialSetting = 20 * 10;	// Device already initialized.
	int m_timeoutAction = 20 * 10;	// Simple commands to device. Device already initialized.

	ProcessInputReader m_processInputReader = null;
	private CaptureDevice m_captureDevice = null;
	private int m_totalCapture = 0;
	private HashSet<Long> captureID = new HashSet<Long>();
	private HashSet<Long> errorID = new HashSet<Long>();
	private HashSet<Long> m_pidChangeIds = new HashSet<Long>();
	private CaptureEngine m_captureEngine = null;

	public CaptureProcess(CaptureDevice cap)
	{
		DataStore store = DataStore.getInstance();
		m_timeoutStartDevice = store.getPropertyInt("Capture.startTimeout", 65) * 10;
		m_timeoutStopDevice = m_timeoutStartDevice;	// Set to same for now.
		m_timeoutStartCapture = m_timeoutStartDevice;
		m_timeoutStopCapture = m_timeoutStartDevice;

		m_timeoutChangeChannel = store.getPropertyInt("Capture.changeTimeout", 20) * 10;
		m_timeoutChangeSetting = m_timeoutChangeChannel;	// Set to same for now.
		m_timeoutChangeInitialSetting = m_timeoutChangeChannel; 
		m_timeoutAction = m_timeoutChangeChannel;

		m_captureDevice = cap;
	}

	public CaptureDevice getCaptureDevice()
	{
		return m_captureDevice;
	}

	public CaptureEngine getCaptureEngine()
	{
		return m_captureEngine;
	}

	public int getTotalCapture()
	{
		return m_totalCapture;
	}

	protected void finalize() throws Throwable 
	{
		DVScheduler.log("Finalize called on CaptureProcess object");

		try
		{
			if(m_captureProcess != null)
			{
				DVScheduler.log("Finalize KILLING external app");
				m_captureProcess.destroy();
			}
		}
		catch(Exception e)
		{
			DVScheduler.log("Error destroying external app on finalize");
			e.printStackTrace();
		}
		finally
		{
			super.finalize();
		}
	}

	private String getCommand(ScheduleItem item, Channel ch, CaptureFile capFile, String template)
	{
		DataStore store = DataStore.getInstance();

		String command = template.replace("$name", m_captureEngine.formatFileName(capFile.getEncodedDirectoryAndName()));

		command = m_captureEngine.replaceKeywords(command);
		command = capFile.replaceKeywords(command);
		command = store.replaceKeywords(command);
		command = m_captureDevice.replaceKeywords(command);
		command = item.replaceKeywords(command);	// Has to be before Channel::replaceKeywords(...)

		return ch.replaceKeywords(command);
	}

	private String getEpgScanCommand(Channel ch)
	{
		DataStore store = DataStore.getInstance();

		if (!store.getProperty("guide.source.type", "Internal").contains("Internal"))
			return null;

		String filename = DataStore.m_saveFolder + "xmltvInternal" + File.separator + ch.getSettingForFilename() + ".xml";
		filename = m_captureEngine.formatFileName(filename);
		String command = store.replaceKeywords(m_captureEngine.getEpgScanCommand());
		return command.replace("$filename", filename);
	}

	public int startDeviceAndEpgScan(Channel channel, CaptureEngine captureEngine2)
	{
		if(m_captureProcess != null)
		{
			DVScheduler.log("Tried to start capture while one was already running!");
			return -1001;
		}

		m_captureEngine = captureEngine2;

		try
		{
			DataStore store = DataStore.getInstance();
			String command = DataStore.m_saveFolder + "xmltvInternal" + File.separator + channel.getSetting() + ".xml";
			command = m_captureEngine.getInitialEpgScanCommand().replace("$filename", command);
			command = store.replaceKeywords(command);
			command = m_captureDevice.replaceKeywords(command);
			command = channel.replaceKeywords(command);
			command = Formatter.formatCommand(command);
			DVScheduler.log("EPG Scan initial command:" + command);
			Runtime runner = Runtime.getRuntime();
			m_captureProcess = runner.exec(Formatter.parseCommand(command));
			//capApp = new ProcessBuilder(command).start();
		}
		catch(Exception e)
		{
			DVScheduler.log("Error launching external capture app!");
			e.printStackTrace();
			return -1002;
		}
		
		m_out = m_captureProcess.getOutputStream();

		try
		{
			// now start up a thread to read responce data
			m_processInputReader = new ProcessInputReader(m_captureProcess, m_captureEngine);
			Thread monitor = new Thread(Thread.currentThread().getThreadGroup(), m_processInputReader, m_processInputReader.getClass().getName());
			monitor.start();
		}
		catch(Exception e)
		{
			e.printStackTrace();

			m_bCaptureRunning = false;

			try
			{
				if(m_captureProcess != null)
					m_captureProcess.destroy();
			}
			catch(Exception e2){}

			return -1005;
		}

		m_bCaptureRunning = true;
		m_bEpgScanRunning = true;
		m_captureDevice.setInUse(channel.getHardSetting(), channel.getSetting(), channel.getDemodulatorName(), this);
		return 0;
	}

	public int startDeviceAndCapture(ScheduleItem item, CaptureFile captureFile, Channel channel, 
			CaptureEngine captureEngine2, 
			StringBuffer log)
	{
		if(m_captureProcess != null)
		{
			DVScheduler.log("Tried to start capture while one was already running!");
			return -1001;
		}

		m_captureEngine = captureEngine2;

		long id = item.getID();

		try
		{
			String initialCommand = null;
			if (item.isPreviewLocal())
				initialCommand = m_captureEngine.getInitialLocalPreviewCommand();
			else if (item.isPreviewRemote())
				initialCommand = m_captureEngine.getInitialRemotePreviewCommand();
			else if (item.isSurfLocal())
				initialCommand = m_captureEngine.getInitialLocalSurfCommand();
			else if (item.isSurfRemote())
				initialCommand = m_captureEngine.getInitialRemoteSurfCommand();
			else
				initialCommand = m_captureEngine.getInitialCommand();

			String args = getCommand(item, channel, captureFile, initialCommand);
			args = args.replace("$command", getCommand(item, channel, captureFile, captureFile.getCommandLine(m_captureEngine)));
			args = Formatter.formatCommand(args);
			DVScheduler.log("Capture engine initial command:" + args);
			Runtime runner = Runtime.getRuntime();
			m_captureProcess = runner.exec(Formatter.parseCommand(args));
			//capApp = new ProcessBuilder(args).start();
		}
		catch(Exception e)
		{
			DVScheduler.log("Error launching external capture app!");
			e.printStackTrace();
			return -1002;
		}

		if (m_captureEngine.isStdin())
		{
			try
			{
				m_out = m_captureProcess.getOutputStream();

				String command = getCommand(item, channel, captureFile, captureFile.getCommandLine(m_captureEngine));
				command = Formatter.formatCommand(command);
				DVScheduler.log("Command: " + command);
				if (!command.isEmpty())
					m_out.write(command.getBytes());

				// Add EPG scan command
				command = getEpgScanCommand(channel);
				DVScheduler.log("EPG scan command: " + command);
				if ((command != null) && !command.isEmpty())
					m_out.write(command.getBytes());

				m_out.flush();
			}
			catch(Exception e)
			{
				DVScheduler.log("Error sending capture config data to external app!");
				e.printStackTrace();
				if (m_captureProcess != null)
					m_captureProcess.destroy();
				m_bCaptureRunning = false;
				return -1003;         
			}
		}

		try
		{
			// now start up a thread to read responce data
			m_processInputReader = new ProcessInputReader(m_captureProcess, m_captureEngine);
			Thread monitor = new Thread(Thread.currentThread().getThreadGroup(), m_processInputReader, m_processInputReader.getClass().getName());
			monitor.start();

			if (m_captureEngine.isStdout())
			{
				int timeOut = m_timeoutStartDevice;

				reader:
					while(true)
					{
						while (m_processInputReader.isResponse())
						{
							String resp = m_processInputReader.getResponse();

							if (m_captureEngine.isRecordingStarted(resp, id))
							{
								break reader;
							}
							else if (m_captureEngine.isRecordingError(resp, id))
							{
								DVScheduler.log("Error when starting capture with external app!");
								m_out.write((m_captureEngine.getExitCommand()).getBytes());
								m_out.flush();

								//if(capApp != null)
								//capApp.destroy();
								m_bCaptureRunning = false;
								//return -1006;
							}
							else
								log.append(resp + "\n");
						}

						if(m_processInputReader.getExitCode() != null)
						{
							m_bCaptureRunning = false;
							int code = m_processInputReader.getExitCode().intValue();

							if(code == -1)
								return -10001;
							else if(code == -2)
								return -10002;
							else if(code == -3)
								return -10003;
							else if(code == -4)
								return -10004;               
							else
								return -11111;
						}
						else
						{
							try
							{
								m_processInputReader.waitResponse();
								//Thread.sleep(100);
								timeOut--;
							}
							catch(Exception e){}
						}

						if(timeOut == 0)
						{
							DVScheduler.log("Timeout reached when starting capture with external app!");
							if(m_captureProcess != null)
								m_captureProcess.destroy();
							m_bCaptureRunning = false;
							return -1004;
						}
					}
			}
		}
		catch(Exception e)
		{
			DVScheduler.log("Error waiting for capture to start");
			e.printStackTrace();

			m_bCaptureRunning = false;

			try
			{
				if(m_captureProcess != null)
					m_captureProcess.destroy();
			}
			catch(Exception e2){}

			return -1005;
		}

		m_bCaptureRunning = true;
		m_totalCapture++;
		captureID.add(id);
		m_captureDevice.setInUse(channel.getHardSetting(), channel.getSetting(), channel.getDemodulatorName(), this);
		return 0;
	}

	public int addCapture(ScheduleItem item, CaptureFile captureFile, Channel channel, StringBuffer log)
	{
		if(m_processInputReader == null || m_processInputReader.getExitCode() != null || !m_captureEngine.isCanRecordMultiple())
			return -1;

		long id = item.getID();

		try
		{
			errorID.remove(id);
			m_pidChangeIds.remove(id);
			captureID.remove(id);

			if (m_captureEngine.isStdin())
			{
				String command = getCommand(item, channel, captureFile, captureFile.getCommandLine(m_captureEngine));
				command = Formatter.formatCommand(command);
				DVScheduler.log("Command: " + command);
				m_out.write(command.getBytes());
				m_out.flush();
			}

			if (m_captureEngine.isStdout())
			{
				int timeOut = m_timeoutStartCapture;

				reader:
					while(true)
					{
						while (m_processInputReader.isResponse())
						{
							String resp = m_processInputReader.getResponse();

							if (m_captureEngine.isRecordingStarted(resp, id))
							{
								break reader;
							}
							else if (m_captureEngine.isRecordingError(resp, id))
							{
								DVScheduler.log("Error with add capture command to external app!");
								m_out.write((m_captureEngine.getStopRecordingCommand().replace("$rid", Long.toString(id))).getBytes());
								m_out.flush();
								return -1008;
							}
							else
							{
								Long id2 = m_captureEngine.getRecordingError(resp);
								if (id2 != null)
									errorID.add(id2);
								id2 = m_captureEngine.getPidChange(resp);
								if (id2 != null)
									m_pidChangeIds.add(id2);
								else
									log.append(resp + "\n");
							}
						}

						if(m_processInputReader.getExitCode() != null)
						{
							m_bCaptureRunning = false;
							int code = m_processInputReader.getExitCode().intValue();

							if(code == -1)
								return -10001;
							else if(code == -2)
								return -10002;
							else if(code == -3)
								return -10003;
							else if(code == -4)
								return -10004;               
							else
								return -11111;
						}
						else
						{
							try
							{
								m_processInputReader.waitResponse();
								//Thread.sleep(100);
								timeOut--;
							}
							catch(Exception e){}
						}

						if(timeOut == 0)
						{
							DVScheduler.log("Timeout reached sending add capture command to external app!");
							m_out.write((m_captureEngine.getStopRecordingCommand().replace("$rid", Long.toString(id))).getBytes());
							m_out.flush();
							return -1006;
						}
					}
			}
		}
		catch(Exception e)
		{
			DVScheduler.log("Error sending add capture command to external app!");
			e.printStackTrace();
			return -1007;
		}

		m_totalCapture++;
		captureID.add(id);

		return 0;
	}

	public int stopCapture(long id)
	{
		if(m_processInputReader == null || m_processInputReader.getExitCode() != null)
			return -1;

		if (!m_captureEngine.isCanRecordMultiple())
			return stopDevice();

		try
		{
			if (m_captureEngine.isStdin())
			{
				/*
				// Add EPG stop command
				String command = m_captureEngine.getStopEpgScanCommand();
				DVScheduler.log("EPG stop command: " + command);
				if ((command != null) && !command.isEmpty())
					m_out.write(command.getBytes());
				*/

				String command = Formatter.formatCommand(m_captureEngine.getStopRecordingCommand().replace("$rid", Long.toString(id)));
				DVScheduler.log("Stop record: " + command);
				m_out.write(command.getBytes());
				m_out.flush();
			}

			if (m_captureEngine.isStdout())
			{
				int timeOut = m_timeoutStopCapture;

				reader:
					while(true)
					{
						while (m_processInputReader.isResponse())
						{
							String resp = m_processInputReader.getResponse();
							//DVScheduler.log(resp);

							if (m_captureEngine.isRecordingStopped(resp, id))
							{
								break reader;
							}
							else if (m_captureEngine.isRecordingError(resp, id))
							{
								DVScheduler.log("Error with stop record command to external capture app!");
								break reader;
							}
							else
							{
								Long id2 = m_captureEngine.getRecordingError(resp);
								if (id2 != null)
									errorID.add(id2);
								id2 = m_captureEngine.getPidChange(resp);
								if (id2 != null)
									m_pidChangeIds.add(id2);
							}
						}

						if(m_processInputReader.getExitCode() != null)
						{
							break;
						}            
						else
						{
							try
							{
								m_processInputReader.waitResponse();
								//Thread.sleep(100);
								timeOut--;
							}
							catch(Exception e){}
						}

						if(timeOut == 0)
						{
							DVScheduler.log("Timeout reached with stop record command to external capture app!");
							//if(capApp != null)
							//capApp.destroy();
							if(--m_totalCapture <= 0)
								stopDevice();
							return -2;
						}
					}
			}
		}
		catch(Exception e)
		{
			DVScheduler.log("Error with stop record command!");
			e.printStackTrace();
			if(--m_totalCapture <= 0)
				stopDevice();
			return -2;
		}

		errorID.remove(id);
		m_pidChangeIds.remove(id);
		captureID.remove(id);

		--m_totalCapture;
		//		if(--m_totalCapture <= 0)
		//		return stopCapture();

		return 0;
	}

	public int stopDevice()
	{
		if(m_processInputReader == null || m_processInputReader.getExitCode() != null)
		{
			m_captureDevice.setInUse(null, null, null, null);
			return 0;
		}

		try
		{
			if (m_captureEngine.isStdin())
			{
				String command = Formatter.formatCommand(m_captureEngine.getExitCommand());
				DVScheduler.log("Calling exit:" + command);
				m_out.write(command.getBytes());
				m_out.flush();
			}

			if (!m_captureEngine.isExitTerminate())
			{
				int timeOut = m_timeoutStopDevice;

				while(true)
				{
					if(m_processInputReader.getExitCode() != null)
					{
						break;
					}
					else
					{
						try
						{
							m_processInputReader.waitResponse();
							//Thread.sleep(100);
							timeOut--;
						}
						catch(Exception e){}
					}

					if(timeOut == 0)
					{
						DVScheduler.log("Timeout reached when stopping external capture app!");
						//if (reader.getExitCode() != null)
						//break;
						if (m_captureProcess != null)
							m_captureProcess.destroy();
						return -2;
					}
				}
			}
		}
		catch(Exception e)
		{
			DVScheduler.log("Error trying to exit capture!");
			e.printStackTrace();
			return -2;
		}
		finally
		{
			try
			{
				if(m_captureProcess != null)
					m_captureProcess.destroy();
			}
			catch(Exception e){}

			m_bCaptureRunning = false;

			m_captureProcess = null;
		}

		m_captureDevice.setInUse(null, null, null, null);
		return 0;
	}

	public int setWriterStates(boolean[] writerStatuses)
	{
		String writerSetData = "writer_states(";

		for(int x = 0; x < writerStatuses.length; x++)
		{
			if(writerStatuses[x] == true)
			{
				writerSetData += "1";
			}
			else
			{
				writerSetData += "0";
			}
		}

		writerSetData += ")\n";

		try
		{
			//DVScheduler.log("Sending writer states : " + writerSetData);
			m_out.write(writerSetData.getBytes());
			m_out.flush();
		}
		catch(Exception e)
		{
			return -1;
		}

		return 0;
	}

	public boolean isCaptureRunning(long id)
	{
		if(m_processInputReader != null && m_processInputReader.getExitCode() == null && m_bCaptureRunning == true 
				&& captureID.contains(id) && getStatus(id) == 0)

			return true;
		else
			return false;
	}

	public boolean isCaptureRunning()
	{
		if (m_processInputReader != null && m_processInputReader.getExitCode() == null && m_bCaptureRunning == true)
			return true;
		else
			return false;
	}

	public boolean isProcessRunning()
	{
		if (m_processInputReader != null && m_processInputReader.getExitCode() == null)
			return true;
		else
			return false;
	}

	public boolean isPidChanged(long id)
	{
		if (m_pidChangeIds.contains(id))
		{
			m_pidChangeIds.remove(id);
			return true;
		}
		else
			return false;
	}

	// Return value
	// 0 not running.
	// 1 running idle.
	// 2 running with captures.
	public int getStatus()
	{
		if (m_processInputReader != null && m_processInputReader.getExitCode() == null && m_bCaptureRunning == true)
			return captureID.isEmpty() ? 1 : 2;
		else
			return 0;
	}

	public Integer getExitCode()
	{
		return m_processInputReader.getExitCode();
	}

	public int getStatus(long id)
	{
		if ((m_processInputReader == null) || (m_processInputReader.getExitCode() != null))
			return -2;

		while(true)
		{
			if(m_processInputReader.isResponse())
			{
				String resp = m_processInputReader.getResponse();

				Long id2 = m_captureEngine.getRecordingError(resp);
				if (id2 != null)
					errorID.add(id2);
				id2 = m_captureEngine.getPidChange(resp);
				if (id2 != null)
					m_pidChangeIds.add(id2);
			}
			else
				break;
		}

		if (errorID.contains(id))
		{
			errorID.remove(id);
			return -1;
		}

		return 0;
	}

	public boolean isSurf()
	{
		DataStore store = DataStore.getInstance();

		Iterator<Long> iter = captureID.iterator();
		while (iter.hasNext())
		{
			ScheduleItem item = store.getScheduleItem(iter.next().toString());
			if ( (item != null) && item.isSurf() )
				return true;
		}

		return false;
	}

	public boolean isPreviewOrSurfOrEpgScan()
	{
		if (m_bEpgScanRunning)
			return true;

		DataStore store = DataStore.getInstance();

		Iterator<Long> iter = captureID.iterator();
		while (iter.hasNext())
		{
			ScheduleItem item = store.getScheduleItem(iter.next().toString());
			if ( (item != null) && (item.isSurf() || item.isPreview() || item.isPlaylist()) )
				return true;
		}

		return false;
	}

	public int changeChannel(ScheduleItem item, CaptureFile captureFile, Channel channel)
	{
		if(m_processInputReader == null || m_processInputReader.getExitCode() != null || !m_captureEngine.isCanRecordMultiple())
			return -1;

		long id = item.getID();

		try
		{
			errorID.remove(id);
			m_pidChangeIds.remove(id);
			//captureID.remove(id);

			if (m_captureEngine.isStdin())
			{
				String command = getCommand(item, channel, captureFile, captureFile.getChangeChannelCommand(m_captureEngine));
				command = Formatter.formatCommand(command);
				DVScheduler.log("Change channel: " + command);
				m_out.write(command.getBytes());
				m_out.flush();
			}

			if (m_captureEngine.isStdout())
			{
				int timeOut = m_timeoutChangeChannel;

				reader:
					while(true)
					{
						while (m_processInputReader.isResponse())
						{
							String resp = m_processInputReader.getResponse();

							if (m_captureEngine.isRecordingStarted(resp, id))
							{
								break reader;
							}
							else if (m_captureEngine.isRecordingError(resp, id))
							{
								DVScheduler.log("Error with change channel command to external app!");
								m_out.write((m_captureEngine.getStopRecordingCommand().replace("$rid", Long.toString(id))).getBytes());
								m_out.flush();
								return -1008;
							}
							else
							{
								Long id2 = m_captureEngine.getRecordingError(resp);
								if (id2 != null)
									errorID.add(id2);
								id2 = m_captureEngine.getPidChange(resp);
								if (id2 != null)
									m_pidChangeIds.add(id2);
							}
						}

						if(m_processInputReader.getExitCode() != null)
						{
							m_bCaptureRunning = false;
							int code = m_processInputReader.getExitCode().intValue();

							if(code == -1)
								return -10001;
							else if(code == -2)
								return -10002;
							else if(code == -3)
								return -10003;
							else if(code == -4)
								return -10004;               
							else
								return -11111;
						}
						else
						{
							try
							{
								m_processInputReader.waitResponse();
								//Thread.sleep(100);
								timeOut--;
							}
							catch(Exception e){}
						}

						if(timeOut == 0)
						{
							DVScheduler.log("Timeout reached sending change channel command to external app!");
							m_out.write((m_captureEngine.getStopRecordingCommand().replace("$rid", Long.toString(id))).getBytes());
							m_out.flush();
							return -1006;
						}
					}
			}
		}
		catch(Exception e)
		{
			DVScheduler.log("Error sending change channel command to external app!");
			e.printStackTrace();
			return -1007;
		}

		//totalCapture++;
		//captureID.add(id);

		return 0;
	}

	public int changeSetting(ScheduleItem item, CaptureFile captureFile, Channel channel)
	{
		if(m_processInputReader == null || m_processInputReader.getExitCode() != null || !m_captureEngine.isCanRecordMultiple())
			return -1;

		if (captureID.size() > 1)
			return -2;

		long id = item.getID();

		try
		{
			errorID.remove(id);
			m_pidChangeIds.remove(id);
			//captureID.remove(id);

			if (m_captureEngine.isStdin())
			{
				// Add EPG stop command
				//if (m_bEpgScanRunning)
				{
					String command = m_captureEngine.getStopEpgScanCommand();
					DVScheduler.log("EPG stop command: " + command);
					if ((command != null) && !command.isEmpty())
						m_out.write(command.getBytes());
					m_bEpgScanRunning = false;
				}

				String command = getCommand(item, channel, captureFile, captureFile.getChangeSettingCommand(m_captureEngine));
				command = Formatter.formatCommand(command);
				DVScheduler.log("Change setting: " + command);
				m_out.write(command.getBytes());

				// Add EPG scan command
				command = getEpgScanCommand(channel);
				DVScheduler.log("EPG scan command: " + command);
				if ((command != null) && !command.isEmpty())
					m_out.write(command.getBytes());

				m_out.flush();
			}

			if (m_captureEngine.isStdout())
			{
				int timeOut = m_timeoutChangeSetting;

				reader:
					while(true)
					{
						while (m_processInputReader.isResponse())
						{
							String resp = m_processInputReader.getResponse();

							if (m_captureEngine.isRecordingStarted(resp, id))
							{
								break reader;
							}
							else if (m_captureEngine.isRecordingError(resp, id))
							{
								DVScheduler.log("Error with change setting command to external app!");
								m_out.write((m_captureEngine.getStopRecordingCommand().replace("$rid", Long.toString(id))).getBytes());
								m_out.flush();
								return -1008;
							}
							else
							{
								Long id2 = m_captureEngine.getRecordingError(resp);
								if (id2 != null)
									errorID.add(id2);
								id2 = m_captureEngine.getPidChange(resp);
								if (id2 != null)
									m_pidChangeIds.add(id2);
							}
						}

						if(m_processInputReader.getExitCode() != null)
						{
							m_bCaptureRunning = false;
							int code = m_processInputReader.getExitCode().intValue();

							if(code == -1)
								return -10001;
							else if(code == -2)
								return -10002;
							else if(code == -3)
								return -10003;
							else if(code == -4)
								return -10004;               
							else
								return -11111;
						}
						else
						{
							try
							{
								m_processInputReader.waitResponse();
								//Thread.sleep(100);
								timeOut--;
							}
							catch(Exception e){}
						}

						if(timeOut == 0)
						{
							DVScheduler.log("Timeout reached sending change setting command to external app!");
							m_out.write((m_captureEngine.getStopRecordingCommand().replace("$rid", Long.toString(id))).getBytes());
							m_out.flush();
							return -1006;
						}
					}
			}
		}
		catch(Exception e)
		{
			DVScheduler.log("Error sending change setting command to external app!");
			e.printStackTrace();
			return -1007;
		}

		//totalCapture++;
		//captureID.add(id);

		m_captureDevice.setSetting(channel.getSetting(), channel.getDemodulatorName());

		return 0;
	}

	public int changeInitialSetting(ScheduleItem item, CaptureFile captureFile, Channel channel, StringBuffer log)
	{
		if(m_processInputReader == null || m_processInputReader.getExitCode() != null || !m_captureEngine.isCanRecordMultiple())
			return -1;

		if (captureID.size() > 0)
			return -2;

		long id = item.getID();

		try
		{
			errorID.remove(id);
			m_pidChangeIds.remove(id);
			captureID.remove(id);

			if (m_captureEngine.isStdin())
			{
				// Add EPG stop command
				//if (m_bEpgScanRunning)
				{
					String command = m_captureEngine.getStopEpgScanCommand();
					DVScheduler.log("EPG stop command: " + command);
					if ((command != null) && !command.isEmpty())
						m_out.write(command.getBytes());
					m_bEpgScanRunning = false;
				}
				
				if ( !channel.getSetting().equals(m_captureDevice.getSetting()) )
				{
					String command = getCommand(item, channel, captureFile, m_captureEngine.getChangeSettingCommand());
					command = Formatter.formatCommand(command);
					DVScheduler.log("Initial change setting: " + command);
					m_out.write(command.getBytes());
				}

				String command = getCommand(item, channel, captureFile, captureFile.getCommandLine(m_captureEngine));
				command = Formatter.formatCommand(command);
				DVScheduler.log("Command: " + command);
				m_out.write(command.getBytes());

				// Add EPG scan command
				command = getEpgScanCommand(channel);
				DVScheduler.log("EPG scan command: " + command);
				if ((command != null) && !command.isEmpty())
					m_out.write(command.getBytes());

				m_out.flush();
			}

			if (m_captureEngine.isStdout())
			{
				int timeOut = m_timeoutChangeInitialSetting;

				reader:
					while(true)
					{
						while (m_processInputReader.isResponse())
						{
							String resp = m_processInputReader.getResponse();

							if (m_captureEngine.isRecordingStarted(resp, id))
							{
								break reader;
							}
							else if (m_captureEngine.isRecordingError(resp, id))
							{
								DVScheduler.log("Error with initial change setting command to external app!");
								m_out.write((m_captureEngine.getExitCommand()).getBytes());
								m_out.flush();
								m_bCaptureRunning = false;
							}
							else
								log.append(resp + "\n");
						}

						if(m_processInputReader.getExitCode() != null)
						{
							m_bCaptureRunning = false;
							int code = m_processInputReader.getExitCode().intValue();

							if(code == -1)
								return -10001;
							else if(code == -2)
								return -10002;
							else if(code == -3)
								return -10003;
							else if(code == -4)
								return -10004;               
							else
								return -11111;
						}
						else
						{
							try
							{
								m_processInputReader.waitResponse();
								//Thread.sleep(100);
								timeOut--;
							}
							catch(Exception e){}
						}

						if(timeOut == 0)
						{
							DVScheduler.log("Timeout reached sending initial change setting command to external app!");
							if(m_captureProcess != null)
								m_captureProcess.destroy();
							m_bCaptureRunning = false;
							return -1006;
						}
					}
			}
		}
		catch(Exception e)
		{
			DVScheduler.log("Error sending initial change setting command to external app!");
			e.printStackTrace();
			m_bCaptureRunning = false;
			try
			{
				if(m_captureProcess != null)
					m_captureProcess.destroy();
			}
			catch(Exception e2){}
			return -1007;
		}

		m_bCaptureRunning = true;
		m_totalCapture++;
		captureID.add(id);

		m_captureDevice.setSetting(channel.getSetting(), channel.getDemodulatorName());

		return 0;
	}

	public String getClientGetAddress(ScheduleItem item, Channel ch, CaptureFile capFile)
	{
		return getCommand(item, ch, capFile, capFile.getClientGetAddress(m_captureEngine));
	}

	public String getSignalStatistic()
	{
		if(m_processInputReader == null || m_processInputReader.getExitCode() != null)
			return null;

		try
		{
			if (m_captureEngine.isStdin())
			{
				String command = Formatter.formatCommand(m_captureEngine.getSignalStatisticCommand());
				//DVScheduler.log("Calling signal statistic:" + command);
				m_out.write(command.getBytes());
				m_out.flush();
			}

			if (m_captureEngine.isStdout())
			{
				int timeOut = m_timeoutAction;

				while(true)
				{
					while (m_processInputReader.isResponse())
					{
						String resp = m_processInputReader.getResponse();

						Long id2 = m_captureEngine.getRecordingError(resp);
						if (id2 != null)
							errorID.add(id2);
						id2 = m_captureEngine.getPidChange(resp);
						if (id2 != null)
							m_pidChangeIds.add(id2);						
						else
							return resp;
					}

					if(m_processInputReader.getExitCode() != null)
					{
						m_bCaptureRunning = false;
						return null;
					}
					else
					{
						try
						{
							m_processInputReader.waitResponse();
							//Thread.sleep(100);
							timeOut--;
						}
						catch(Exception e){}
					}

					if(timeOut == 0)
					{
						DVScheduler.log("Timeout reached sending signal statistic command to external app!");
						return null;
					}
				}
			}
		}
		catch(Exception e)
		{
			DVScheduler.log("Error with signal statistic command!");
			e.printStackTrace();
			return null;
		}

		return null;
	}
}
