/*
   Copyright (C) 2007-2010 contributors

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

 */

import java.util.*;
import java.io.*;
import java.text.*;

class CaptureTask
{
	private ScheduleItem m_item = null;
	private DataStore m_store = null;
	private long m_start = -1;
	private long m_stop = -1;
	private CaptureFile m_captureFile = null;
	private CaptureProcess m_capture = null;
	private CaptureEngine captureEngine = new CaptureEngine();
	private long m_lastResponseTime = 0;
	private long m_keepAliveTime = 0;
	private long m_idleTimeout = 0;
	private boolean m_bLiveToNormal = false;
	private int m_streamPort = -1;
	private String m_createFileName = "";
	private String m_createFileType = "";
	private String m_createFileContent = "";
	private String m_redirect = "";

	protected static String DEVICE_RUNNING = "1";
	protected static String DEVICE_RUNNINGSURF = "2";
	protected static String DEVICE_IDLE = "3";	// Idle, same setting
	protected static String DEVICE_IDLEHARD = "4";	// Idle, different setting but same hard setting
	protected static String DEVICE_FREE = "5";

	public CaptureTask(ScheduleItem it) throws Exception
	{
		m_store = DataStore.getInstance();
		m_item = it;
	}

	public ScheduleItem getScheduleItem()
	{
		return m_item;
	}

	public Channel getBestChannel(Channel ch) throws ResultException
	{
		return getBestChannel(ch, false, m_capture == null ? null : m_capture.getCaptureDevice());
	}

	private Channel getBestChannel(Channel ch, boolean bSetState, CaptureDevice changeableDevice) throws ResultException
	{
		Vector<String> channels = ch.getChannels();
		if (channels.isEmpty())
		{
			if (!ch.isDataValid())
			{
				m_item.setStatus("Error channel object has invalid data");
				m_item.log("Channel object has invalid configuration data");
				if (bSetState)
				{
					m_item.incrementErrorCount();
					m_item.setState(ScheduleItem.ERROR);
					m_store.m_nRecordingErrorCount++;
					m_item.notifyResponse();
				}
				throw new ResultException(-2);
			}

			return ch;
		}
		else
		{
			Vector<ScheduleItem> overlaps = new Vector<ScheduleItem>();
			ScheduleItem[] items = (ScheduleItem[]) m_store.getScheduleList().values().toArray(new ScheduleItem[0]);
			for (int x = 0; x < items.length; x++)
			{
				int state = items[x].getState();
				if ( (state == ScheduleItem.WAITING || state == ScheduleItem.RUNNING || state == ScheduleItem.RESTART) && 
						(items[x].toString().equals(m_item.toString()) == false) && (m_item.isOverlapping(items[x])) )
				{
					overlaps.add(items[x]);
				}
			}

			m_item.log("Number of immediate overlapping schedules:" + overlaps.size());

			Vector<String> availableDevices = new Vector<String>();
			Channel firstChannel = null;
			for (int i=0; i<channels.size(); i++)
			{
				Channel ch2 = m_store.getChannel(channels.get(i));
				if ( (ch2 == null) || !ch2.isDataValid() )
					continue;

				if (firstChannel == null)
					firstChannel = ch2;

				getDevicesStatus(ch2, String.format("%1$08x", i), overlaps, availableDevices, changeableDevice);
			}

			if (availableDevices.size() == 1)
			{
				m_item.log("Found one available device-channel pair");
				String[] data = availableDevices.firstElement().split("\n");
				ch = m_store.getChannel(data[4]);
				return ch;
			}
			else if (!availableDevices.isEmpty())
			{
				m_item.log("Found " + availableDevices.size() + " available device-channel pairs");
				Collections.sort(availableDevices);
				Vector<String> freeDevices = new Vector<String>();
				CaptureDeviceList devList = CaptureDeviceList.getInstance();
				for(int x = 0; x < availableDevices.size(); x++)
				{
					String[] data = availableDevices.get(x).split("\n");
					int deviceIndex = Integer.parseInt(data[3], 16);
					CaptureDevice capDev2 = devList.getDevice(deviceIndex);
					m_item.log((x+1) + ". Potential conflict:" + Integer.parseInt(data[0], 16) + ", Device:" + capDev2.getName() + ", Channel:" + data[4]);
					if (data[1].equals(DEVICE_FREE))
					{
						if (m_item.isCaptureDeviceUsed(capDev2.getNumber()) == false)
						{
							ch = m_store.getChannel(data[4]);
							return ch;
						}
						else
						{
							m_item.log("Skipping device due to previous failure");
							freeDevices.add(availableDevices.get(x));
						}
					}
					else
					{
						ch = m_store.getChannel(data[4]);
						return ch;
					}
				}

				m_item.log("All free device-channel pair(s) have failed previously, choosing one randomly");
				int freeIndex = m_store.m_rand.nextInt(freeDevices.size());

				String[] data = freeDevices.get(freeIndex).split("\n");
				ch = m_store.getChannel(data[4]);
				return ch;
			}
			else if (firstChannel != null)
			{
				if (m_item.isLive())
				{
					m_item.setStatus("No channel has an available device");
					DVScheduler.log("No channel has an available device");
					if (bSetState)
					{
						m_item.setState(ScheduleItem.ERROR);
						m_item.notifyResponse();
					}
					throw new ResultException(-4);
				}
				else
				{
					m_item.setStatus("Waiting for retry");
					m_item.log("No channel has an available device, will try again");
					if (bSetState)
					{
						m_item.setState(ScheduleItem.WAITING);
						m_item.notifyResponse();
					}
					throw new ResultException(-4);
				}
			}
			else
			{
				m_item.incrementErrorCount();
				m_item.setState(ScheduleItem.ERROR);
				m_item.setStatus("Error no valid channel");
				m_item.log("Error no valid channel");
				m_store.m_nRecordingErrorCount++;
				m_item.notifyResponse();
				throw new ResultException(-2);
			}
		}
	}

	public int startCapture()
	{
		//
		// Get the Channel info
		//
		Channel ch = m_store.getChannel(m_item.getChannel());

		if(ch == null)
		{
			m_item.incrementErrorCount();
			m_item.setState(ScheduleItem.ERROR);
			m_item.setStatus("Error channel not found");
			m_item.log("Channel object was not found in Channel List");
			m_store.m_nRecordingErrorCount++;
			m_item.notifyResponse();
			return -2;
		}

		m_item.log("Channel data loaded from Item : " + ch.getName());
		Vector<String> channels = ch.getChannels();
		if (channels.isEmpty())
		{
			if (!ch.isDataValid())
			{
				m_item.incrementErrorCount();
				m_item.setState(ScheduleItem.ERROR);
				m_item.setStatus("Error channel object has invalid data");
				m_item.log("Channel object has invalid configuration data");
				m_store.m_nRecordingErrorCount++;
				m_item.notifyResponse();
				return -2;
			}
			m_item.setRunningChannel(ch.getName());

			// This must be called first, as it sets the capture engine.
			// However, ScheduleItem.setRunningchannel(.) must be called first before getFullFileName()
			m_captureFile = getFullFileName();

		}
		// Channel group.
		else
		{
			// We just set to first channel for now to read the capture engine, as all channels in the group must use the same engine.
			m_item.setRunningChannel(channels.firstElement());

			// This must be called first, as it sets the capture engine.
			// However, ScheduleItem.setRunningchannel(.) must be called first before getFullFileName()
			m_captureFile = getFullFileName();

			if (m_captureFile == null)
			{
				m_item.notifyResponse();
				return -1;
			}

			try
			{
				ch = getBestChannel(ch, true, null);
			}
			catch (ResultException e)
			{
				return e.getResult();
			}

			m_item.setRunningChannel(ch.getName());
			m_captureFile = getFullFileName();
		}

		if (m_captureFile == null)
		{
			m_item.notifyResponse();
			return -1;
		}

		String dirString = m_captureFile.getDirectory();
		if ( (dirString != null) && !dirString.isEmpty() && !dirExists(dirString) )
		{
			m_item.incrementErrorCount();
			m_item.setState(ScheduleItem.ERROR);
			m_item.setStatus("Error capture directory");
			m_item.log("The capture directory (" + dirString
					+ ") does not exist, can not capture");
			m_store.m_nRecordingErrorCount++;
			m_item.notifyResponse();
			return -1;
		}

		writeLog(ch);
		return startCapture(ch);
	}

	protected void writeLog(Channel ch)
	{
		DVScheduler.log("About to start capture");

		m_item.log("Schedule Name : " + m_item.getName());
		m_item.log("File Pattern : " + m_item.getFilePattern());

		m_item.log("File Name : " + new File(m_captureFile.getFullPath()).getName());
		m_item.log("File Path : " + new File(m_captureFile.getFullPath()).getParent());

		m_item.log("Channel Name : " + ch.getName());
		m_item.log("Setting : " + (ch.getDemodulatorName().isEmpty() ? "" : ch.getDemodulatorName() + " : ") + ch.getSetting());
		//item.log("Frequency : " + ch.getFrequency());
		//item.log("Bandwidth : " + ch.getBandWidth());
		m_item.log("Channel : " + ch.getSelectionDescription());
		//item.log("Prog ID   : " + ch.getProgramID());
		//item.log("Video Pid : " + ch.getVideoPid());
		//item.log("Audio Pid : " + ch.getAudioPid());
		m_item.log("Capture Type : " + m_item.getCaptureCommandFile());
	}

	protected int getMax(int val1, int val2)
	{
		return val1 > val2 ? val1 : val2;
	}

	protected long getMax(long val1, long val2)
	{
		return val1 > val2 ? val1 : val2;
	}

	protected int getMinConflictChannelGroup(String netType, String setting, Set<String> conflictSettings, Vector<ScheduleItem> overlaps, int index)
	{
		int minCount = Integer.MAX_VALUE;
		ScheduleItem item2 = overlaps.get(index);
		Channel ch2 = m_store.getChannel(item2.getChannel());
		Vector<String> childs = ch2.getChannels();
		for (int i=0; i<childs.size(); i++)
		{
			Set<String> conflictSettings2 = new HashSet<String>();
			conflictSettings2.addAll(conflictSettings);
			ch2 = m_store.getChannel(childs.get(i));
			if (ch2 == null || !ch2.getNetworkType().equals(netType))
				continue;
			String setting2 = ch2.getSetting();
			int count = 0;

			if ( !setting.equals(setting2) && !conflictSettings2.contains(setting2) )
			{
				conflictSettings2.add(setting2);
				count++;
			}

			for (int j=index+1; j<overlaps.size(); j++)
			{
				item2 = overlaps.get(j);
				ch2 = m_store.getChannel(item2.getChannel());
				if (ch2 == null || !ch2.getNetworkType().equals(netType))
					continue;
				else if ( ch2.isGroup() && (item2.getState() != ScheduleItem.RUNNING) )
					count += getMinConflictChannelGroup(netType, setting, conflictSettings2, overlaps, j);
				else
				{
					if ( ch2.isGroup() && (item2.getState() == ScheduleItem.RUNNING) )
						ch2 = m_store.getChannel(item2.getRunningChannel());

					setting2 = ch2.getSetting();
					if ( !setting.equals(setting2) && !conflictSettings2.contains(setting2) )
					{
						conflictSettings2.add(setting2);
						count++;
					}
				}
			}

			if (count < minCount)
				minCount = count;
		}

		return minCount;
	}

	protected int getMinConflict(String netType, String setting, Vector<ScheduleItem> overlaps)
	{
		CaptureDeviceList devList = CaptureDeviceList.getInstance();
		Set<String> conflictSettings = new HashSet<String>();
		int count = 1 - devList.getDeviceCount(netType);

		for (int j=0; j<overlaps.size(); j++)
		{
			ScheduleItem item2 = overlaps.get(j);
			Channel ch2 = m_store.getChannel(item2.getChannel());
			if (ch2 == null || !ch2.getNetworkType().equals(netType))
				continue;
			else if ( ch2.isGroup() && (item2.getState() != ScheduleItem.RUNNING) )
				count += getMinConflictChannelGroup(netType, setting, conflictSettings, overlaps, j);
			else
			{
				if ( ch2.isGroup() && (item2.getState() == ScheduleItem.RUNNING) )
					ch2 = m_store.getChannel(item2.getRunningChannel());

				String setting2 = ch2.getSetting();
				if ( !setting.equals(setting2) && !conflictSettings.contains(setting2) )
				{
					conflictSettings.add(setting2);
					count++;
				}
			}
		}

		return count < 0 ? 0 : count;
	}

	protected void getDevicesStatus(Channel ch, String channelOrder, Vector<ScheduleItem> overlaps, Vector<String> availableDevices, CaptureDevice changeableDevice)
	{
		String netType = ch.getNetworkType();
		String setting = ch.getSetting();
		CaptureDeviceList devList = CaptureDeviceList.getInstance();

		// Find a suitable capture device.
		for(int x = 0; x < devList.getDeviceCount(); x++)
		{
			CaptureDevice cap = devList.getDevice(x);
			CaptureEngine captureEngine2 = cap.getCaptureProcess() == null ? null : cap.getCaptureProcess().getCaptureEngine();
			int capStatus = cap.getStatus();
			if ( (capStatus == 2) && (cap == changeableDevice) && (cap.getCaptureProcess().getTotalCapture() <= 1) )
				capStatus = 1;

			if( (capStatus == 2) && cap.getSetting().equals(setting) && 
					(captureEngine.getName().equals(captureEngine2.getName())) && 
					captureEngine.isCanRecordMultiple() && captureEngine2.isCanRecordMultiple() )
			{
				if (cap.getCaptureProcess().isSurf())
				{
					availableDevices.add(String.format("%1$08x", getMinConflict(netType, setting, overlaps)) + "\n" + DEVICE_RUNNINGSURF + "\n" + channelOrder + "\n" + String.format("%1$08x", x) + "\n" + ch.getName());
				}
				else
				{
					availableDevices.add(String.format("%1$08x", getMinConflict(netType, setting, overlaps)) + "\n" + DEVICE_RUNNING + "\n" + channelOrder + "\n" + String.format("%1$08x", x) + "\n" + ch.getName());
				}
			}
			else if ( (capStatus == 1) && cap.getSetting().equals(setting) && 
					captureEngine.getName().equals(captureEngine2.getName()) && 
					captureEngine.isCanRecordMultiple() && captureEngine2.isCanRecordMultiple() )
			{
				availableDevices.add(String.format("%1$08x", getMinConflict(netType, setting, overlaps)) + "\n" + DEVICE_IDLE + "\n" + channelOrder + "\n" + String.format("%1$08x", x) + "\n" + ch.getName());
			}
			else if ( (capStatus == 1) && cap.isHardSettingCompatible(ch) && 
					captureEngine.getName().equals(captureEngine2.getName()) && 
					captureEngine.isCanRecordMultiple() && captureEngine2.isCanRecordMultiple() )
			{
				availableDevices.add(String.format("%1$08x", getMinConflict(netType, setting, overlaps)) + "\n" + DEVICE_IDLEHARD + "\n" + channelOrder + "\n" + String.format("%1$08x", x) + "\n" + ch.getName());
			}
			else if ( cap.isNetworkTypeCompatible(netType) && (capStatus == 0) )
			{
				availableDevices.add(String.format("%1$08x", getMinConflict(netType, setting, overlaps)) + "\n" + DEVICE_FREE + "\n" + channelOrder + "\n" + String.format("%1$08x", x) + "\n" + ch.getName());
			}
		}
	}

	protected int startCapture(Channel ch)
	{
		m_item.setRunningChannel(ch.getName());
		CaptureDeviceList devList = CaptureDeviceList.getInstance();
		StringBuffer log = new StringBuffer();
		int startCode = -1;
		if (m_streamPort == 0)
		{
			m_streamPort = m_store.getStreamPort();
			m_captureFile.setStreamPort(m_streamPort);
		}

		// Find a suitable capture device.
		Vector<CaptureDevice> capFreeList = new Vector<CaptureDevice>();
		Vector<CaptureDevice> capIdleList = new Vector<CaptureDevice>();
		Vector<CaptureDevice> capIdleHardList = new Vector<CaptureDevice>();
		Vector<CaptureDevice> capSurfList = new Vector<CaptureDevice>();
		for(int x = 0; x < devList.getDeviceCount(); x++)
		{
			CaptureDevice cap = devList.getDevice(x);
			CaptureEngine captureEngine2 = cap.getCaptureProcess() == null ? null : cap.getCaptureProcess().getCaptureEngine() ;

			if( (cap.getStatus() == 2) && cap.getSetting().equals(ch.getSetting()) && 
					(captureEngine.getName().equals(captureEngine2.getName())) && 
					captureEngine.isCanRecordMultiple() && captureEngine2.isCanRecordMultiple() )
			{
				if (cap.getCaptureProcess().isSurf())
				{
					capSurfList.add(cap);
				}
				else
				{
					if (!m_item.isLive())
						sendEmail("email.captureStart", "DV Scheduler about to capture using a device already in use");
					log = new StringBuffer();
					m_capture = cap.getCaptureProcess();
					m_item.log("Capture Device : " + m_capture.getCaptureDevice().getName());
					startCode = m_capture.addCapture(
							m_item, 
							m_captureFile,
							ch,
							log);

					if(startCode == 0)
					{
						m_start = System.nanoTime();
						m_item.log("Capture added with code : " + startCode);

						m_item.log("Start Log:\n\n" + log.toString());

						if (!m_item.isLive() && m_store.getProperty("Capture.details", "none").endsWith("beginning"))
							new CaptureDetails().writeCaptureDetails(this, m_capture.getCaptureDevice());

						m_lastResponseTime = System.nanoTime();

						m_item.setState(ScheduleItem.RUNNING);
						m_item.setRunningStatus(new Date());
						m_item.log("Capture added and started");

						m_store.m_captureTasks.put(m_item.toString(), this);
						m_item.notifyResponse();
						//return capture.getCaptureDevice().getID();
						return 1;
					}
					else if (m_item.isLive())
					{
						releaseStreamPort();
						m_item.incrementErrorCount();
						m_item.setStatus("Error adding capture");
						m_item.setState(ScheduleItem.ERROR);
						DVScheduler.log("Error adding capture");
						m_store.m_nRecordingErrorCount++;
						m_item.notifyResponse();
						return -3;
					}
					else
					{
						releaseStreamPort();
						m_item.incrementErrorCount();
						m_item.setStatus("Waiting for retry");
						m_item.log("Error adding capture using (" + m_capture.getCaptureDevice().getName() + ")");
						m_item.setState(ScheduleItem.WAITING);
						m_store.m_nRecordingErrorCount++;
						m_item.notifyResponse();
						return -3;
					}
				}
			}
			else if ( (cap.getStatus() == 1) && cap.getSetting().equals(ch.getSetting()) && 
					captureEngine.getName().equals(captureEngine2.getName()) && 
					captureEngine.isCanRecordMultiple() && captureEngine2.isCanRecordMultiple() )
			{
				capIdleList.add(cap);
			}
			else if ( (cap.getStatus() == 1) && cap.isHardSettingCompatible(ch) && 
					captureEngine.getName().equals(captureEngine2.getName()) && 
					captureEngine.isCanRecordMultiple() && captureEngine2.isCanRecordMultiple() )
			{
				capIdleHardList.add(cap);
			}
			else if ( cap.isNetworkTypeCompatible(ch.getNetworkType()) && (cap.getStatus() == 0) )
			{
				capFreeList.add(cap);
			}
		}

		if (!capSurfList.isEmpty())
		{
			if (!m_item.isLive())
				sendEmail("email.captureStart", "DV Scheduler about to capture using a device being used for Channel Surfing");
			log = new StringBuffer();
			m_capture = capSurfList.firstElement().getCaptureProcess();
			m_item.log("Capture Device (surf) : " + m_capture.getCaptureDevice().getName());
			startCode = m_capture.addCapture(
					m_item, 
					m_captureFile,
					ch,
					log);

			if(startCode == 0)
			{
				m_start = System.nanoTime();
				m_item.log("Capture added in surf with code : " + startCode);

				m_item.log("Start Log:\n\n" + log.toString());

				if (!m_item.isLive() && m_store.getProperty("Capture.details", "none").endsWith("beginning"))
					new CaptureDetails().writeCaptureDetails(this, m_capture.getCaptureDevice());

				m_lastResponseTime = System.nanoTime();

				m_item.setState(ScheduleItem.RUNNING);
				m_item.setRunningStatus(new Date());
				m_item.log("Capture added in surf and started");

				m_store.m_captureTasks.put(m_item.toString(), this);
				m_item.notifyResponse();
				//return capture.getCaptureDevice().getID();
				return 1;
			}
			else if (m_item.isLive())
			{
				releaseStreamPort();
				m_item.incrementErrorCount();
				m_item.setStatus("Error adding capture in surf");
				m_item.setState(ScheduleItem.ERROR);
				DVScheduler.log("Error adding capture in surf");
				m_store.m_nRecordingErrorCount++;
				m_item.notifyResponse();
				return -3;
			}
			else
			{
				releaseStreamPort();
				m_item.incrementErrorCount();
				m_item.setStatus("Waiting for retry");
				m_item.log("Error adding capture in surf using (" + m_capture.getCaptureDevice().getName() + ")");
				m_item.setState(ScheduleItem.WAITING);
				m_store.m_nRecordingErrorCount++;
				m_item.notifyResponse();
				return -3;
			}
		}
		else if (!capIdleList.isEmpty())
		{
			if (!m_item.isLive())
				sendEmail("email.captureStart", "DV Scheduler about to capture using an idle device with same setting");
			log = new StringBuffer();
			m_capture = capIdleList.firstElement().getCaptureProcess();
			m_item.log("Capture Device (idle) : " + m_capture.getCaptureDevice().getName());
			startCode = m_capture.changeInitialSetting(
					m_item, 
					m_captureFile,
					ch,
					log);

			if(startCode == 0)
			{
				m_start = System.nanoTime();
				m_item.log("Capture added idle with code : " + startCode);

				m_item.log("Start Log:\n\n" + log.toString());

				if (!m_item.isLive() && m_store.getProperty("Capture.details", "none").endsWith("beginning"))
					new CaptureDetails().writeCaptureDetails(this, m_capture.getCaptureDevice());

				m_lastResponseTime = System.nanoTime();

				m_item.setState(ScheduleItem.RUNNING);
				m_item.setRunningStatus(new Date());
				m_item.log("Capture started");

				m_store.m_captureTasks.put(m_item.toString(), this);
				m_item.notifyResponse();
				//return capture.getCaptureDevice().getID();
				return 1;
			}
			else if (m_item.isLive())
			{
				releaseStreamPort();
				m_item.incrementErrorCount();
				m_item.setStatus("Error adding capture idle");
				m_item.setState(ScheduleItem.ERROR);
				DVScheduler.log("Error adding capture idle");
				m_store.m_nRecordingErrorCount++;
				m_item.notifyResponse();
				return -3;
			}
			else
			{
				releaseStreamPort();
				m_item.incrementErrorCount();
				m_item.setStatus("Waiting for retry");
				m_item.log("Error adding capture idle using (" + m_capture.getCaptureDevice().getName() + ")");
				m_item.setState(ScheduleItem.WAITING);
				m_store.m_nRecordingErrorCount++;
				m_item.notifyResponse();
				return -3;
			}
		}
		else if (!capIdleHardList.isEmpty())
		{
			if (!m_item.isLive())
				sendEmail("email.captureStart", "DV Scheduler about to capture using an idle device with different setting");
			log = new StringBuffer();
			m_capture = capIdleHardList.firstElement().getCaptureProcess();
			m_item.log("Capture Device (hard idle) : " + m_capture.getCaptureDevice().getName());
			startCode = m_capture.changeInitialSetting(
					m_item, 
					m_captureFile,
					ch,
					log);

			if(startCode == 0)
			{
				m_start = System.nanoTime();
				m_item.log("Change setting succeeded with code : " + startCode);

				m_item.log("Start Log:\n\n" + log.toString());

				if (!m_item.isLive() && m_store.getProperty("Capture.details", "none").endsWith("beginning"))
					new CaptureDetails().writeCaptureDetails(this, m_capture.getCaptureDevice());

				m_lastResponseTime = System.nanoTime();

				m_item.setState(ScheduleItem.RUNNING);
				m_item.setRunningStatus(new Date());
				m_item.log("Capture started");

				m_store.m_captureTasks.put(m_item.toString(), this);
				m_item.notifyResponse();
				//return capture.getCaptureDevice().getID();
				return 1;
			}
			else if (m_item.isLive())
			{
				releaseStreamPort();
				m_item.incrementErrorCount();
				m_item.setStatus("Error changing setting");
				m_item.setState(ScheduleItem.ERROR);
				DVScheduler.log("Error changing setting");
				m_store.m_nRecordingErrorCount++;
				m_item.notifyResponse();
				return -3;
			}
			else
			{
				releaseStreamPort();
				m_item.incrementErrorCount();
				m_item.setStatus("Waiting for retry");
				m_item.log("Error changing setting using (" + m_capture.getCaptureDevice().getName() + ")");
				m_item.setState(ScheduleItem.WAITING);
				m_store.m_nRecordingErrorCount++;
				m_item.notifyResponse();
				return -3;
			}
		}
		// See if we have a free capture device.
		else if (!capFreeList.isEmpty())
		{
			m_capture = null;
			boolean bFound = false;
			for (int x = 0; x < capFreeList.size(); x++)
			{
				CaptureDevice capDev2 = capFreeList.get(x);

				if (m_item.isCaptureDeviceUsed(capDev2.getNumber()) == false)
				{
					if (x > 0)
					{
						m_item.log("First " + x + " available device(s) have failed previously, using next available device (" + capDev2.getName() + ")");
					}
					bFound = true;
					m_item.addCaptureDeviceUsed(capDev2.getNumber());

					if (m_item.isLive())
					{
						runPresetTask(capDev2, "Tasks.PreLiveTask");
					}
					else
					{
						sendEmail("email.captureStart", "DV Scheduler about to capture using a free device");
						runPresetTask(capDev2, "Tasks.PreTask");
					}
					m_capture = new CaptureProcess(capDev2);
					break;
				}
			}

			// Choose one randomly.
			if (!bFound)
			{
				m_item.log("All available device(s) have failed previously, choosing one randomly");
				int randInt = m_store.m_rand.nextInt(capFreeList.size());
				if (m_item.isLive())
				{
					runPresetTask(capFreeList.get(randInt), "Tasks.PreLiveTask");
				}
				else
				{
					sendEmail("email.captureStart", "DV Scheduler about to capture, all devices have failed previously");
					runPresetTask(capFreeList.get(randInt), "Tasks.PreTask");
				}
				m_capture = new CaptureProcess(capFreeList.get(randInt));
			}

			m_item.log("Capture Device : " + m_capture.getCaptureDevice().getName());

			log = new StringBuffer();
			startCode = m_capture.startDeviceAndCapture(
					m_item, 
					m_captureFile,
					ch,
					captureEngine,
					log);

			if(startCode == 0)
			{
				m_start = System.nanoTime();
				//capFree.setInUse(true, ch.getFrequency(), capture);

				m_item.log("CaptureProcess started with code : " + startCode);

				m_item.log("Start Log:\n\n" + log.toString());

				if (!m_item.isLive() && m_store.getProperty("Capture.details", "none").endsWith("beginning"))
					new CaptureDetails().writeCaptureDetails(this, m_capture.getCaptureDevice());

				m_lastResponseTime = System.nanoTime();

				m_item.setState(ScheduleItem.RUNNING);
				m_item.setRunningStatus(new Date());
				m_item.log("Capture started");

				m_store.m_captureTasks.put(m_item.toString(), this);
				m_item.notifyResponse();
				//return capture.getCaptureDevice().getID();
				return 1;
			}
			else if (m_item.isLive())
			{
				releaseStreamPort();
				m_item.incrementErrorCount();
				m_item.setStatus("Error starting capture");
				m_item.setState(ScheduleItem.ERROR);
				DVScheduler.log("Error starting capture");
				m_store.m_nRecordingErrorCount++;
				//sendEmail("email.captureError", "DV Scheduler error starting capture");
				//runPresetTask(capture.getCaptureDevice(), "Tasks.StartRecordErrorTask", "DV Scheduler error starting capture");
				m_item.notifyResponse();
				return -3;
			}
			else
			{
				releaseStreamPort();
				m_item.incrementErrorCount();
				m_item.setStatus("Waiting for retry");
				m_item.log("Error starting capture, will try again");
				m_store.m_nRecordingErrorCount++;
				sendEmail("email.captureError", "DV Scheduler error starting capture");
				runPresetTask(m_capture.getCaptureDevice(), "Tasks.StartRecordErrorTask");
				m_item.setState(ScheduleItem.WAITING);
				m_item.notifyResponse();
				return -3;
			}
		}
		else if (m_item.isLive())
		{
			releaseStreamPort();
			m_item.setStatus("No free capture device");
			m_item.setState(ScheduleItem.ERROR);
			DVScheduler.log("No free capture device");
			m_item.notifyResponse();
			return -3;
		}
		else
		{
			releaseStreamPort();
			m_item.setStatus("Waiting for retry");
			m_item.log("No free capture device, will try again");
			m_item.setState(ScheduleItem.WAITING);
			m_item.notifyResponse();
			return -3;
		}
	}

	public boolean isCaptureRunning()
	{
		return m_capture.isCaptureRunning(m_item.getID());
	}

	public boolean isFinished()
	{
		if (m_capture.isCaptureRunning(m_item.getID()) == false)
		{
			if (m_item.isLive() == false)
			{
				m_item.incrementErrorCount();
				m_item.setState(ScheduleItem.RESTART);
				m_item.setStatus("Restarting");
				m_item.log("Capture not running, will restart, app exit code: " + m_capture.getExitCode());
				m_store.m_nRecordingErrorCount++;
				sendEmail("email.captureError", "DV Scheduler error during capture");
				runPresetTask(m_capture.getCaptureDevice(), "Tasks.RecordErrorTask");
			}
			else
			{
				//m_item.incrementErrorCount();
				m_item.abort();
				m_item.setStatus("Aborting");
				m_item.setState(ScheduleItem.ABORTED);
				m_item.log("Item not running, marked for abortion");
				//store.m_nRecordingErrorCount++;
				//sendEmail("email.captureError", "DV Scheduler error during capture");
				//runPresetTask(capture.getCaptureDevice(), "Tasks.RecordErrorTask");
			}
			return true;
		}

		if(m_item.isAborted())
		{
			m_item.log("Action Aborted By User");
			return true;
		}

		if ( m_item.isLive() && (m_idleTimeout > 0) && ((System.nanoTime() - m_lastResponseTime) > m_idleTimeout) )
			return true;

		// is the action finished

		Date now = new Date();
		if (now.after(m_item.getStop()))
		{
			// Add more time.
			if(m_item.isLive())
			{
				try
				{
					m_item.setDuration((int)((now.getTime() - m_item.getStart().getTime()) / (1000 * 60)) + 5);
					return false;
				}
				catch (Exception e01)
				{
				}
			}

			m_item.log("Action Finished");
			return true;
		}

		// We assume the capture is stopped on change in PIDs.
		if (m_capture.isPidChanged(m_item.getID()))
		{
			if (m_item.isLive() == false)
			{
				m_item.incrementErrorCount();
				m_item.setState(ScheduleItem.RESTART);
				m_item.setStatus("Restarting");
				m_item.log("Change in PIDs detected, restarting");
				//store.m_nRecordingErrorCount++;
				runPresetTask(m_capture.getCaptureDevice(), "Tasks.PidChangeTask");
			}
			else
			{
				//m_item.incrementErrorCount();
				m_item.abort();
				m_item.setStatus("Aborting");
				m_item.setState(ScheduleItem.ABORTED);
				m_item.log("Change in PIDs detected, aborting");
				//store.m_nRecordingErrorCount++;
				//runPresetTask(capture.getCaptureDevice(), "Tasks.PidChangeTask");
			}
			return true;
		}

		m_item.setRunningStatus(now);

		/*
		boolean[] writerStatuses = item.getWriterStatuses(now, store);

		if(writerStates == null || isStatusChanged(writerStatuses))
		{
			writerStates = writerStatuses;

			String states = "Setting FileWriters (";
			for(int x = 0; x < writerStatuses.length; x++)
			{
				if(writerStatuses[x])
					states += "1";
				else
					states += "0";
			}
			states += ")";
			item.log(states);  

			int setState = capture.setWriterStates(writerStatuses);
			if(setState != 0)
			{
				item.log("Setting filewriters failed!");
			}
		}
		 */

		return false;
	}

	public void stopCapture()
	{
		// Now try to stop the capture
		// Make sure we clean up and release the card afterwards

		if(m_capture.isProcessRunning())
		{
			int stopCode = m_capture.stopCapture(m_item.getID());
			Integer exitCode = m_capture.getExitCode();
			m_item.log("Capture Stopped(" + stopCode + ") with exit code:" + exitCode);
		}
		else
		{
			Integer exitCode = m_capture.getExitCode();
			m_item.log("Capture already not running, it exited with:" + exitCode);
		}
		m_stop = System.nanoTime();

		releaseStreamPort();

		CaptureDevice deviceID = m_capture.getCaptureDevice();
		//capture.getCaptureDevice().setInUse(false, 0, null);
		m_capture = null;

		NumberFormat nf = NumberFormat.getInstance();

		// if the item is set to restart just mark it as waiting and return. 
		if(m_item.getState() == ScheduleItem.RESTART)
		{
			m_item.setState(ScheduleItem.WAITING);
			m_item.setStatus("Waiting");
			m_item.log("Schedule Item Restarted and Set to WAITING");

			// Capture File Sizes
			boolean bAllZero = true;
			for(int y = 0; y < m_captureFile.getFilenameCount(); y++)
			{
				String fileName = m_captureFile.getFullPath(y);
				if (fileName == null || fileName.isEmpty())
					continue;

				File capFile = new File(fileName);

				if (capFile.length() > 0)
					bAllZero = false;

				if (capFile.exists())
					m_item.log("Capture : " + capFile.getName() + " (" + nf.format(capFile.length()/1024) + " KB)");
				else
					m_item.log("Capture : " + capFile.getName() + " (doesn't exist)");

				String fullPath = capFile.getParent();
				if (fullPath == null || fullPath.isEmpty())
					continue;
				long freeSpace = getMax(capFile.getFreeSpace(), new File(fullPath).getFreeSpace());
				m_item.log("Free Space on \"" + fullPath + "\" : (" + nf.format(freeSpace/1024) + " KB)");
			}

			if (!m_item.isLive() && !bAllZero && m_store.getProperty("Capture.details", "none").endsWith("end"))
				new CaptureDetails().writeCaptureDetails(this, deviceID);

			deleteCapture(0);

			if (m_item.isAutoDeletable())
			{
				for(int y = 0; y < m_captureFile.getFilenameCount(); y++)
				{
					m_store.addAutoDeleteItem(m_captureFile.getFullPath(y), m_item.getKeepFor() * 24 * 60);
				}
				m_store.saveAutoDelList();
			}

			/*
			///////////////////////////////////////////////////////////////////////////
			// run post task if needed
			HashMap<?, ?> tasks = m_store.getTaskList();

			TaskCommand taskCommand = (TaskCommand) tasks.get(m_item.getPostTask());
			if(taskCommand != null)
			{
				//File capFile = new File(captureFiles.get(0).getFullPath());
				runTask(taskCommand, false, deviceID, m_captureFile);
			}
			 */

			m_store.m_captureTasks.remove(m_item.toString());
			m_item.notifyResponse();
			m_store.saveSchedules(m_item);
			return;
		}

		boolean bDeleteFiles = m_item.getState() == ScheduleItem.ABORTED_DELETEFILES ? true : false;
		m_item.setState(ScheduleItem.FINISHED);
		m_item.setStatus("Finished");

		// Capture File Sizes
		boolean bAllZero = true;
		for(int y = 0; y < m_captureFile.getFilenameCount(); y++)
		{
			String fileName = m_captureFile.getFullPath(y);
			if (fileName == null || fileName.isEmpty())
				continue;

			File capFile = new File(fileName);

			if (capFile.length() > 0)
				bAllZero = false;

			if (capFile.exists())
				m_item.log("Capture : " + capFile.getName() + " (" + nf.format(capFile.length()/1024) + " KB)");
			else
				m_item.log("Capture : " + capFile.getName() + " (doesn't exist)");

			String fullPath = capFile.getParent();
			if (fullPath == null || fullPath.isEmpty())
				continue;
			long freeSpace = getMax(capFile.getFreeSpace(), new File(fullPath).getFreeSpace());
			m_item.log("Free Space on \"" + fullPath + "\" : (" + nf.format(freeSpace/1024) + " KB)");
		}

		if (!bDeleteFiles && !m_item.isLive() && !bAllZero && m_store.getProperty("Capture.details", "none").endsWith("end"))
			new CaptureDetails().writeCaptureDetails(this, deviceID);

		// if the capture files size is zero nuke it.
		if (bDeleteFiles)
		{
			if (!deleteCapture(-2))
			{
				m_item.log("Capture Finished, Recording files have been put in Auto-Deletable File List and will be deleted");
				for(int y = 0; y < m_captureFile.getFilenameCount(); y++)
				{
					m_store.addAutoDeleteItem(m_captureFile.getFullPath(y), -1);
				}
				m_store.saveAutoDelList();
			}

			m_store.m_captureTasks.remove(m_item.toString());
			m_item.notifyResponse();
			m_store.saveSchedules(m_item);
			return;
		}
		else if (m_item.isLive())
		{
			if (m_item.getKeepFor() > 0)
			{
				for(int y = 0; y < m_captureFile.getFilenameCount(); y++)
				{
					m_store.addAutoDeleteItem(m_captureFile.getFullPath(y), m_item.getKeepFor() * 24 * 60);
				}
				m_store.saveAutoDelList();
			}
			else
			{
				int keepForMinutes = 0;
				try
				{
					keepForMinutes = Integer.parseInt(m_store.getProperty("timeshift.keepforMinutes", "0"));
				}
				catch(Exception e){}

				if ( (keepForMinutes == 0) && (!deleteCapture(-1)) )
				{
					for(int y = 0; y < m_captureFile.getFilenameCount(); y++)
					{
						m_store.addAutoDeleteItem(m_captureFile.getFullPath(y), -1);
					}
					m_store.saveAutoDelList();
				}
				else if (keepForMinutes > 0)
				{
					for(int y = 0; y < m_captureFile.getFilenameCount(); y++)
					{
						m_store.addAutoDeleteItem(m_captureFile.getFullPath(y), keepForMinutes);
					}
					m_store.saveAutoDelList();
				}
			}

			m_store.removeScheduleItem(m_item.toString());
			m_store.saveSchedules(m_item);

			m_store.m_captureTasks.remove(m_item.toString());
			m_item.notifyResponse();
			return;
		}
		else
			deleteCapture(0);

		///////////////////////////////////////////////////////////////////////////
		// add auto delete actions
		if (m_item.isAutoDeletable())
		{
			for(int y = 0; y < m_captureFile.getFilenameCount(); y++)
			{
				m_store.addAutoDeleteItem(m_captureFile.getFullPath(y), m_item.getKeepFor() * 24 * 60);
			}
			m_store.saveAutoDelList();
		}

		m_store.saveSchedules(m_item);

		sendEmail("email.captureFinished", "DV Scheduler finished capture");

		///////////////////////////////////////////////////////////////////////////
		// run post task if needed
		HashMap<?, ?> tasks = m_store.getTaskList();

		TaskCommand taskCommand = (TaskCommand) tasks.get(m_item.getPostTask());
		if(taskCommand != null)
		{
			//File capFile = new File(captureFiles.get(0).getFullPath());
			runTask(taskCommand, false, deviceID, m_captureFile);
		}

		m_item.log("Capture Finished");
		m_store.m_captureTasks.remove(m_item.toString());
		m_item.notifyResponse();
		m_store.saveSchedules(m_item);
	}

	private void runPresetTask(CaptureDevice deviceID, String presetName)
	{
		String taskName = m_store.getProperty(presetName, "");
		HashMap<String, TaskCommand> tasks = m_store.getTaskList();
		TaskCommand taskCommand = (TaskCommand) tasks.get(taskName);

		if(taskCommand == null)
		{
			return;
		}

		boolean bRunTask = true;
		boolean bWhenNotCapturing = taskCommand.isWhenNotCapturing();
		int minIdleTime = taskCommand.getMinIdleTime();
		int maxConcurrentTask = taskCommand.getMaxConcurrentTask();
		CaptureDeviceList devList = CaptureDeviceList.getInstance();
		if ( (m_store.m_bSoonCapturing || !m_store.m_captureTasks.isEmpty() || (devList.getRunningDeviceCount() > 0)) && (bWhenNotCapturing || (minIdleTime > 0)) )
			bRunTask = false;

		if ( bRunTask && (maxConcurrentTask > 0) && (m_store.m_runningTaskList.size() >= maxConcurrentTask) )
			bRunTask = false;

		if ( bRunTask && (minIdleTime > 0) && ((m_store.m_nextScheduleTime - Calendar.getInstance().getTimeInMillis()) <= (minIdleTime*60*1000)) )
			bRunTask = false;

		if (bRunTask)
		{
			//runTask(taskCommand, true, deviceID, null);
			runTask(taskCommand, true, deviceID, m_captureFile);
		}
		else
		{
			m_item.log("Task (" + taskCommand.getName() + ") cannot be run due to condition, skipping.");
		}
	}

	private void runTask(TaskCommand taskCommand, boolean waitForExit, CaptureDevice capDev, CaptureFile capFile)
	{  
		String fileName = "";
		if(capFile != null)
			fileName = capFile.getName();

		m_item.log("Running task (" + taskCommand.getName() + ") on file (" + fileName + ")");

		if(taskCommand.getEnabled() == false)
		{
			m_item.log("Task (" + taskCommand.getName() + ") is disabled.");
			return;
		}

		String command = taskCommand.getCommand();
		boolean autoRem = taskCommand.getAutoRemove();
		int delay = taskCommand.getDelay();

		if(command != null && command.length() > 1)
		{
			command = captureEngine.replaceKeywords(command);
			command = capFile.replaceKeywords(command);
			command = m_store.replaceKeywords(command);
			command = capDev.replaceKeywords(command);
			command = m_item.replaceKeywords(command);	// Has to be before Channel::replaceKeywords(...)

			Channel ch = m_store.getChannel(m_item.getRunningChannel());
			if (ch != null)
				command = ch.replaceKeywords(command);

			if ( taskCommand.isConditional() && (waitForExit == false) )
			{
				m_item.log("Queuing task (" + taskCommand.getName() + ")");
				TaskItem taskItem = new TaskItem(taskCommand, command, m_item);
				ThreadLock.getInstance().getLock();
				try
				{
					m_store.m_waitingTaskList.add(taskItem);
				}
				finally
				{
					ThreadLock.getInstance().releaseLock();
				}
			}
			else
			{
				try
				{
					runCommand(taskCommand, command, m_item, waitForExit, autoRem, delay);
				}
				catch (Exception e)
				{
					m_item.log("There was an Exception thrown when trying to run the task.");
					e.printStackTrace();
				}
			}
		}
	}

	private int runCommand(TaskCommand taskCommand, String command, ScheduleItem item, boolean wait,
			boolean autoRem, int delay) throws Exception
			{
		item.log("Command string: " + command);

		long timeout = 20;

		if(delay > 0)
		{
			item.log("The command has a delayed start of " + delay + " seconds");
		}

		CommandWaitThread comThread = new CommandWaitThread(command, delay);
		Thread comWait = new Thread(Thread.currentThread().getThreadGroup(), comThread, comThread.getClass().getName());
		comWait.start();

		if(wait == false)
		{
			ThreadLock.getInstance().getLock();
			try
			{
				String key = new Long(new Date().getTime()).toString() + "-" + new Long(m_store.m_rand.nextLong()).toString();
				TaskItemThread taskItem = new TaskItemThread(key, comThread, command, autoRem);
				taskItem.m_allowSuspend = taskCommand.isAllowSuspend();
				Thread taskThread = new Thread(Thread.currentThread().getThreadGroup(), taskItem, taskItem.getClass().getName());
				taskThread.start();
				m_store.m_runningTaskList.put(key, taskItem);
			}
			finally
			{
				ThreadLock.getInstance().releaseLock();
			}
			return 1;
		}

		item.log("Will wait for " + timeout + " seconds for it to complete");

		long started = System.nanoTime();
		while (true)
		{
			if(comThread.isFinished())
			{
				item.log("Command finished normally");
				break;
			}
			else if((System.nanoTime() - started) > (1000000000L * timeout))
			{
				item.log("Command timed out, killing it now");
				comThread.stop();
				break;
			}
			else if(item.isAborted())
			{
				item.log("Action Aborted so Killing off Command");
				comThread.stop();
				break;
			}
			DVScheduler.log("Waiting for command to finish.....");
			Thread.sleep(1000);
		}

		item.log("Task finished, Output of task follows:");
		item.log("*****************************");

		item.log("Standard Out:\n" + comThread.getInputStream());
		item.log("Standard Error:\n" + comThread.getErrorStream());

		item.log("*****************************");

		item.log("Task Command Finished");

		return 0;
			}

	private boolean dirExists(String dirString)
	{
		File dir = new File(dirString);

		if(dir.exists() && dir.isDirectory())
			return true;
		else
			return false;
	}

	protected String getValue(String text)
	{
		return text.substring(text.indexOf("->") + 2);
	}

	protected CaptureFile getFullFileName()
	{
		// Check capture type.
		String capType = m_item.getCaptureCommandFile();
		String commandTemplate = "";
		Vector<String> fullPathTemplates = new Vector<String>();
		int filenameEncoding = Formatter.TYPE_ASCII;
		String changeChannelTemplate = "";
		String changeSettingTemplate = "";
		String clientGetAddressTemplate = "";
		boolean bSeekable = false;

		String capTypeFile = m_store.getProperty("Capture.engine.path", "util") + File.separator + capType;
		File dir = new File(capTypeFile);

		if (!dir.exists() || !dir.isFile())
		{
			m_item.incrementErrorCount();
			m_item.setState(ScheduleItem.ERROR);
			m_item.setStatus("Error capture type");
			m_item.log("Capture type file (" + capType + ") doesn't exist");
			m_store.m_nRecordingErrorCount++;
			return null;
		}
		else if (dir.length() > 4096)
		{
			m_item.incrementErrorCount();
			m_item.setState(ScheduleItem.ERROR);
			m_item.setStatus("Error capture type");
			m_item.log("Capture type file (" + capType + ") > 4096 bytes - too big, check for error");
			m_store.m_nRecordingErrorCount++;
			return null;
		}
		else
		{
			m_idleTimeout = 0;
			m_keepAliveTime = 0;

			try
			{
				BufferedReader r = new BufferedReader(new FileReader(capTypeFile));
				String line = null;

				while ((line = r.readLine()) != null)
				{
					line = line.trim();

					if (line.startsWith("Capture Engine->"))
					{
						if ( !captureEngine.SetEngine(line.substring(16).trim(), m_store.getProperty("Capture.engine.path", "util") + File.separator + line.substring(16).trim() + ".ce") )
						{
							r.close();
							m_item.incrementErrorCount();
							m_item.setState(ScheduleItem.ERROR);
							m_item.setStatus("Error capture type");
							m_item.log("Error reading capture engine file " + line.substring(16).trim() + ".ce");
							m_store.m_nRecordingErrorCount++;

							return null;
						}
					}
					else if (line.startsWith("Filename Encoding->"))
					{
						String enc = line.substring(19).trim();

						filenameEncoding = Formatter.getEncoding(enc, Formatter.TYPE_ASCII);
					}
					else if (line.startsWith("Filename->"))
					{
						fullPathTemplates.add(line.substring(10).trim());
					}
					else if (line.startsWith("||"))
					{
						fullPathTemplates.add(line.substring(2).trim());
					}
					else if (line.startsWith("Client Get Address->"))
					{
						clientGetAddressTemplate = line.substring(20).trim();
					}
					else if (line.startsWith("Is Seekable->"))
					{
						bSeekable = line.substring(13).trim().equals("1") ? true : false;
					}
					else if (line.startsWith("Idle Timeout->"))
					{
						m_idleTimeout = Long.parseLong(getValue(line)) * 1000000000;
					}
					else if (line.startsWith("Keep Alive Time->"))
					{
						m_keepAliveTime = Long.parseLong(getValue(line)) * 1000000000;
					}
					else if ( line.startsWith("Local Change Channel Command->") && 
							HTTPResponse.isLocal(m_item.getClientName()) )
					{
						changeChannelTemplate += line.substring(30).trim() + "\n";
					}
					else if ( line.startsWith("Remote Change Channel Command->") && 
							!HTTPResponse.isLocal(m_item.getClientName()) )
					{
						changeChannelTemplate += line.substring(31).trim() + "\n";
					}
					else if (line.startsWith("Change Channel Command->"))
					{
						changeChannelTemplate += line.substring(24).trim() + "\n";
					}
					else if ( line.startsWith("Local Change Setting Command->") && 
							HTTPResponse.isLocal(m_item.getClientName()) )
					{
						changeSettingTemplate += line.substring(30).trim() + "\n";
					}
					else if ( line.startsWith("Remote Change Setting Command->") && 
							!HTTPResponse.isLocal(m_item.getClientName()) )
					{
						changeSettingTemplate += line.substring(31).trim() + "\n";
					}
					else if (line.startsWith("Change Setting Command->"))
					{
						changeSettingTemplate += line.substring(24).trim() + "\n";
					}
					else if ( line.startsWith("Local Command->") && 
							HTTPResponse.isLocal(m_item.getClientName()) )
					{
						commandTemplate += line.substring(15).trim() + "\n";
					}
					else if ( line.startsWith("Remote Command->") && 
							!HTTPResponse.isLocal(m_item.getClientName()) )
					{
						commandTemplate += line.substring(16).trim() + "\n";
					}
					else if (line.startsWith("Create File Name->"))
					{
						m_createFileName = getValue(line).trim();
					}
					else if (line.startsWith("Create File Type->"))
					{
						m_createFileType = getValue(line).trim();
					}
					else if (line.startsWith("Create File Content->"))
					{
						m_createFileContent += getValue(line) + "\r\n";
					}
					else if (line.startsWith("Redirect->"))
					{
						m_redirect += getValue(line) + "\n";
					}
					else if (!line.startsWith("//") && !line.isEmpty())
						commandTemplate += line + "\n";
				}

				r.close();

				if ( (m_createFileName.isEmpty() != m_createFileType.isEmpty()) ? true : (m_createFileName.isEmpty() != m_createFileContent.isEmpty()) ) 
				{
					m_item.incrementErrorCount();
					m_item.setState(ScheduleItem.ERROR);
					m_item.setStatus("Error capture type");
					m_item.log("Error reading capture type file (" + capType + "): One or more of CreateFileName, CreateFileType, CreateFileContent missing.");
					m_store.m_nRecordingErrorCount++;
					return null;
				}
			}
			catch (Exception e)
			{
				m_item.incrementErrorCount();
				m_item.setState(ScheduleItem.ERROR);
				m_item.setStatus("Error capture type");
				m_item.log("Error reading capture type file (" + capType + ")");
				m_store.m_nRecordingErrorCount++;
				e.printStackTrace();
				return null;
			}
		}

		String bestFullPath = null;
		String bestFullPathOriginal = null;
		if (fullPathTemplates.isEmpty())
		{
			//bestFullPath = DVScheduler.getArray(m_store.getProperty("Capture.path", "capture"))[0];
			bestFullPath = "";
			bestFullPathOriginal = "";
		}
		else
		{
			long minFreeSpace = 1;
			try
			{
				minFreeSpace = Long.parseLong(m_store.getProperty("Capture.minFreeSpace", "20000")) * 1024 * 1024;
			}
			catch (Exception e) {}
			if (minFreeSpace < 1)
				minFreeSpace = 1;

			int nPathSelection = 0;
			String customFullPath = "";
			String customFullPathOriginal = "";
			long customPathFreeSpace = 0;
			File cap = null;
			String itemCapturePath = m_item.getCapturePathActual();
			String[] paths = DVScheduler.getArray(m_store.getProperty("Capture.path", "capture"));
			if (itemCapturePath != null && !itemCapturePath.equals("*"))
			{
				boolean bFound = false;
				for (int x = 0; x < paths.length; x++)
				{
					if (paths[x].equals(itemCapturePath))
					{
						bFound = true;
						break;
					}
				}
				if (!bFound)
				{
					m_item.log("Selected capture path not found in list of capture paths; using default path selection rule");
				}
				else
				{
					if (itemCapturePath.isEmpty())
						itemCapturePath = ".";

					cap = new File(itemCapturePath);
					if (!cap.exists())
						cap.mkdirs();
					if (!cap.exists())
					{
						m_item.log("Capture directory (" + cap.getAbsolutePath() + ") cannot be created");
					}
					else if (cap.getFreeSpace() >= minFreeSpace)
					{
						bestFullPath = cap.getAbsolutePath();
						bestFullPathOriginal = itemCapturePath;
						nPathSelection = 2;
					}
					else
					{
						customPathFreeSpace = cap.getFreeSpace();
						customFullPath = cap.getAbsolutePath();
						customFullPathOriginal = itemCapturePath;
						nPathSelection = 1;
					}
				}
			}

			if (nPathSelection < 2)
			{
				long bestFreeSpace = -1;
				Vector<String> possibleFullPaths = new Vector<String>();
				Vector<String> possibleFullPathsOriginal = new Vector<String>();
				for(int x = 0; x < paths.length; x++)
				{
					if (paths[x].isEmpty())
						paths[x] = ".";

					cap = new File(paths[x]);
					if (!cap.exists())
						cap.mkdirs();
					if (!cap.exists())
					{
						//					item.setState(ScheduleItem.WAITING);
						//					item.setStatus("Error capture directory");
						m_item.log("Capture directory (" + cap.getAbsolutePath() + ") cannot be created");
						//					return null;
					}
					else
					{
						String fullPath = cap.getAbsolutePath();
						long freeSpace = cap.getFreeSpace();
						if (freeSpace >= minFreeSpace)
						{
							possibleFullPaths.add(fullPath);
							possibleFullPathsOriginal.add(paths[x]);
						}
						if (freeSpace > bestFreeSpace)
						{
							bestFreeSpace = freeSpace;
							bestFullPath = fullPath;
							bestFullPathOriginal = paths[x];
						}
					}
				}

				if ( (nPathSelection == 1) && ((bestFullPath == null) || ((customPathFreeSpace > (bestFreeSpace - 1000000) && (customPathFreeSpace > 0)))) )
				{
					bestFullPath = customFullPath;
					bestFullPathOriginal = customFullPathOriginal;
				}
				else
				{
					if (bestFullPath == null)
					{
						m_item.incrementErrorCount();
						m_item.setState(ScheduleItem.WAITING);
						m_item.setStatus("No free space");
						m_item.log("No free space, will retry");
						m_store.m_nRecordingErrorCount++;
						return null;
					}

					if (nPathSelection == 1)
					{
						m_item.log("Selected capture path has less than minimum free space; using default path selection rule");
					}

					if (possibleFullPaths.size() > 1)
					{
						String pathSelection = m_store.getProperty("Capture.pathSelection", "Max Space");
						if (pathSelection.equals("Sequential"))
						{
							bestFullPath = possibleFullPaths.firstElement();
							bestFullPathOriginal = possibleFullPathsOriginal.firstElement();
						}
						else if (pathSelection.equals("Random"))
						{
							int randValue = m_store.m_rand.nextInt(possibleFullPaths.size());
							bestFullPath = possibleFullPaths.get(randValue);
							bestFullPathOriginal = possibleFullPathsOriginal.get(randValue);
						}
						else if (pathSelection.equals("Evenly Distributed"))
						{
							int[] count = new int[possibleFullPaths.size()];
							for (int i=0; i<count.length; i++)
								count[i] = 0;

							Iterator<String> iter = m_store.m_captureTasks.keySet().iterator();

							while (iter.hasNext())
							{
								CaptureTask task = m_store.m_captureTasks.get(iter.next());
								if ( (task == this) || (task.getCaptureFile() == null) )
									continue;

								int index = possibleFullPaths.indexOf(task.getCaptureFile().getUnformattedDirectory());

								if (index > -1)
									count[index]++;
							}

							int indexMin = -1;
							int countMin = Integer.MAX_VALUE;
							for (int i=0; i<count.length; i++)
							{
								if (count[i] < countMin)
								{
									indexMin = i;
									countMin = count[i];
								}
							}

							bestFullPath = possibleFullPaths.get(indexMin);
							bestFullPathOriginal = possibleFullPathsOriginal.get(indexMin);
						}
					}
				}
			}
		}

		if (commandTemplate.contains("$streamport"))
			m_streamPort = 0;

		CaptureFile cap = m_item.getFileName();
		cap.setEncoding(filenameEncoding);
		cap.setDirectory(bestFullPath);
		cap.setOriginalDirectory(bestFullPathOriginal);

		cap.setOriginalName(cap.getUnformattedName());
		String tempName = getUniqueName(cap.getName(), cap.getDirectory());
		if (tempName == null)
			return null;
		cap.setName( tempName );
		cap.setCommandTemplate(commandTemplate);
		cap.setFullPathTemplates(fullPathTemplates);
		cap.setChangeChannelTemplate(changeChannelTemplate);
		cap.setChangeSettingTemplate(changeSettingTemplate);
		cap.setClientGetAddressTemplate(clientGetAddressTemplate);
		cap.setSeekable(bSeekable);

		return cap;
	}

	protected String getUniqueName(String fileName, String directory)
	{
		if ( (directory == null) || directory.isEmpty() )
			return fileName;

		int count = 1;
		File dir = new File(directory + File.separator + fileName);
		String fileName2 = dir.getName();
		String[] contents = dir.getParentFile().list(new FilenameExistFilter(fileName2));

		String newFileName = fileName2;
		while ( (contents != null) && (contents.length > 0) )
		{
			newFileName = fileName2 + "-" + ++count;
			contents = dir.getParentFile().list(new FilenameExistFilter(newFileName));
		}

		File parentPath = dir.getParentFile();
		if(!parentPath.exists())
		{
			parentPath.mkdirs();
		}

		if(!parentPath.exists())
		{
			m_item.incrementErrorCount();
			m_item.setState(ScheduleItem.WAITING);
			m_item.setStatus("Error filename directory");
			m_item.log("Filename directory (" + parentPath.getAbsolutePath() + ") cannot be created, will retry");
			m_store.m_nRecordingErrorCount++;
			return null;
		}

		return count == 1 ? fileName : fileName + "-" + count;
	}

	protected boolean deleteCapture(long minSize)
	{
		boolean bFail = false;
		for(int y = 0; y < m_captureFile.getFilenameCount(); y++)
		{
			File file = new File(m_captureFile.getFullPath(y));

			if (!file.exists())
				continue;

			if (minSize == -1)
			{
				if ( !file.delete() )
					bFail = true;
			}
			else if ( (file.length() <= minSize) || (minSize == -2) )
			{
				if ( !file.delete() )
				{
					bFail = true;
					m_item.log("Cannot delete capture file : " + file.getName());
				}
				else
				{
					m_item.log("Deleted capture file : " + file.getName());
				}
			}
		}

		return !bFail;
	}

	public CaptureFile getCaptureFile()
	{
		return m_captureFile;
	}

	public int getCardID()
	{
		return m_capture.getCaptureDevice().getNumber();
	}

	public String getCardName()
	{
		return m_capture.getCaptureDevice().getName();
	}

	protected int setName()
	{
		if (m_item.isLive())
		{
			if (m_item.getKeepFor() > 0)
			{
				for(int y = 0; y < m_captureFile.getFilenameCount(); y++)
				{
					m_store.addAutoDeleteItem(m_captureFile.getFullPath(y), m_item.getKeepFor() * 24 * 60);
				}
				m_store.saveAutoDelList();

				String newFileName = getUniqueName(m_captureFile.getOriginalName(), m_captureFile.getDirectory());
				if (newFileName == null)
					return -1;
				m_captureFile.setName(newFileName);
			}
			else
			{
				int keepForMinutes = 0;
				try
				{
					keepForMinutes = Integer.parseInt(m_store.getProperty("timeshift.keepforMinutes", "0"));
				}
				catch(Exception e){}

				if ( (keepForMinutes == 0) && (!deleteCapture(-1)) )
				{
					for(int y = 0; y < m_captureFile.getFilenameCount(); y++)
					{
						m_store.addAutoDeleteItem(m_captureFile.getFullPath(y), -1);
					}
					m_store.saveAutoDelList();

					String newFileName = getUniqueName(m_captureFile.getOriginalName(), m_captureFile.getDirectory());
					if (newFileName == null)
						return -1;
					m_captureFile.setName(newFileName);
				}
				else if (keepForMinutes > 0)
				{
					for(int y = 0; y < m_captureFile.getFilenameCount(); y++)
					{
						m_store.addAutoDeleteItem(m_captureFile.getFullPath(y), keepForMinutes);
					}
					m_store.saveAutoDelList();

					String newFileName = getUniqueName(m_captureFile.getOriginalName(), m_captureFile.getDirectory());
					if (newFileName == null)
						return -1;
					m_captureFile.setName(newFileName);
				}
				else if (keepForMinutes < 0)
				{
					String newFileName = getUniqueName(m_captureFile.getOriginalName(), m_captureFile.getDirectory());
					if (newFileName == null)
						return -1;
					m_captureFile.setName(newFileName);
				}
			}
		}
		else
		{
			String newFileName = getUniqueName(m_captureFile.getOriginalName(), m_captureFile.getDirectory());
			if (newFileName == null)
				return -1;
			m_captureFile.setName(newFileName);
		}

		return 0;
	}

	public int changeChannel(Channel newChannel)
	{
		if (setName() < 0)
			return -1;

		m_item.setRunningChannel(newChannel.getName());
		int startCode = m_capture.changeChannel(m_item, m_captureFile, newChannel);

		if (startCode == 0)
		{
			m_item.log("Change channel succeeded with code : " + startCode);
			return 1;
		}
		else
		{
			m_item.log("Change channel failed with code : " + startCode);
			return startCode;
		}
	}

	public int changeSetting(Channel newChannel)
	{
		if (setName() < 0)
			return -1;

		m_item.setRunningChannel(newChannel.getName());
		int startCode = m_capture.changeSetting(m_item, m_captureFile, newChannel);

		if (startCode == 0)
		{
			m_item.log("Change setting succeeded with code : " + startCode);
			return 1;
		}
		else
		{
			m_item.log("Change setting failed with code : " + startCode);
			return startCode;
		}
	}

	public int changeDevice(Channel newChannel)
	{
		if (setName() < 0)
			return -1;

		m_item.setRunningChannel(newChannel.getName());
		int result = m_capture.stopCapture(m_item.getID());
		if ( result != 0 )
			return result;

		return startCapture(newChannel);
	}

	public String getClientGetAddress()
	{
		Channel ch = m_store.getChannel(m_item.getRunningChannel());
		return m_capture.getClientGetAddress(m_item, ch, m_captureFile);
	}

	public boolean isSeekable()
	{
		return m_captureFile.isSeekable();
	}

	public boolean isFile()
	{
		return m_captureFile.getFilenameCount() > 0;
	}

	public void setLastTime()
	{
		m_lastResponseTime = System.nanoTime();
	}

	public long getKeepAliveTime()
	{
		return m_keepAliveTime / 1000000000;
	}

	public boolean isSettingChangeable()
	{
		return captureEngine.isCanRecordMultiple();
	}

	public void setLiveToNormal(boolean liveToNormal)
	{
		m_bLiveToNormal = liveToNormal;
	}

	public boolean isLiveToNormal()
	{
		return m_bLiveToNormal;
	}

	public String getSignalStatistic()
	{
		return m_capture.getSignalStatistic();
	}

	private void releaseStreamPort()
	{
		if (m_streamPort > 0)
			m_store.returnStreamPort(m_streamPort);
		m_streamPort = -1;
	}

	private String replaceKeywords(String text)
	{
		text = captureEngine.replaceKeywords(text);	// Has to be first.
		text = m_captureFile.replaceKeywords(text);
		text = m_store.replaceKeywords(text);
		text = m_capture.getCaptureDevice().replaceKeywords(text);
		text = m_item.replaceKeywords(text);	// Has to be before Channel::replaceKeywords(...)

		Channel ch = m_store.getChannel(m_item.getRunningChannel());
		if (ch != null)
			text = ch.replaceKeywords(text);

		return text;
	}

	public String getCreateFileName()
	{
		return replaceKeywords(m_createFileName);
	}

	public String getCreateFileType()
	{
		return replaceKeywords(m_createFileType);
	}

	public String getRedirect()
	{
		return replaceKeywords(m_redirect);
	}

	public String getCreateFileContent()
	{
		return replaceKeywords(m_createFileContent);
	}

	private void sendEmail(String property, String subject)
	{
		if ( !m_store.getProperty(property, "No").equals("Yes") )
			return;

		DVScheduler.log("Sending Email: " + subject);

		EmailSender sender = new EmailSender();

		//sender.setSubject(subject);
		sender.setSubject("Recording");
		sender.setBody(m_item.getLog());

		Thread mailThread = new Thread(Thread.currentThread().getThreadGroup(), sender, sender.getClass().getName());
		mailThread.start();
	}

	// Return value in millisecond.
	public long getCaptureTime()
	{
		if ( (m_start < 0) || (m_stop < 0) )
		{
			DVScheduler.log("Error in CaptureTask.getCaptureTime()");
			return 0;
		}

		return (m_stop - m_start) / 1000000;
	}
}
