/*
   Copyright (C) 2007-2010 contributors

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

 */


import java.io.*;
import java.net.*;
import java.nio.charset.Charset;
import java.text.DateFormat;
import java.text.SimpleDateFormat;
import java.util.*;

public class DVScheduler
{
	private static int port = 8007;

	static boolean m_bLogToFile = false;
	static String m_portNum = "";
	static Date m_start = null;
	static long m_startNano = -1;

	public static void main(String[] args)
	{
		m_startNano = System.nanoTime();
		m_start = new Date();
		Runtime.getRuntime().addShutdownHook(new ShutDownHook());

		try
		{
			System.setOut(new PrintStream(System.out, true, "UTF-8"));
			parseArgs(args);

			// Since version 5.63
			BufferedReader r = new BufferedReader(new FileReader("paths.list"));
			String pathUserData = r.readLine().trim();
			String saveToAppData = r.readLine().trim();
			r.close();

			DataStore.m_saveFile = saveToAppData.equals("1") ? pathUserData + File.separator + "server.prop" : "server.prop";
			DataStore.m_saveFolder = saveToAppData.equals("1") ? pathUserData + File.separator : "";
			DataStore.m_saveFolderData = saveToAppData.equals("1") ? pathUserData + File.separator : "data" + File.separator;

			// if started with -log
			// redirect the output to the log file
			if(m_bLogToFile)
			{
				File outStream = getLogFile();
				//PrintStream log = new PrintStream(new FileOutputStream(outStream, true));
				PrintStream log = new PrintStream(new FileOutputStream(outStream, true), true, "UTF-8");

				System.out.println("******************************************************");
				System.out.println("DVScheduler Starting : " + new Date().toString() );
				System.out.println("******************************************************");
				System.out.println("For errors see the following log files:\n");
				System.out.println(outStream.getAbsoluteFile());
				//System.out.println(new File(".").getAbsoluteFile().getParent() + File.separator + "DirectShow.log");

				log.println("");
				log.println("******************************************************");
				log.println("DVScheduler Starting : " + new Date().toString() );
				log.println("******************************************************");

				System.setOut(log);
				System.setErr(log);
			}

			printSystemProp();

			deleteOldLogs();

			DataStore store = DataStore.getInstance();
			DVScheduler.log(store.toString());
			CaptureDeviceList devList = CaptureDeviceList.getInstance();
			DVScheduler.log(devList.toString());
			GuideStore guide = GuideStore.getInstance();
			DVScheduler.log(guide.toString());
			store.setProperty("path.userData", pathUserData);
			store.setProperty("saveTo.appData", saveToAppData);

			if(m_portNum.length() > 0)
				port = Integer.parseInt(m_portNum);

			try
			{
				checkOldVersions();
			}
			catch(Exception e)
			{
				DVScheduler.log("check old versions : " + e.toString());
			}

			TimerThread timer = new TimerThread();
			store.m_timerThread = timer;
			Thread timerThread = new Thread(Thread.currentThread().getThreadGroup(), timer, timer.getClass().getName());
			timerThread.start();

			boolean bHttps = (store.getWebGuiProtocol() == DataStore.WEBGUI_HTTPS);
			int httpsPort = store.getHttpsPort();
			HTTPSThread https = null;
			Thread httpsThread = null;
			if (bHttps)
			{
				https = new HTTPSThread(httpsPort, store.getKeyStore(), store.getKeyStorePassword(), store.getKeyPassword());
				httpsThread = new Thread(Thread.currentThread().getThreadGroup(), https, https.getClass().getName());
				httpsThread.start();
			}

			ServerSocket listener = new ServerSocket(port);

			Socket server;

			DVScheduler.log("Starting server listening on (" + listener.getInetAddress().toString() + ":" + port + ")");

			if (bHttps)
			{
				while(true)
				{
					try
					{
						server = listener.accept();
						RequestRedir con = new RequestRedir(server, httpsPort);
						Thread conThread = new Thread(Thread.currentThread().getThreadGroup(), con, con.getClass().getName());
						conThread.start();
					}
					catch (Exception e1)
					{
						e1.printStackTrace();
					}
				}
			}
			else
			{
				while(true)
				{
					try
					{
						server = listener.accept();
						RequestObject con = new RequestObject(server);
						Thread conThread = new Thread(Thread.currentThread().getThreadGroup(), con, con.getClass().getName());
						conThread.start();
					}
					catch (Exception e1)
					{
						e1.printStackTrace();
					}
				}
			}
		}
		catch (Exception e)
		{
			DVScheduler.log("Base server Exception: " + e);
			e.printStackTrace();
			System.exit(0);
		}
	}

	private static void parseArgs(String[] args)
	{
		for(int x = 0; x < args.length; x++)
		{
			if(args[x].equals("-log"))
			{
				m_bLogToFile = true;
			}
			else if(x < args.length-1 && args[x].equals("-port"))
			{
				String tempPort = args[x+1];
				if(tempPort.startsWith("-") == false)
				{
					m_portNum = tempPort;
					x++;
				}
			}         
		}
	}

	private static File getLogFile()
	{
		String fileName = DataStore.m_saveFolder + "log" + File.separator + "Scheduler_";

		Calendar now = Calendar.getInstance();
		SimpleDateFormat fm = new SimpleDateFormat("yyyy-MM-dd_(HH.mm.ss.SSS)");
		fileName += fm.format(now.getTime()) + ".log";

		File log = new File(fileName);
		File parent = log.getParentFile();
		if(parent.exists() == false)
		{
			parent.mkdirs();
		}

		return log;
	}

	private static void deleteOldLogs()
	{
		File logDir = new File("log");

		if(logDir.exists() == false)
		{
			return;
		}

		File[] logFiles = logDir.listFiles();
		long ageLimit = (new Date().getTime()) - (1000 * 60 * 60 * 24 * 21); //3 weeks

		for(int x = 0; x < logFiles.length; x++)
		{
			String name = logFiles[x].getName();
			if(name.length() > 4 && name.substring(name.length() - 4, name.length()).equalsIgnoreCase(".log"))
			{
				if(logFiles[x].lastModified() < ageLimit)
				{
					try
					{
						DVScheduler.log("Deleting old log file : " + name);
						logFiles[x].delete();
					}
					catch(Exception e)
					{
						DVScheduler.log("Could not delete old log file : " + name);
					}
				}
			}  
		}
	}

	public static void printSystemProp()
	{ 
		System.out.println("System Info");
		System.out.println("--------------------------------------------------------------------------");
		System.out.println("Current Time: " + new Date().toString());
		System.out.println("java.home: " + System.getProperty("java.home"));
		System.out.println("java.class.path: " + "java.home" + System.getProperty("java.class.path"));
		System.out.println("java.specification.version: " + System.getProperty("java.specification.version"));
		System.out.println("java.specification.vendor: " + System.getProperty("java.specification.vendor"));
		System.out.println("java.specification.name: " + System.getProperty("java.specification.name"));
		System.out.println("java.version: " + System.getProperty("java.version"));
		System.out.println("java.vendor: " + System.getProperty("java.vendor"));
		System.out.println("java.vendor.url: " + System.getProperty("java.vendor.url"));
		System.out.println("java.vm.specification.version: " + System.getProperty("java.vm.specification.version"));
		System.out.println("java.vm.specification.vendor: " + System.getProperty("java.vm.specification.vendor"));
		System.out.println("java.vm.specification.name: " + System.getProperty("java.vm.specification.name"));
		System.out.println("java.vm.version: " + System.getProperty("java.vm.version"));
		System.out.println("java.vm.vendor: " + System.getProperty("java.vm.vendor"));
		System.out.println("java.vm.name: " + System.getProperty("java.vm.name"));
		System.out.println("java.class.version: " + System.getProperty("java.class.version"));
		System.out.println("os.home: " + System.getProperty("os.name"));
		System.out.println("os.arch: " + System.getProperty("os.arch"));
		System.out.println("os.version: " + System.getProperty("os.version"));
		System.out.println("user.name: " + System.getProperty("user.name"));
		System.out.println("user.home: " + System.getProperty("user.home"));
		System.out.println("user.dir: " + System.getProperty("user.dir"));
		System.out.println("--------------------------------------------------------------------------");
	}   

	public static void log(StringBuffer text)
	{
		System.out.println(DateFormat.getDateTimeInstance().format(new Date()) + ": " + text);
	}

	public static void log(String text)
	{
		System.out.println(DateFormat.getDateTimeInstance().format(new Date()) + ": " + text);
	}

	public static void log(String text, Date date)
	{
		System.out.println(DateFormat.getDateTimeInstance().format(date) + ": " + text);
	}

	public static void logCounter(String text)
	{
		System.out.println(System.nanoTime() + ": " + text);
	}

	private static void checkOldVersions() throws Exception
	{
		DataStore store = DataStore.getInstance();

		// Prior to v5.39 User-agent "Windows CE" is hard-coded to use theme "simple".
		// This has been removed. In place is this user-removable option.
		if ( store.getProperty("path.agentThemes.version", "1").equals("1") )
		{
			if (!store.isThemeExist("Windows CE"))
			{
				store.addTheme(0, "Windows CE", "simple");
			}

			store.setProperty("path.agentThemes.version", "2");
		}

		// From v5.39 the theme "default" is absorbed/moved to root "http" itself.
		if ( store.getProperty("path.theme", "..").equalsIgnoreCase("default") )
		{
			store.setProperty("path.theme", "..");
		}

		// From v5.39 change in theme structure means early status.xsl is no longer compatible, rename so it won't be used.
		renameStatusXsl("default");
		renameStatusXsl("default-Horizontal");
		renameStatusXsl("unknown2");

		// From v5.39 no more default directory. If it exists then probably from earlier version not uninstalled.
		String sDir = store.getProperty("path.httproot", "http") + File.separator + "themes" + File.separator;
		File dir = new File(sDir + "default");
		if (dir.exists() && dir.isDirectory())
		{
			int count = 2;
			sDir += "unknown";
			File newDir;
			while (true)
			{
				newDir = new File(sDir + count++);
				if (!newDir.exists())
				{
					if ( dir.renameTo(newDir) )
						DVScheduler.log("Renamed theme directory \"default\" to \"" + newDir.getName() + "\"");
					else
						DVScheduler.log("Cannot rename disallowed theme directory \"default\" to \"" + newDir.getName() + "\"");
					break;
				}
			}
		}
	}

	private static void renameStatusXsl(String directory)
	{
		DataStore store = DataStore.getInstance();
		Calendar earliestDate = Calendar.getInstance();
		earliestDate.set(2009, 8, 19, 0, 0, 0);	// 2009 September 19
		String sDir = store.getProperty("path.httproot", "http") + File.separator + "themes" + File.separator + directory + File.separator + "xsl" + File.separator;
		File dir = new File(sDir + "status.xsl");
		if (dir.exists() && dir.isFile() && (dir.lastModified() < earliestDate.getTimeInMillis()))
		{
			int count = 2;
			sDir += "status - old version doesnt work in v5.39 or newer (";
			File newDir;
			while (true)
			{
				newDir = new File(sDir + count++ + ").xsl");
				if (!newDir.exists())
				{
					if ( dir.renameTo(newDir) )
						DVScheduler.log("Renamed old \"" + directory + "\\status.xsl\" to \"" + directory + "\\" + newDir.getName() + "\"");
					else
						DVScheduler.log("Cannot rename old \"" + directory + "\\status.xsl\" to \"" + directory + "\\" + newDir.getName() + "\"");
					break;
				}
			}
		}
	}

	public static String checkBreaks(String text)
	{
		return text == null ? null : text.replace("\r\n", "\n");
	}

	public static String[] getArray(String text)
	{
		return text == null ? null : text.split("\n");
	}

	public static int getHttpPort()
	{
		return port;
	}
}
