/*
   Copyright (C) 2007

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

*/

#include "graph.h"
#include "iMpeg2PsiParser.h"
#include "Bdatif.h"
#include "Sbe.h"
#include "MediaFormats.h"
#include <initguid.h>
#include "../../FlowMeter/src/iFlowMeter.h"
#include "../../FileWriter/src/iFileWriter.h"
#include "CStreamScan.h"

#include "Dvdmedia.h"

//
// An application can advertise the existence of its filter graph
// by registering the graph with a global Running Object Table (ROT).
// The GraphEdit application can detect and remotely view the running
// filter graph, allowing you to 'spy' on the graph with GraphEdit.
//
// To enable registration in this sample, define REGISTER_FILTERGRAPH.
//
#define REGISTER_FILTERGRAPH

// We use channel 46 internally for testing.  Change this constant to any value.
#define DEFAULT_PHYSICAL_CHANNEL    46L

int messageText(std::string *buffer, char *sz,...)
{
	if(buffer != NULL)
	{
		char tach[2000];

		va_list va;
		va_start(va, sz);
		StringCchVPrintf(tach, 2000, sz, va);
		va_end(va);

		buffer->append(tach);
	}

    return FALSE;
}

// Constructor, initializes member variables
// and calls InitializeGraphBuilder
CBDAFilterGraph::CBDAFilterGraph() :
    m_fGraphBuilt(FALSE),
    m_fGraphRunning(FALSE),
    m_NetworkType(DVB_T),
    m_dwGraphRegister (0),
	m_lFrequency(191625),
	m_lBandwidth(7)
{
    if(FAILED(InitializeGraphBuilder()))
        m_fGraphFailure = TRUE;
    else
        m_fGraphFailure = FALSE;
}


// Destructor
CBDAFilterGraph::~CBDAFilterGraph()
{
    if(m_fGraphRunning)
    {
        StopGraph();
    }

    if(m_fGraphBuilt || m_fGraphFailure)
    {
        TearDownGraph();
    }
}


// Instantiate graph object for filter graph building
HRESULT
CBDAFilterGraph::InitializeGraphBuilder()
{
    HRESULT hr = S_OK;
    
    // we have a graph already
    if (m_pFilterGraph)
        return S_OK;

    // create the filter graph

	hr = m_pFilterGraph.CoCreateInstance (CLSID_FilterGraph);

    if (FAILED(hr))
    {
        log("win32-DLL-BDA : Couldn't CoCreate IGraphBuilder\n");
        m_fGraphFailure = true;
        return hr;
    }
    
    return hr;
}

int
CBDAFilterGraph::getProgramInfo(std::vector <CProgramInfo> *ppinfo)
{
	Sleep(2000);

	int found = 0;
	for(int x = 0; x < 6; x++)
	{
		scanForProgramInfo(&found, ppinfo);
		if(found > 0)
			break;
		log("No Programs found, sleeping for 2 sec and try again\n");
		Sleep(2000);
	}

	char buff[256];
	StringCchPrintf(buff, 256, "Number of programs found = %d", found); 
	DbgOutString(buff);

	return found;
}

// BuildGraph sets up devices, adds and connects filters
HRESULT
CBDAFilterGraph::BuildGraph(NETWORK_TYPE NetType, int sourceIndex)
{
    HRESULT hr = S_OK;
    m_NetworkType = NetType;

    // if we have already have a filter graph, tear it down
    if(m_fGraphBuilt)
    {
        if(m_fGraphRunning)
        {
            hr = StopGraph ();
        }

        hr = TearDownGraph ();
    }

	log("win32-DLL-BDA : Starting build of capture graph\n");

    // STEP 1: load network provider first so that it can configure other
    // filters, such as configuring the demux to sprout output pins.
    // We also need to submit a tune request to the Network Provider so it will
    // tune to a channel
	log("win32-DLL-BDA : Loading Network Provider\n");
    if(FAILED (hr = LoadNetworkProvider()))
    {
        log("win32-DLL-BDA : Cannot load network provider\n");
        TearDownGraph();
        m_fGraphFailure = true;
        return hr;
    }

    hr = m_pNetworkProvider->QueryInterface(__uuidof (ITuner), reinterpret_cast <void**> (&m_pITuner));
    if(FAILED (hr))
    {
        log("win32-DLL-BDA : pNetworkProvider->QI: Can't QI for ITuner.\n");
        TearDownGraph();
        m_fGraphFailure = true;
        return hr;
    }

	log("win32-DLL-BDA : Creating Tune Request\n");
    // create a tune request to initialize the network provider
    // before connecting other filters
    CComPtr <IDVBTuneRequest>  pDVBTTuneRequest;
    if(FAILED (hr = CreateDVBTTuneRequest(
                                        0, //m_lFrequency,
                                        0, //m_lBandwidth, 
                                        &pDVBTTuneRequest
                                        )))
    {
        log("win32-DLL-BDA : Cannot create tune request\n");
        TearDownGraph();
        m_fGraphFailure = true;
        return hr;
    }

	log("win32-DLL-BDA : Submiting Tune Request\n");
    //submit the tune request to the network provider
    hr = m_pITuner->put_TuneRequest(pDVBTTuneRequest);
    if(FAILED(hr))
    {
        log("win32-DLL-BDA : Cannot submit the tune request\n");
        TearDownGraph();
        m_fGraphFailure = true;
        return hr;
    }

	log("win32-DLL-BDA : Loading Tuner Filter\n");
    // Load tuner device and connect to network provider
    if(FAILED (hr = LoadFilter (
                                KSCATEGORY_BDA_NETWORK_TUNER, 
                                &m_pTunerDevice,
                                m_pNetworkProvider, 
                                TRUE, sourceIndex
                                )))
    {
        log("win32-DLL-BDA : Cannot load tuner device and connect network provider\n");
        TearDownGraph();
        m_fGraphFailure = true;
        return hr;
    }

	// Now lets try to add the Main Demuxer
	// First try to connect directly to the Tuner Filter
	// If this fails try to add a capture filter for
	// this device and tray again.
	if(FAILED (hr = RenderDemux(1)))
	{
		log("win32-DLL-BDA : Loading Capture Filter\n");
		
		if(FAILED (hr = LoadFilter (
									KSCATEGORY_BDA_RECEIVER_COMPONENT, 
									&m_pCaptureDevice,
									m_pTunerDevice, 
									TRUE, -1
									)))
		{
			log("win32-DLL-BDA : Cannot load capture device and connect tuner\n");
			TearDownGraph();
			m_fGraphFailure = true;
			return hr;
		}

		if(FAILED (hr = RenderDemux(2)))
		{
			log("win32-DLL-BDA : Cannot Render Main Demux\n");
			TearDownGraph();
			m_fGraphFailure = true;
			return hr;
		}
	}

    m_fGraphBuilt = true;
    m_fGraphFailure = false;

	log("win32-DLL-BDA : Capture Graph Built\n");
    
	log("win32-DLL-BDA : Registering Graph in ROT\n");


#ifdef REGISTER_FILTERGRAPH
    hr = AddGraphToRot (m_pFilterGraph, &m_dwGraphRegister);
    if (FAILED(hr))
    {
        log("win32-DLL-BDA : Failed to register filter graph with ROT!  hr=0x%x", hr);
        m_dwGraphRegister = 0;
    }
#endif

    return S_OK;
}


// Loads the correct tuning space based on NETWORK_TYPE that got
// passed into BuildGraph()
HRESULT
CBDAFilterGraph::LoadTuningSpace()
{   
	HRESULT hr = S_OK;
/*
    CComPtr <ITuningSpaceContainer>  pITuningSpaceContainer;

    // get the tuningspace container for all the tuning spaces from SYSTEM_TUNING_SPACES
    hr = pITuningSpaceContainer.CoCreateInstance(CLSID_SystemTuningSpaces);
    if (FAILED (hr))
    {
        printf("win32-DLL-BDA : Could not CoCreate SystemTuningSpaces\n");
        return hr;
    }

    CComVariant var (m_NetworkType);

    hr = pITuningSpaceContainer->get_Item(var, &m_pITuningSpace);

    if(FAILED(hr))
    {
        printf("win32-DLL-BDA : Unable to retrieve Tuning Space\n");
    }
*/


	m_pITuningSpace.CoCreateInstance(CLSID_DVBTuningSpace);

	CComQIPtr <IDVBTuningSpace2> piDVBTuningSpace(m_pITuningSpace);

	if (!piDVBTuningSpace)
	{
		m_pITuningSpace.Release();
		return E_FAIL;
	}

	if (FAILED(hr = piDVBTuningSpace->put_SystemType(DVB_Terrestrial)))
	{
		piDVBTuningSpace.Release();
		m_pITuningSpace.Release();
		return E_FAIL;
	}

	CComBSTR bstrNetworkType = "{216C62DF-6D7F-4E9A-8571-05F14EDB766A}";

	if (FAILED(hr = piDVBTuningSpace->put_NetworkType(bstrNetworkType)))
	{
		piDVBTuningSpace.Release();
		m_pITuningSpace.Release();
		return E_FAIL;
	}

	piDVBTuningSpace.Release();


    return hr;
}



// Creates an DVB Tune Request
HRESULT
CBDAFilterGraph::CreateDVBTTuneRequest(
        LONG lFreq,
        LONG lBand, 
        IDVBTuneRequest**   pTuneRequest
    )
{
    HRESULT hr = S_OK;

    if (pTuneRequest == NULL)
    {
        log("win32-DLL-BDA : Invalid pointer\n");
        return E_POINTER;
    }

    // Making sure we have a valid tuning space
    if (m_pITuningSpace == NULL)
    {
        log("win32-DLL-BDA : Tuning Space is NULL\n");
        return E_FAIL;
    }

    //  Create an instance of the DVB-T tuning space
    CComQIPtr <IDVBTuningSpace2> pDVBTTuningSpace (m_pITuningSpace);
    if (!pDVBTTuningSpace)
    {
        log("win32-DLL-BDA : Cannot QI for an IDVBTTuningSpace\n");
        return E_FAIL;
    }

    //  Create an empty tune request.
    CComPtr <ITuneRequest> pNewTuneRequest;
    hr = pDVBTTuningSpace->CreateTuneRequest(&pNewTuneRequest);

    if (FAILED (hr))
    {
        log("win32-DLL-BDA : CreateTuneRequest: Can't create tune request.\n");
        return hr;
    }


    CComQIPtr <IDVBTuneRequest> pDVBTTuneRequest (pNewTuneRequest);
    if (!pDVBTTuneRequest)
    {
        log("win32-DLL-BDA : CreateDVBTuneRequest: Can't QI for IDVBTuneRequest.\n");
        return E_FAIL;
    }


	//
	// Create a new locator and set it values
	//
    CComPtr <IDVBTLocator> pDVBTLocator;
    hr = pDVBTLocator.CoCreateInstance (CLSID_DVBTLocator);
    if (FAILED( hr))
    {
        log("win32-DLL-BDA : Cannot create the DVBT locator failed\n");
        return hr;
    }
	hr = pDVBTLocator->put_Bandwidth(lBand);
    if (FAILED( hr))
    {
        log("win32-DLL-BDA : Cannot set Bandwidth on Locator\n");
        return hr;
    }
	hr = pDVBTLocator->put_CarrierFrequency(lFreq);
    if (FAILED( hr))
    {
        log("win32-DLL-BDA : Cannot set CarrierFrequency on Locator\n");
        return hr;
    }

	//
	// Add the locator to our created Tune Request
	//
    hr = pDVBTTuneRequest->put_Locator(pDVBTLocator);
    if (FAILED (hr))
    {
        log("win32-DLL-BDA : Cannot put the locator\n");
        return hr;
    }

	pDVBTTuneRequest->put_ONID(-1);
	pDVBTTuneRequest->put_SID(-1);
	pDVBTTuneRequest->put_TSID(-1);

    hr = pDVBTTuneRequest.QueryInterface (pTuneRequest);

    return hr;
}


// LoadNetworkProvider loads network provider
HRESULT
CBDAFilterGraph::LoadNetworkProvider()
{
    HRESULT     hr = S_OK;
    CComBSTR    bstrNetworkType;
    CLSID       CLSIDNetworkType;

    // obtain tuning space then load network provider
    if(m_pITuningSpace == NULL)
    {
        hr = LoadTuningSpace();
        if(FAILED(hr))
        {
            log("win32-DLL-BDA : Cannot load TuningSpace\n");
            return hr;
        }
    }

    // Get the current Network Type clsid
    hr = m_pITuningSpace->get_NetworkType(&bstrNetworkType);
    if (FAILED (hr))
    {
        log("win32-DLL-BDA : ITuningSpace::Get Network Type failed\n");
        return hr;
    }

    hr = CLSIDFromString(bstrNetworkType, &CLSIDNetworkType);
    if (FAILED (hr))
    {
        log("win32-DLL-BDA : Couldn't get CLSIDFromString\n");
        return hr;
    }

    // create the network provider based on the clsid obtained from the tuning space
    hr = CoCreateInstance(CLSIDNetworkType, NULL, CLSCTX_INPROC_SERVER,
                          IID_IBaseFilter, 
                          reinterpret_cast<void**>(&m_pNetworkProvider));
    if (FAILED (hr))
    {
        log("win32-DLL-BDA : Couldn't CoCreate Network Provider\n");
        return hr;
    }

    //add the Network Provider filter to the graph
    hr = m_pFilterGraph->AddFilter(m_pNetworkProvider, L"Network Provider");

    return hr;
}


// enumerates through registered filters
// instantiates the the filter object and adds it to the graph
// it checks to see if it connects to upstream filter
// if not,  on to the next enumerated filter
// used for tuner, capture, MPE Data Filters and decoders that
// could have more than one filter object
// if pUpstreamFilter is NULL don't bother connecting
HRESULT
CBDAFilterGraph::LoadFilter(
    REFCLSID clsid, 
    IBaseFilter** ppFilter,
    IBaseFilter* pConnectFilter, 
    BOOL fIsUpstream,
	int index
    )
{
    HRESULT                 hr = S_OK;
    BOOL                    fFoundFilter = FALSE;
    CComPtr <IMoniker>      pIMoniker;
    CComPtr <IEnumMoniker>  pIEnumMoniker;

    if (!m_pICreateDevEnum)
    {
        hr = m_pICreateDevEnum.CoCreateInstance(CLSID_SystemDeviceEnum);
        if (FAILED (hr))
        {
            log("win32-DLL-BDA : LoadFilter(): Cannot CoCreate ICreateDevEnum\n");
            return hr;
        }
    }

    // obtain the enumerator
    hr = m_pICreateDevEnum->CreateClassEnumerator(clsid, &pIEnumMoniker, 0);
    // the call can return S_FALSE if no moniker exists, so explicitly check S_OK
    if (FAILED (hr))
    {
        log("win32-DLL-BDA : LoadFilter(): Cannot CreateClassEnumerator\n");
        return hr;
    }
    if (S_OK != hr)  // Class not found
    {
        log("win32-DLL-BDA : LoadFilter(): Class not found, CreateClassEnumerator returned S_FALSE");
        return E_UNEXPECTED;
    }

	int count = 0;
    // next filter
    while(pIEnumMoniker->Next(1, &pIMoniker, 0) == S_OK)
    {

		if(index == -1 || index == count)
		{
		
			// obtain filter's friendly name
			CComPtr <IPropertyBag>  pBag;
			hr = pIMoniker->BindToStorage(
										NULL, 
										NULL, 
										IID_IPropertyBag,
										reinterpret_cast<void**>(&pBag)
										);

			if(FAILED(hr))
			{
				OutputDebugString (TEXT("LoadFilter(): Cannot BindToStorage"));
				return hr;
			}

			CComVariant varBSTR;
			hr = pBag->Read(L"FriendlyName", &varBSTR, NULL);
			if(FAILED(hr))
			{
				OutputDebugString (TEXT("LoadFilter(): IPropertyBag->Read method failed"));
				pIMoniker = NULL;
				continue;
			}

			log("win32-DLL-BDA : LoadFilter(): %S\n", varBSTR.bstrVal);

			// bind the filter
			CComPtr <IBaseFilter>   pFilter;
			hr = pIMoniker->BindToObject(
										NULL, 
										NULL, 
										IID_IBaseFilter,
										reinterpret_cast<void**>(&pFilter)
										);

			if (FAILED(hr))
			{
				pIMoniker = NULL;
				pFilter = NULL;
				continue;
			}


			hr = m_pFilterGraph->AddFilter (pFilter, varBSTR.bstrVal);

			if (FAILED(hr))
			{
				OutputDebugString (TEXT("Cannot add filter\n"));
				return hr;
			}

			//MessageBox (NULL, _T(""), _T(""), MB_OK);
			// test connections
			// to upstream filter
			if (pConnectFilter)
			{
				if(fIsUpstream)
				{
					hr = ConnectFilters (pConnectFilter, pFilter);
				}
				else
				{
					hr = ConnectFilters (pFilter, pConnectFilter);
				}

				if(SUCCEEDED(hr))
				{
					//that's the filter we want
					fFoundFilter = TRUE;
					pFilter.QueryInterface (ppFilter);
					break;
				}
				else
				{
					fFoundFilter = FALSE;
					// that wasn't the the filter we wanted
					// so unload and try the next one
					hr = m_pFilterGraph->RemoveFilter(pFilter);

					if(FAILED(hr))
					{
						OutputDebugString(TEXT("Failed unloading Filter\n"));
						return hr;
					}
				}
			}
			else
			{
				fFoundFilter = TRUE;
				pFilter.QueryInterface (ppFilter);
				break;
			}


			pFilter = NULL;
		}

        pIMoniker = NULL;
		count++;

    } // while

	if(fFoundFilter == TRUE)
		return S_OK;
	else
		return -1;
}

HRESULT CBDAFilterGraph::LoadDump2()
{
    HRESULT hr = S_OK;

	CLSID dumpCID;
	CComBSTR bstrNetworkType ("{36A5F770-FE4C-11CE-A8ED-00AA002FEAB5}");

	hr = CLSIDFromString(bstrNetworkType, &dumpCID);
    
    hr = CoCreateInstance(
                        dumpCID, 
                        NULL, 
                        CLSCTX_INPROC_SERVER,
                        IID_IBaseFilter, 
                        reinterpret_cast<void**>(&m_pDump2)
                        );
    if (FAILED (hr))
    {
        log("win32-DLL-BDA : Could not CoCreateInstance Dump Filter\n");
        return hr;
    }

    hr = m_pFilterGraph->AddFilter(m_pDump2, L"Dump2");
    if(FAILED(hr))
    {
        log("win32-DLL-BDA : Unable to add Dump filter to graph\n");
        return hr;
    }

    return hr;
}

HRESULT CBDAFilterGraph::LoadDump()
{
    HRESULT hr = S_OK;

	CLSID dumpCID;
	CComBSTR bstrNetworkType ("{36A5F770-FE4C-11CE-A8ED-00AA002FEAB5}");

	hr = CLSIDFromString(bstrNetworkType, &dumpCID);
    
    hr = CoCreateInstance(
                        dumpCID, 
                        NULL, 
                        CLSCTX_INPROC_SERVER,
                        IID_IBaseFilter, 
                        reinterpret_cast<void**>(&m_pDump)
                        );
    if (FAILED (hr))
    {
        log("win32-DLL-BDA : Could not CoCreateInstance Dump Filter\n");
        return hr;
    }

    hr = m_pFilterGraph->AddFilter(m_pDump, L"Dump");
    if(FAILED(hr))
    {
        log("win32-DLL-BDA : Unable to add Dump filter to graph\n");
        return hr;
    }

    return hr;
}

HRESULT CBDAFilterGraph::LoadFlow01()
{
    HRESULT hr = S_OK;

	CLSID cID;
	CComBSTR bstrCID ("{FC13D753-DB71-46d7-9974-98635EAF60F0}");

	hr = CLSIDFromString(bstrCID, &cID);
    
    hr = CoCreateInstance(
                        cID, 
                        NULL, 
                        CLSCTX_INPROC_SERVER,
                        IID_IBaseFilter, 
                        reinterpret_cast<void**>(&m_pFlow01)
                        );
    if (FAILED (hr))
    {
        log("win32-DLL-BDA : Could not CoCreateInstance Flow01 Filter\n");
        return hr;
    }

    hr = m_pFilterGraph->AddFilter(m_pFlow01, L"Flow01");
    if(FAILED(hr))
    {
        log("win32-DLL-BDA : Unable to add Flow01 filter to graph\n");
        return hr;
    }

    return hr;
}

HRESULT CBDAFilterGraph::LoadFlow02()
{
    HRESULT hr = S_OK;

	CLSID cID;
	CComBSTR bstrCID ("{FC13D753-DB71-46d7-9974-98635EAF60F0}");

	hr = CLSIDFromString(bstrCID, &cID);
    
    hr = CoCreateInstance(
                        cID, 
                        NULL, 
                        CLSCTX_INPROC_SERVER,
                        IID_IBaseFilter, 
                        reinterpret_cast<void**>(&m_pFlow02)
                        );
    if (FAILED (hr))
    {
        log("win32-DLL-BDA : Could not CoCreateInstance Flow02 Filter\n");
        return hr;
    }

    hr = m_pFilterGraph->AddFilter(m_pFlow02, L"Flow02");
    if(FAILED(hr))
    {
        log("win32-DLL-BDA : Unable to add Flow02 filter to graph\n");
        return hr;
    }

    return hr;
}

// loads the demux into the FilterGraph
HRESULT
CBDAFilterGraph::LoadDemux()
{
    HRESULT hr = S_OK;
    
    hr = CoCreateInstance(
                        CLSID_MPEG2Demultiplexer, 
                        NULL, 
                        CLSCTX_INPROC_SERVER,
                        IID_IBaseFilter, 
                        reinterpret_cast<void**>(&m_pDemux)
                        );
    if (FAILED (hr))
    {
        log("win32-DLL-BDA : Could not CoCreateInstance CLSID_MPEG2Demultiplexer\n");
        return hr;
    }

    hr = m_pFilterGraph->AddFilter(m_pDemux, L"Demux");
    if(FAILED(hr))
    {
        log("win32-DLL-BDA : Unable to add demux filter to graph\n");
        return hr;
    }

    return hr;
}


// loads the demux into the FilterGraph
HRESULT
CBDAFilterGraph::LoadCapDemux()
{
    HRESULT hr = S_OK;
    
    hr = CoCreateInstance(
                        CLSID_MPEG2Demultiplexer, 
                        NULL, 
                        CLSCTX_INPROC_SERVER,
                        IID_IBaseFilter, 
                        reinterpret_cast<void**>(&m_pCapDemux)
                        );
    if (FAILED (hr))
    {
        log("win32-DLL-BDA : Could not CoCreateInstance CLSID_MPEG2Demultiplexer\n");
        return hr;
    }

    hr = m_pFilterGraph->AddFilter(m_pCapDemux, L"Cap Demux");
    if(FAILED(hr))
    {
        log("win32-DLL-BDA : Unable to add Cap demux filter to graph\n");
        return hr;
    }

    return hr;
}


// loads the demux into the FilterGraph
HRESULT
CBDAFilterGraph::LoadInfTee()
{
    HRESULT hr = S_OK;
    
    hr = CoCreateInstance(
                        CLSID_InfTee, 
                        NULL, 
                        CLSCTX_INPROC_SERVER,
                        IID_IBaseFilter, 
                        reinterpret_cast<void**>(&m_pInfTee)
                        );
    if (FAILED (hr))
    {
        log("win32-DLL-BDA : Could not CoCreateInstance CLSID_InfTee\n");
        return hr;
    }

    hr = m_pFilterGraph->AddFilter(m_pInfTee, L"Inf Tee");
    if(FAILED(hr))
    {
        log("win32-DLL-BDA : Unable to add Inf Tee filter to graph\n");
        return hr;
    }

    return hr;
}

// renders demux output pins
HRESULT
CBDAFilterGraph::RenderDemux(int type)
{
	log("win32-DLL-BDA : Doing Main Demux Render with type %d capture device.\n", type);

    HRESULT             hr = S_OK;
    CComPtr <IPin>      pIPin;
    CComPtr <IPin>      pDownstreamPin;
    CComPtr <IEnumPins> pIEnumPins;

	// load the demux
	log("win32-DLL-BDA : Loading Main Demux\n");
    if(FAILED (hr = LoadDemux()))
    {
        log("win32-DLL-BDA : Cannot load Main Demux\n");
		if(m_pDemux != NULL)
		{
			m_pFilterGraph->RemoveFilter(m_pDemux);
			m_pDemux.Release();
			m_pDemux = NULL;
		}
        m_fGraphFailure = true;
        return hr;
    }

	// Load the Inf Tee
	hr = LoadInfTee();
    if (FAILED (hr))
    {
        log("win32-DLL-BDA : Cannot Load Inf Tee filter\n");
        return hr;
    }

	if(type == 1)
	{
		// Connecting the Int Tee to the Tuner Filter
		log("win32-DLL-BDA : Connecting the Inf Tee to the Tuner Filter\n");
		hr = ConnectFilters (m_pTunerDevice, m_pInfTee);
		if (FAILED (hr))
		{
			log("win32-DLL-BDA : Cannot connect Inf Tee to Tuner filter\n");
			m_pFilterGraph->RemoveFilter(m_pInfTee);
			m_pInfTee.Release();
			m_pInfTee = NULL;
			m_pFilterGraph->RemoveFilter(m_pDemux);
			m_pDemux.Release();
			m_pDemux = NULL;
			return hr;
		}
	}
	else if(type == 2)
	{
		// Connecting the Int Tee to the Cap Filter
		log("win32-DLL-BDA : Connecting the Inf Tee to the Cap Filter\n");
		hr = ConnectFilters (m_pCaptureDevice, m_pInfTee);
		if (FAILED (hr))
		{
			log("win32-DLL-BDA : Cannot connect Inf Tee to capture filter\n");
			m_pFilterGraph->RemoveFilter(m_pInfTee);
			m_pInfTee.Release();
			m_pInfTee = NULL;
			m_pFilterGraph->RemoveFilter(m_pDemux);
			m_pDemux.Release();
			m_pDemux = NULL;
			return hr;
		}
	}

    // connect the demux to inf tee
    hr = ConnectFilters (m_pInfTee, m_pDemux);

    if (FAILED (hr))
    {
        log("win32-DLL-BDA : Cannot connect Inf Tee to demux\n");
		m_pFilterGraph->RemoveFilter(m_pInfTee);
		m_pInfTee.Release();
		m_pInfTee = NULL;
		m_pFilterGraph->RemoveFilter(m_pDemux);
		m_pDemux.Release();
		m_pDemux = NULL;
        return hr;
    }

	// Get an instance of the Demux control interface
	CComPtr <IMpeg2Demultiplexer> muxInterface;
	hr = m_pDemux.QueryInterface (&muxInterface);

    if (FAILED (hr))
    {
        log("win32-DLL-BDA : Cannot QI IMpeg2Demultiplexer on Demux\n");
        return hr;
    }

	//muxInterface->DeleteOutputPin(L"1");
	//muxInterface->DeleteOutputPin(L"2");
	//muxInterface->DeleteOutputPin(L"3");
	//muxInterface->DeleteOutputPin(L"4");
	//muxInterface->DeleteOutputPin(L"5");

	m_pCapDemux = NULL;

    // load transform information filter and connect it to the demux
    hr = LoadFilter (
                    KSCATEGORY_BDA_TRANSPORT_INFORMATION, 
                    &m_pTIF, 
                    m_pDemux, 
                    TRUE, -1
                    );
    if (FAILED (hr))
    {
        log("win32-DLL-BDA : Cannot load TIF\n");
        return hr;
    }

    return hr;
}

__int64 CBDAFilterGraph::getFlow02()
{
	__int64 count = -1;

	if(m_pFlow02 == NULL)
		return count;

	IFlowMeter *flow02 = NULL;

	try
	{
		HRESULT hr = m_pFlow02->QueryInterface(IID_IFlowMeter, (void **)&flow02);

		if (FAILED (hr) || flow02 == NULL)
		{
			log("win32-DLL-BDA : Cannot QI IFlowMeter on flow02 filter\n");
			return count;
		}

		flow02->get_FlowData(&count);

		flow02->Release();

	}
	catch(...)
	{
		log("win32-DLL-BDA : Exception Thrown when trying to get flow count for Flow02\n");
	}

	return count;
}

__int64 CBDAFilterGraph::getFlow01()
{
	__int64 count = -1;

	if(m_pFlow01 == NULL)
		return count;

	IFlowMeter *flow01 = NULL;

	try
	{
		HRESULT hr = m_pFlow01->QueryInterface(IID_IFlowMeter, (void **)&flow01);

		if (FAILED (hr) || flow01 == NULL)
		{
			log("win32-DLL-BDA : Cannot QI IFlowMeter on flow01 filter\n");
			return count;
		}

		flow01->get_FlowData(&count);

		flow01->Release();
	}
	catch(...)
	{
		log("win32-DLL-BDA : Exception Thrown when trying to get flow count for Flow01\n");
	}

	return count;
}


HRESULT
CBDAFilterGraph::SetUpDvrMSCap()
{
	HRESULT hr = S_OK;

    hr = CoCreateInstance(
                        CLSID_StreamBufferSink, 
                        NULL, 
                        CLSCTX_INPROC_SERVER,
                        IID_IBaseFilter, 
                        reinterpret_cast<void**>(&m_pSBsink)
                        );
	
    if (FAILED (hr))
    {
        log("win32-DLL-BDA : Cannot Create Stream Buffer Sink Filter\n");
        return hr;
    }

    hr = m_pFilterGraph->AddFilter(m_pSBsink, L"Stream Buffer Sink");
    if(FAILED(hr))
    {
        log("win32-DLL-BDA : Unable to add Stream Buffer Sink Filter to graph\n");
        return hr;
    }

	ChangeChannel(m_lFrequency, m_lBandwidth);

	
    CComPtr <IPin> pIVPin;
	CComPtr <IPin> pIAPin;

	hr = LoadCapDemux();
    if (FAILED (hr))
    {
        log("win32-DLL-BDA : Cannot Load Cap Demux filter\n");
        return hr;
    }

	hr = ConnectFilters (m_pInfTee, m_pCapDemux);
	if (FAILED (hr))
    {
        log("win32-DLL-BDA : Cannot connect Infinite Tee to capture Demux filter\n");
        return hr;
    }

	// Get an instance of the Demux control interface
	CComPtr <IMpeg2Demultiplexer> muxCapInterface;
	hr = m_pCapDemux.QueryInterface (&muxCapInterface);

	//
	// Set up and connect the Video output PIN
	//
	AM_MEDIA_TYPE mtVideo;
	ZeroMemory(&mtVideo, sizeof(AM_MEDIA_TYPE));

	mtVideo.majortype = MEDIATYPE_Video;
	mtVideo.subtype = MEDIASUBTYPE_MPEG2_VIDEO;
	mtVideo.formattype = FORMAT_MPEG2Video;

	BYTE SeqHdr[] = { 
	0x00, 0x00, 0x01, 0xb3, 0x2d, 0x02, 0x40, 0x33, 
	0x24, 0x9f, 0x23, 0x81, 0x10, 0x11, 0x11, 0x12, 
	0x12, 0x12, 0x13, 0x13, 0x13, 0x13, 0x14, 0x14, 
	0x14, 0x14, 0x14, 0x15, 0x15, 0x15, 0x15, 0x15, 
	0x15, 0x16, 0x16, 0x16, 0x16, 0x16, 0x16, 0x16, 
	0x17, 0x17, 0x17, 0x17, 0x17, 0x17, 0x17, 0x17, 
	0x18, 0x18, 0x18, 0x19, 0x18, 0x18, 0x18, 0x19, 
	0x1a, 0x1a, 0x1a, 0x1a, 0x19, 0x1b, 0x1b, 0x1b, 
	0x1b, 0x1b, 0x1c, 0x1c, 0x1c, 0x1c, 0x1e, 0x1e, 
	0x1e, 0x1f, 0x1f, 0x21 };

	// Allocate the format block, including space for the sequence header. 
	mtVideo.cbFormat = sizeof(MPEG2VIDEOINFO) + sizeof(SeqHdr);
	mtVideo.pbFormat = (BYTE*)CoTaskMemAlloc(mtVideo.cbFormat);
	ZeroMemory(mtVideo.pbFormat, mtVideo.cbFormat);

	// Cast the buffer pointer to an MPEG2VIDEOINFO struct.
	MPEG2VIDEOINFO *pMVIH = (MPEG2VIDEOINFO*)mtVideo.pbFormat;

	RECT rcSrc = {0, 576, 0, 720};        // Source rectangle.
	pMVIH->hdr.rcSource = rcSrc;
	pMVIH->hdr.dwBitRate = 6000000;       // Bit rate 6 meg bits sec.
	pMVIH->hdr.AvgTimePerFrame = 400000;  // 25 fps.
	pMVIH->hdr.dwPictAspectRatioX = 16;   // 16:9 aspect ratio.
	pMVIH->hdr.dwPictAspectRatioY = 9;

	// BITMAPINFOHEADER information.
	pMVIH->hdr.bmiHeader.biSize = 40;
	pMVIH->hdr.bmiHeader.biWidth = 720;
	pMVIH->hdr.bmiHeader.biHeight = 576;

	pMVIH->dwLevel = AM_MPEG2Profile_Main;  // MPEG-2 profile. 
	pMVIH->dwProfile = AM_MPEG2Level_Main;  // MPEG-2 level.

	pMVIH->cbSequenceHeader = sizeof(SeqHdr); // Size of sequence header.
	memcpy(pMVIH->dwSequenceHeader, SeqHdr, sizeof(SeqHdr));

	muxCapInterface->CreateOutputPin(&mtVideo, L"Video" ,&pIVPin);

	// Get the Pid Map interface of the pin
	// and map the pids we want.
	CComPtr <IMPEG2PIDMap> muxVMapPid;

	hr = pIVPin.QueryInterface (&muxVMapPid);

	log("win32-DLL-BDA : Mapping Video PID %d\n", m_lvPID);

	ULONG pid = m_lvPID;
	hr = muxVMapPid->MapPID(1, &pid, MEDIA_ELEMENTARY_STREAM);

	// Load and add the Flow Meter filter for video
	hr = LoadFlow01();
	if (FAILED (hr))
    {
        log("win32-DLL-BDA : Cannot Load FlowFilter for Video\n");
        return hr;
    }

	// Connect the flow meter filter for video
	hr = ConnectFilters(m_pCapDemux, m_pFlow01);
    if (FAILED (hr))
    {
        log("win32-DLL-BDA : Cannot connect CapDemux and Flow01 0x%x\n", hr);
        return hr;
    }

	hr = ConnectFilters(m_pFlow01, m_pSBsink);
    if (FAILED (hr))
    {
        log("win32-DLL-BDA : Cannot connect Flow01 and Stream Buffer Sink (video) 0x%x\n", hr);
        return hr;
    }

	log("win32-DLL-BDA : Video connected to SB\n");

	//
	// Set up and connect the Audio output PIN
	//

	AM_MEDIA_TYPE mtAudio;
	ZeroMemory(&mtAudio, sizeof(AM_MEDIA_TYPE));
	mtAudio.majortype = MEDIATYPE_Audio;
	mtAudio.subtype = MEDIASUBTYPE_MPEG1AudioPayload;
	mtAudio.bFixedSizeSamples = TRUE;
	mtAudio.bTemporalCompression = 0;
	mtAudio.lSampleSize = 1;
	mtAudio.formattype = FORMAT_WaveFormatEx;
	mtAudio.pUnk = NULL;

	if (m_iaType == TYPE_AUDIO_MPG)
	{
		mtAudio.subtype = MEDIASUBTYPE_MPEG2_AUDIO;
	}
	else
	{
		mtAudio.subtype = MEDIASUBTYPE_DOLBY_AC3;
	}

	mtAudio.cbFormat = sizeof g_MPEG1AudioFormat;
	mtAudio.pbFormat = g_MPEG1AudioFormat;
	muxCapInterface->CreateOutputPin(&mtAudio, L"Audio" ,&pIAPin);

	CComPtr <IMPEG2PIDMap> muxAMapPid;

	hr = pIAPin.QueryInterface(&muxAMapPid);

	log("win32-DLL-BDA : Mapping Audio PID %d of type %d\n", m_laPID, m_iaType);

	pid = m_laPID;
	hr = muxAMapPid->MapPID(1, &pid, MEDIA_ELEMENTARY_STREAM);

	// Load and add the Flow Meter filter for audio
	hr = LoadFlow02();
	if (FAILED (hr))
    {
        log("win32-DLL-BDA : Cannot Load FlowFilter for Audio\n");
        return hr;
    }

	// Connect the flow meter filter for audio
	hr = ConnectFilters(m_pCapDemux, m_pFlow02);
    if (FAILED (hr))
    {
        log("win32-DLL-BDA : Cannot connect CapDemux and Flow02 0x%x\n", hr);
        return hr;
    }


	hr = ConnectFilters (m_pFlow02, m_pSBsink);
    if (FAILED (hr))
    {
        log("win32-DLL-BDA : Cannot connect Flow02 and Stream Buffer Sink (audio) 0x%x\n", hr);
        return hr;
    }

	log("win32-DLL-BDA : Audio connected to SB\n");

	//
	// Setup and start the Stream Buffer capture
	//

	CComPtr <IStreamBufferSink> pISBSi;
	hr = m_pSBsink.QueryInterface (&pISBSi);
    if (FAILED (hr))
    {
        log("win32-DLL-BDA : Cannot QueryInterface for IStreamBufferSink on Stream Buffer Sink 0x%x\n", hr);
        return hr;
    }

	hr = pISBSi->LockProfile(NULL);
    if (FAILED (hr))
    {
        log("win32-DLL-BDA : Cannot LockProfile on Stream Buffer Sink 0x%x\n", hr);
        return hr;
    }

	std::string name = dests.at(0);
	WCHAR test[512];
	mbstowcs(test, name.c_str(), 512); 

	hr = pISBSi->CreateRecorder(test, RECORDING_TYPE_CONTENT, &m_pRecUnk);
    if (FAILED (hr))
    {
        log("win32-DLL-BDA : Cannot CreateRecorder on Stream Buffer Sink 0x%x\n", hr);
        return hr;
    }

	CComQIPtr<IStreamBufferRecordControl> pRecControl;
	REFERENCE_TIME rtStart = 0;

	hr = m_pRecUnk.QueryInterface (&pRecControl);
    if (FAILED (hr))
    {
        log("win32-DLL-BDA : Cannot IQ IStreamBufferRecordControl 0x%x\n", hr);
        return hr;
    }

	hr = pRecControl->Start(&rtStart);
    if (FAILED (hr))
    {
        log("win32-DLL-BDA : Cannot Start Capture on Stream Buffer 0x%x\n", hr);
        return hr;
    }

	return hr;
}

HRESULT
CBDAFilterGraph::SetUpSepFileCap()
{
	ChangeChannel(m_lFrequency, m_lBandwidth);

	HRESULT hr = S_OK;
    CComPtr <IPin> pIVPin;
	CComPtr <IPin> pIAPin;

	hr = LoadCapDemux();
    if (FAILED (hr))
    {
        log("win32-DLL-BDA : Cannot Load Cap Demux filter\n");
        return hr;
    }

	hr = ConnectFilters (m_pInfTee, m_pCapDemux);
	if (FAILED (hr))
    {
        log("win32-DLL-BDA : Cannot connect Infinite Tee to capture Demux filter\n");
        return hr;
    }

	// Get an instance of the Demux control interface
	CComPtr <IMpeg2Demultiplexer> muxCapInterface;
	hr = m_pCapDemux.QueryInterface (&muxCapInterface);

	//
	// Set up and connect the Video output PIN
	//
	AM_MEDIA_TYPE type;
	ZeroMemory(&type, sizeof(AM_MEDIA_TYPE));
	type.majortype = MEDIATYPE_Video;
	type.subtype = MEDIASUBTYPE_MPEG2_VIDEO;
	type.formattype = FORMAT_None;

	ULONG pid = -1;
	char fileName[512];
	WCHAR test[512];

	if(m_lvPID > -1)
	{
		muxCapInterface->CreateOutputPin(&type, L"Video" ,&pIVPin);

		// Get the Pid Map interface of the pin
		// and map the pids we want.
		CComPtr <IMPEG2PIDMap> muxVMapPid;

		hr = pIVPin.QueryInterface (&muxVMapPid);

		log("win32-DLL-BDA : Mapping Video PID %d\n", m_lvPID);

		pid = m_lvPID;
		hr = muxVMapPid->MapPID(1, &pid, MEDIA_ELEMENTARY_STREAM);

		// Load and add the Flow Meter filter for video
		hr = LoadFlow01();
		if (FAILED (hr))
		{
			log("win32-DLL-BDA : Cannot Load FlowFilter for Video\n");
			return hr;
		}

		// Connect the flow meter filter for video
		hr = ConnectFilters(m_pCapDemux, m_pFlow01);
		if (FAILED (hr))
		{
			log("win32-DLL-BDA : Cannot connect CapDemux and Flow01 0x%x\n", hr);
			return hr;
		}

		// Load the dump filter
		hr = LoadDump();
		if (FAILED (hr))
		{
			log("win32-DLL-BDA : Cannot add Dump Filter\n");
			return hr;
		}

		//
		// Set file name the dump filter will dump to
		//
		CComPtr <IFileSinkFilter> pIFileSink;
		hr = m_pDump.QueryInterface(&pIFileSink);
		if(FAILED (hr))
		{
			log("win32-DLL-BDA : Cannot find File Sink Interface on Dump Filter\n");
			TearDownGraph();
			m_fGraphFailure = true;
			return hr;
		}

		std::string name = dests.at(0);
		StringCchCopy(fileName, 512, name.c_str());
		StringCchCat(fileName, 512, ".mpv");
		mbstowcs(test, fileName, 512); 

		hr = pIFileSink->SetFileName(test, NULL);
		if (FAILED (hr))
		{
			log("win32-DLL-BDA : Can not set the dump filter capture file name.\n");
			TearDownGraph();
			m_fGraphFailure = true;
			return hr;
		}

		hr = ConnectFilters (m_pFlow01, m_pDump);
		if (FAILED (hr))
		{
			log("win32-DLL-BDA : Cannot connect CapDemux and Dump (video) 0x%x\n", hr);
			TearDownGraph();
			m_fGraphFailure = true;
			return hr;
		}
	}

	//
	// Set up and connect the Audio output PIN
	//
	if(m_laPID > -1)
	{
		type.majortype = MEDIATYPE_Audio;
		type.subtype = MEDIASUBTYPE_MPEG1Payload;
		type.formattype = FORMAT_None;

		muxCapInterface->CreateOutputPin(&type, L"Audio" ,&pIAPin);

		CComPtr <IMPEG2PIDMap> muxAMapPid;

		hr = pIAPin.QueryInterface(&muxAMapPid);

		log("win32-DLL-BDA : Mapping Audio PID %d\n", m_laPID);

		pid = m_laPID;
		hr = muxAMapPid->MapPID(1, &pid, MEDIA_ELEMENTARY_STREAM);

		// Load and add the Flow Meter filter for video
		hr = LoadFlow02();
		if (FAILED (hr))
		{
			log("win32-DLL-BDA : Cannot Load FlowFilter for Audio\n");
			return hr;
		}

		// Connect the flow meter filter for video
		hr = ConnectFilters(m_pCapDemux, m_pFlow02);
		if (FAILED (hr))
		{
			log("win32-DLL-BDA : Cannot connect CapDemux and Flow02 0x%x\n", hr);
			return hr;
		}

		// Load the dump filter
		hr = LoadDump2();
		if (FAILED (hr))
		{
			log("win32-DLL-BDA : Cannot add Dump Filter\n");
			return hr;
		}

		//
		// Set file name the dump filter will dump to
		//
		CComPtr <IFileSinkFilter> pIFileSink2;
		hr = m_pDump2.QueryInterface(&pIFileSink2);
		if(FAILED (hr))
		{
			log("win32-DLL-BDA : Cannot find File Sink Interface on Dump Filter\n");
			TearDownGraph();
			m_fGraphFailure = true;
			return hr;
		}

		std::string name = dests.at(0);
		StringCchCopy(fileName, 512, name.c_str());
		StringCchCat(fileName, 512, ".mpa");	
		mbstowcs(test, fileName, 512);

		hr = pIFileSink2->SetFileName(test, NULL);
		if (FAILED (hr))
		{
			log("win32-DLL-BDA : Can not set the dump filter capture file name.\n");
			return hr;
		}

		hr = ConnectFilters (m_pFlow02, m_pDump2);
		if (FAILED (hr))
		{
			log("win32-DLL-BDA : Cannot connect CapDemux and Dump (audio) 0x%x\n", hr);
			TearDownGraph();
			m_fGraphFailure = true;
			return hr;
		}
	}

	return hr;
}

// removes each filter from the graph
HRESULT
CBDAFilterGraph::addDumpFilters(std::vector <std::string> destinations)
{
	HRESULT hr = S_OK;

	//
	// create and add inf 'T'
	//
    hr = CoCreateInstance(
                        CLSID_InfTee, 
                        NULL, 
                        CLSCTX_INPROC_SERVER,
                        IID_IBaseFilter, 
                        reinterpret_cast<void**>(&m_pDumpInfT)
                        );
    if (FAILED (hr))
    {
        log("win32-DLL-BDA : Could not CoCreateInstance CLSID_InfTee\n");
        return hr;
    }

    hr = m_pFilterGraph->AddFilter(m_pDumpInfT, L"Dump Inf Tee");
    if(FAILED(hr))
    {
        log("win32-DLL-BDA : Unable to add Inf Tee filter to graph\n");
        return hr;
    }

	// Connect the dump inf 'T' to flowmeter
    hr = ConnectFilters (m_pFlow01, m_pDumpInfT);

    if (FAILED (hr))
    {
        log("win32-DLL-BDA : Cannot connect dump infT to flowmeter filter 0x%x\n", hr);
        return hr;
    }

	//
	// create add and setup dump filters
	//

	// Load the dump filter
	CLSID dumpCID;
	CComBSTR bstrNetworkType ("{36A5F770-FE4C-11CE-A8ED-00AA002FEAB5}");
	hr = CLSIDFromString(bstrNetworkType, &dumpCID);
    
	for(std::vector<std::string>::size_type x = 0; x < destinations.size(); x++)
	{
		CComPtr <IBaseFilter> m_pDumpFilterX;

		hr = CoCreateInstance(
							dumpCID, 
							NULL, 
							CLSCTX_INPROC_SERVER,
							IID_IBaseFilter, 
							reinterpret_cast<void**>(&m_pDumpFilterX)
							);
		if (FAILED (hr))
		{
			log("win32-DLL-BDA : Could not CoCreateInstance Dump Filter (%d)\n", x);
			return hr;
		}

		wchar_t filterName[56];
		StringCchPrintfW(filterName, 56, L"Dump Filter (%d)", x);
		hr = m_pFilterGraph->AddFilter(m_pDumpFilterX, filterName);
		if(FAILED(hr))
		{
			log("win32-DLL-BDA : Unable to add Dump filter to graph (%d) 0x%x\n", x , hr);
			return hr;
		}

		// add dump filter to dump list
		dumpFilters.push_back(m_pDumpFilterX);

		// Connect the Dump filter to the mux
		hr = ConnectFilters (m_pDumpInfT, m_pDumpFilterX);

		if (FAILED (hr))
		{
			log("win32-DLL-BDA : Cannot connect dump infT to dump filter %d\n", x);
			return hr;
		}

		//
		// Set file name the dump filter will dump to
		//
		CComPtr <IFileSinkFilter> pIFileSink;
		hr = m_pDumpFilterX.QueryInterface(&pIFileSink);
		if(FAILED (hr))
		{
			log("win32-DLL-BDA : Cannot find File Sink Interface on Dump Filter\n");
			TearDownGraph();
			m_fGraphFailure = true;
			return hr;
		}

		wchar_t capPath[512];
		std::string name = dests.at(x);
		mbstowcs(capPath, name.c_str(), 512); 

		hr = pIFileSink->SetFileName(capPath, NULL);
		if (FAILED (hr))
		{
			log("win32-DLL-BDA : Can not set the dump filter capture file name.\n");
			return hr;
		}

		//
		// enable the file writer, the first one is enabled and the rest are diabled
		//
		CComPtr <IFileWriter> pIFileWriter;
		HRESULT qhr = m_pDumpFilterX.QueryInterface(&pIFileWriter);
		if(FAILED (qhr) || pIFileWriter == NULL)
		{
			log("win32-DLL-BDA : Cannot find FileWriter Interface on Dump Filter\n");
		}
		else
		{
			if(x == 0)
			{
				log("win32-DLL-BDA : Enabling FileWriter Filter (%d)\n", x);
				pIFileWriter->WriteEnabled(TRUE);
			}
			else
			{
				log("win32-DLL-BDA : Disabling FileWriter Filter (%d)\n", x);
				pIFileWriter->WriteEnabled(FALSE);
			}
			pIFileWriter->SetFileType(1);
		}
	}

	return hr;
}

// removes each filter from the graph
HRESULT
CBDAFilterGraph::SetUpTSMuxCap()
{
	ChangeChannel(m_lFrequency, m_lBandwidth);

	HRESULT hr = S_OK;
    CComPtr <IPin>      pIPin;

	// Load and add the PSI filter so we can parse all the required PIDS
	m_pMpegSections = NULL;
	LoadPsiFilter();

	hr = LoadCapDemux();
    if (FAILED (hr))
    {
        log("win32-DLL-BDA : Cannot Load Cap Demux filter\n");
        return hr;
    }

	hr = ConnectFilters (m_pInfTee, m_pCapDemux);
	if (FAILED (hr))
    {
        log("win32-DLL-BDA : Cannot connect Infinite Tee to capture Demux filter\n");
        return hr;
    }

	// Get an instance of the Demux control interface
	CComPtr <IMpeg2Demultiplexer> muxCapInterface;
	hr = m_pCapDemux.QueryInterface (&muxCapInterface);

	// Create out new pin of type GUID_NULL
	AM_MEDIA_TYPE type;
	ZeroMemory(&type, sizeof(AM_MEDIA_TYPE));
	type.majortype = GUID_NULL;
	type.subtype = GUID_NULL;
	type.formattype = FORMAT_None;

	muxCapInterface->CreateOutputPin(&type, L"TS" ,&pIPin);

	// Load and add the Flow Meter filter
	hr = LoadFlow01();
	if (FAILED (hr))
    {
        log("win32-DLL-BDA : Cannot Load FlowFilter for Video\n");
        return hr;
    }

	// Connect the flow meter filter
	hr = ConnectFilters(m_pCapDemux, m_pFlow01);
    if (FAILED (hr))
    {
        log("win32-DLL-BDA : Cannot connect CapDemux and Flow01 0x%x\n", hr);
        return hr;
    }

	hr = addDumpFilters(dests);
    if (FAILED (hr))
    {
        log("win32-DLL-BDA : addDumpFilters Failed to add all the dump filters\n");
        return hr;
    }

	return hr;
}

// removes each filter from the graph
HRESULT
CBDAFilterGraph::SetUpTSCap()
{
	ChangeChannel(m_lFrequency, m_lBandwidth);
	HRESULT hr = S_OK;

	// Load and add the Flow Meter filter
	hr = LoadFlow01();
	if (FAILED (hr))
    {
        log("win32-DLL-BDA : Cannot Load FlowFilter for Video\n");
        return hr;
    }

	// Connect the Flow Meter Filter
	hr = ConnectFilters(m_pInfTee, m_pFlow01);
    if (FAILED (hr))
    {
        log("win32-DLL-BDA : Cannot connect CapDemux and Flow01 0x%x\n", hr);
        return hr;
    }

	// Load the dump filter
	hr = LoadDump();
    if (FAILED (hr))
    {
        log("win32-DLL-BDA : Cannot add Dump Filter\n");
        return hr;
    }

	// Connect the Dump filter to the mux
    hr = ConnectFilters (m_pFlow01, m_pDump);
    if (FAILED (hr))
    {
        log("win32-DLL-BDA : Cannot connect Infinite Tee to Dump filter\n");
        return hr;
    }

	//
	// Set file name the dump filter will dump to
	//
	CComPtr <IFileSinkFilter> pIFileSink;
	hr = m_pDump.QueryInterface(&pIFileSink);
    if(FAILED (hr))
    {
        log("win32-DLL-BDA : Cannot find File Sink Interface on Dump Filter\n");
        TearDownGraph();
        m_fGraphFailure = true;
        return hr;
    }

	WCHAR test[512];
	std::string name = dests.at(0);
	mbstowcs(test, name.c_str(), 512); 

	hr = pIFileSink->SetFileName(test, NULL);
    if (FAILED (hr))
    {
        log("win32-DLL-BDA : Can not set the dump filter capture file name.\n");
        return hr;
    }

	return hr;
}

// removes each filter from the graph
HRESULT
CBDAFilterGraph::TearDownGraph()
{
    HRESULT hr = S_OK;
    CComPtr <IBaseFilter> pFilter;
    CComPtr <IEnumFilters> pIFilterEnum;

    m_pITuningSpace = NULL;

    if(m_fGraphBuilt || m_fGraphFailure)
    {
        // unload manually added filters
        m_pFilterGraph->RemoveFilter(m_pIPSink);
        m_pFilterGraph->RemoveFilter(m_pMPE);
        m_pFilterGraph->RemoveFilter(m_pTIF);
        m_pFilterGraph->RemoveFilter(m_pDemux);
        m_pFilterGraph->RemoveFilter(m_pNetworkProvider);
        m_pFilterGraph->RemoveFilter(m_pTunerDevice);
        m_pFilterGraph->RemoveFilter(m_pCaptureDevice);
		m_pFilterGraph->RemoveFilter(m_pDump);
		m_pFilterGraph->RemoveFilter(m_pDump2);
		m_pFilterGraph->RemoveFilter(m_pCapDemux);
		m_pFilterGraph->RemoveFilter(m_pInfTee);
		m_pFilterGraph->RemoveFilter(m_pDumpInfT);

		for(std::vector<std::string>::size_type x = 0; x < dumpFilters.size(); x++)
		{
			m_pFilterGraph->RemoveFilter(dumpFilters.at(x));
		}
		
        m_pIPSink = NULL;
        m_pMPE = NULL;
        m_pTIF = NULL;
        m_pDemux = NULL;
        m_pNetworkProvider = NULL;
        m_pTunerDevice = NULL;
        m_pDemodulatorDevice = NULL;
        m_pCaptureDevice = NULL;
		m_pDump = NULL;
		m_pDump2 = NULL;
		m_pCapDemux = NULL;
		m_pInfTee = NULL;
		m_pDumpInfT = NULL;
		dumpFilters.empty();

		if (m_pMpegSections != NULL)
			m_pFilterGraph->RemoveFilter(m_pMpegSections);
		m_pMpegSections = NULL;

        // now go unload rendered filters
        hr = m_pFilterGraph->EnumFilters(&pIFilterEnum);

        if(FAILED(hr))
        {
            log("win32-DLL-BDA : TearDownGraph: cannot EnumFilters\n");
            return E_FAIL;
        }

        pIFilterEnum->Reset();

        while(pIFilterEnum->Next(1, &pFilter, 0) == S_OK) // addrefs filter
        {
            hr = m_pFilterGraph->RemoveFilter(pFilter);

            if (FAILED (hr))
                return hr;

            pIFilterEnum->Reset();
            pFilter.Release ();
        }
    }

#ifdef REGISTER_FILTERGRAPH
    if (m_dwGraphRegister)
    {
        RemoveGraphFromRot(m_dwGraphRegister);
        m_dwGraphRegister = 0;
    }
#endif

    m_fGraphBuilt = FALSE;
    return S_OK;
}


// ConnectFilters is called from BuildGraph
// to enumerate and connect pins
HRESULT
CBDAFilterGraph::ConnectFilters(
    IBaseFilter* pFilterUpstream, 
    IBaseFilter* pFilterDownstream
    )
{
    HRESULT         hr = E_FAIL;

    CComPtr <IPin>  pIPinUpstream;


    PIN_INFO        PinInfoUpstream;
    PIN_INFO        PinInfoDownstream;

    // validate passed in filters
    ASSERT (pFilterUpstream);
    ASSERT (pFilterDownstream);

    // grab upstream filter's enumerator
    CComPtr <IEnumPins> pIEnumPinsUpstream;
    hr = pFilterUpstream->EnumPins(&pIEnumPinsUpstream);

    if(FAILED(hr))
    {
        log("win32-DLL-BDA : Cannot Enumerate Upstream Filter's Pins\n");
        return hr;
    }

    // iterate through upstream filter's pins
    while (pIEnumPinsUpstream->Next (1, &pIPinUpstream, 0) == S_OK)
    {
        hr = pIPinUpstream->QueryPinInfo (&PinInfoUpstream);
        if(FAILED(hr))
        {
            log("win32-DLL-BDA : Cannot Obtain Upstream Filter's PIN_INFO\n");
            return hr;
        }

        CComPtr <IPin>  pPinDown;
        pIPinUpstream->ConnectedTo (&pPinDown);

        // bail if pins are connected
        // otherwise check direction and connect
        if ((PINDIR_OUTPUT == PinInfoUpstream.dir) && (pPinDown == NULL))
        {
            // grab downstream filter's enumerator
            CComPtr <IEnumPins> pIEnumPinsDownstream;
            hr = pFilterDownstream->EnumPins (&pIEnumPinsDownstream);
            if(FAILED(hr))
            {
                log("win32-DLL-BDA : Cannot enumerate pins on downstream filter!\n");
                return hr;
            }

            // iterate through downstream filter's pins
            CComPtr <IPin>  pIPinDownstream;
            while (pIEnumPinsDownstream->Next (1, &pIPinDownstream, 0) == S_OK)
            {
                // make sure it is an input pin
                hr = pIPinDownstream->QueryPinInfo(&PinInfoDownstream);
                if(SUCCEEDED(hr))
                {
                    CComPtr <IPin>  pPinUp;

                    // Determine if the pin is already connected.  Note that 
                    // VFW_E_NOT_CONNECTED is expected if the pin isn't yet connected.
                    hr = pIPinDownstream->ConnectedTo (&pPinUp);
                    if(FAILED(hr) && hr != VFW_E_NOT_CONNECTED)
                    {
                        log("win32-DLL-BDA : Failed in pIPinDownstream->ConnectedTo()!\n");
						pIPinDownstream = NULL;
                        continue;
                    }

                    if ((PINDIR_INPUT == PinInfoDownstream.dir) && (pPinUp == NULL))
                    {
                        if (SUCCEEDED (m_pFilterGraph->Connect(
                                        pIPinUpstream,
                                        pIPinDownstream))
                                        )
                        {
                            PinInfoDownstream.pFilter->Release();
                            PinInfoUpstream.pFilter->Release();
							pIPinDownstream = NULL;
                            return S_OK;
                        }
                    }

                }

                PinInfoDownstream.pFilter->Release();
                pIPinDownstream = NULL;
            } // while next downstream filter pin

            //We are now back into the upstream pin loop
        } // if output pin

        pIPinUpstream = NULL;
        PinInfoUpstream.pFilter->Release();
    } // while next upstream filter pin

    return E_FAIL;
}

// RunGraph checks to see if a graph has been built
// if not it calls BuildGraph
// RunGraph then calls MediaCtrl-Run
HRESULT
CBDAFilterGraph::RunGraph(std::string *resultText)
{
    // check to see if the graph is already running
    if(m_fGraphRunning)
        return S_OK;

    HRESULT hr01 = S_OK;
    if (m_pIMediaControl == NULL)
	{
        hr01 = m_pFilterGraph.QueryInterface (&m_pIMediaControl);
	}

    if (SUCCEEDED (hr01))
    {
        // run the graph
        hr01 = m_pIMediaControl->Run();
        if(SUCCEEDED(hr01))
        {
            m_fGraphRunning = true;
        }
        else
        {
            // stop parts of the graph that ran
            m_pIMediaControl->Stop();
            log("win32-DLL-BDA : Cannot run graph %x\n", hr01);
			messageText(resultText, "Cannot run graph %x\n", hr01);
			return hr01;
        }
    }
	else
	{
		log("win32-DLL-BDA : Cannot QI IMediaControl %x\n", hr01);
		return hr01;
	}

	// at this point the graph should be running and 
	// data should be flowing

	// wait for 2 sec before we do anything
	Sleep(2000);

	// Get the Signal Strength and show in log

	long sig = -99;
	m_pITuner->get_SignalStrength(&sig);
	log("win32-DLL-BDA : Signal Strength : %d\n", sig);
	messageText(resultText, "Signal Strength : %d\n", sig);
	
	// Now get some signal stats from the tuner filter 
	// using the device Topology system

	HRESULT hr03;
	int signalLocked = -1;
	CComPtr <IBDA_Topology> bdaNetTop;
	hr03 = m_pTunerDevice.QueryInterface(&bdaNetTop);

    if(SUCCEEDED(hr03))
	{
		CComPtr <IUnknown> unknown;
		hr03 = bdaNetTop->GetControlNode(0, 1, 1, &unknown);

		if(SUCCEEDED(hr03))
		{
			//log("win32-DLL-BDA : Got Control Node 1 %x\n", hr03);

			CComPtr <IBDA_SignalStatistics> pSigStats;
			hr03 = unknown.QueryInterface(&pSigStats);

			if(SUCCEEDED(hr03))
			{
				//log("win32-DLL-BDA : Found IBDA_SignalStatistics interface %x\n", hr03);

				// we now have the signal stats interface
				// get the values we want.

				// Get the Signal Quality value
				long quality = -1;
				hr03 = pSigStats->get_SignalQuality(&quality);

				if(hr03 == E_PROP_SET_UNSUPPORTED)
					log("win32-DLL-BDA : get_SignalQuality Not supported (%x)\n", hr03);
				else if(FAILED(hr03))
					log("win32-DLL-BDA : Error getting get_SignalQuality %x\n", hr03);
				else
				{
					log("win32-DLL-BDA : Signal Quality : %d\n", quality);
					messageText(resultText, "Signal Quality : %d\n", quality);
				}
					
				// Get the signal locked value form the 
				// tuner, this may return not supported
				BOOLEAN locked = FALSE;
				hr03 = pSigStats->get_SignalLocked(&locked);

				if(hr03 == E_PROP_SET_UNSUPPORTED)
					log("win32-DLL-BDA : get_SignalLocked Not supported %x\n", hr03);
				else if(FAILED(hr03))
					log("win32-DLL-BDA : Error getting get_SignalLocked %x\n", hr03);
				else
				{
					log("win32-DLL-BDA : Got Signal Locked : %d\n", locked);
					if(locked)
						signalLocked = 1;
					else
						signalLocked = 0;
				}
			}
			else
			{
				log("win32-DLL-BDA : Cannot QI IBDA_SignalStatistics %x\n", hr03);
			}
		}
		else
		{
			log("win32-DLL-BDA : Cannot GetControlNode %x\n", hr03);
		}
	}
	else
    {
		log("win32-DLL-BDA : Cannot Find IID_IBDA_Topology %x\n", hr03);
	}

	// if the signalLocked value is set to zero this
	// means the get_SignalLocked returned false
	// and the tuner is not locked to a frequency

	if(signalLocked == 0)
	{
		log("win32-DLL-BDA : Tuner Not Locked, returning with error -1\n");
		messageText(resultText, "Tuner Not Locked, returning with error -1\n");
		return -1;
	}

	// Stop the TIF filter, this is due to a bug in the TIF
	// filter and one day may be removed, the tiff filter can not
	// be running in two different graphs at the same time

	//log("win32-DLL-BDA : Stopping TIF filter\n");
	//m_pTIF->Stop();

	//
	// If we are a TS Mux capture parse all the extra PIDS we need
	//

	HRESULT hr02 = S_OK;
	CComPtr <IPin> pIPin;
	if(m_pCapDemux != NULL && SUCCEEDED(m_pCapDemux->FindPin(L"TS", &pIPin)))
	{

		CStreamScan scanner;
		scanner.scanForProgs(m_pMpegSections);

		int programID = scanner.findProgramID(m_lvPID);

		if(programID < 1)
			programID = scanner.findProgramID(m_laPID);

		if(programID > 0)
		{

			ULONG pidsArray[30];

			int numberPIDS = scanner.getTsPidArray(programID, pidsArray, 29);

			char buff[1024];
			StringCchPrintf(buff, 1024, "%d Pids mapped : ", numberPIDS);
			char pidString[10];

			for (int i = 0; i < numberPIDS; i++)
			{
				StringCchPrintf(pidString, 10, " %d ", pidsArray[i]);
				StringCchCat(buff, 1024, pidString);
			}

			StringCchCat(buff, 1024, "\n");
			log(buff);

			messageText(resultText, buff);

			CComPtr <IMPEG2PIDMap> muxMapPid;
			hr02 = pIPin.QueryInterface(&muxMapPid);
			if (SUCCEEDED (hr02))
			{
				if (SUCCEEDED (hr01))
				{
					hr02 = muxMapPid->MapPID(numberPIDS, pidsArray, MEDIA_TRANSPORT_PACKET);
				}
				else
				{
					log("win32-DLL-BDA : Cannot MapPID on IMPEG2PIDMap interface %x\n", hr02);
					return hr02;
				}
			}
			else
			{
				log("win32-DLL-BDA : Cannot QI IMPEG2PIDMap %x\n", hr02);
				return hr02;
			}

		}
		else
		{
			//Map Pids the old way
			CComPtr <IMPEG2PIDMap> muxMapPid;
			hr02 = pIPin.QueryInterface (&muxMapPid);

			messageText(resultText, "Stream scan failed, mapping default PIDS : %d %d\n", m_lvPID, m_laPID);

			log("win32-DLL-BDA : Mapping Default PIDS, pid scan failed!\n");
			log("win32-DLL-BDA : Setting streams stream1 = %d stream2 = %d\n", m_lvPID, m_laPID);

			if (SUCCEEDED (hr02))
			{
				ULONG pid = m_lvPID;
				hr02 = muxMapPid->MapPID(1, &pid, MEDIA_TRANSPORT_PACKET);
				if (FAILED(hr02))
				{
					log("win32-DLL-BDA : Cannot MapPID (video) on IMPEG2PIDMap interface %x\n", hr02);
					return hr02;
				}

				pid = m_laPID;
				hr02 = muxMapPid->MapPID(1, &pid, MEDIA_TRANSPORT_PACKET);
				if (FAILED(hr02))
				{
					log("win32-DLL-BDA : Cannot MapPID (audio) on IMPEG2PIDMap interface %x\n", hr02);
					return hr02;
				}
			}
			else
			{
				log("win32-DLL-BDA : Cannot QI IMPEG2PIDMap %x\n", hr02);
				return hr02;
			}
		}
	}

	// ok so all is well and just return with no errors
	log("win32-DLL-BDA : About to leave RunGraph with code 0\n");
	return 0;
}

HRESULT
CBDAFilterGraph::scanForProgramInfo(int *numFound, std::vector <CProgramInfo> *ppinfo)
{
	HRESULT hr;
	(*numFound) = 0;

	if(!m_fGraphRunning)
	{
		log("win32-DLL-BDA : Can not scan for program info as Graph is not running!\n");
		return -1;
	}

	// Get the GuideData interface from the TIF filter
	CComPtr <IGuideData> pGuideData;
	hr = m_pTIF.QueryInterface(&pGuideData);
    if (FAILED(hr))
    {
		log("win32-DLL-BDA : Failed to QI for IGuideData on TIF Filter\n");
	}

	// Get the TuneRequestinfo interface from the TIF filter
	CComPtr <ITuneRequestInfo> pTuneRequestInfo;
	hr = m_pTIF.QueryInterface(&pTuneRequestInfo);
    if (FAILED(hr))
    {
		log("win32-DLL-BDA : Failed to QI for ITuneRequestInfo on pGuideData\n");
		return hr;
	}

	// Get a list of services from the GuideData interface
	CComPtr <IEnumTuneRequests>  piEnumTuneRequests; 
	hr = pGuideData->GetServices(&piEnumTuneRequests); 
    if (FAILED(hr))
    {
		log("win32-DLL-BDA : Failed to get IEnumTuneRequests\n");
		return hr;
	}

	CProgramInfo info;

	//
	// Now loop through all the availabel services on the frequency
	// 

	unsigned long ulRetrieved = 1;
	ITuneRequest *piTuneRequest = NULL; 

	while (SUCCEEDED(piEnumTuneRequests->Next(1, &piTuneRequest, &ulRetrieved)) && ulRetrieved > 0)
	{
		memset(&info, 0, sizeof(CProgramInfo));
		StringCchCopy(info.programName, 256, "Not Available");

		// Get the service properties
		CComPtr <IEnumGuideDataProperties> serviceProp;

		hr = pGuideData->GetServiceProperties(piTuneRequest, &serviceProp);
		if(FAILED(hr))
		{
			log("win32-DLL-BDA : Cannot Get GetServiceProperties\n");
		}
		else
		{

			// Now Print the service Properties
			IGuideDataProperty *servProp;
			unsigned long num = 1;
			int nameNum = 0;

			while(SUCCEEDED(serviceProp->Next(1, &servProp, &num)) && num > 0)
			{
				char buff[256];

				CComBSTR name;
				servProp->get_Name(&name);

				CComVariant value;
				servProp->get_Value(&value);
				StringCchPrintf(buff, 256, "%S : %S\n", name.m_str, value.bstrVal);
				log(buff);

				if(nameNum == 1)
				{
					StringCchPrintf(info.programName, 256, "%S", value.bstrVal);
				}

				nameNum++;
			}
		}

		// Fill in the Components lists for the tune request
		hr = pTuneRequestInfo->CreateComponentList(piTuneRequest);
		if(FAILED(hr))
		{
			log("win32-DLL-BDA : Failed to GetNextProgram from TIF ITuneRequestInfo (0x%x)\n", hr);
		}
		else
		{
			enumComponents(piTuneRequest, &info);
		}

		
		ppinfo->push_back(info);

		(*numFound)++;

	}

	//
	// Now do the internal PID scan for this channel
	//
	CStreamScan scanner;
	int numProgsFound = scanner.scanForProgs(m_pMpegSections);
	log("win32-DLL-BDA : Internal Scan for Programs found %d\n", numProgsFound);


	//
	// Now update the programs with AC3 data
	//

	int size = ppinfo->size();

	for(int x = 0; x < size; x++)
	{
		scanner.updateProgramInfo(&(ppinfo->at(x)));
	}

	return hr;
}

HRESULT
CBDAFilterGraph::enumComponents(ITuneRequest *pTuneRequest, CProgramInfo *info)
{
	HRESULT hr;

	CComPtr <IComponents> pConponents;
	hr = pTuneRequest->get_Components(&pConponents);
    if(FAILED(hr))
    {
		log("win32-DLL-BDA : Cannot Get Components\n");
		return hr;
	}

    CComPtr<IEnumComponents> pEnum;
    hr = pConponents->EnumComponents(&pEnum);

    if (SUCCEEDED(hr))
    {
        CComPtr <IComponent> pComponent;
        ULONG cFetched = 1;

        while (SUCCEEDED(pEnum->Next(1, &pComponent, &cFetched)) && cFetched > 0)
        {

			char buff[256];

			CComPtr <IMPEG2Component> mpegComponent;

            hr = pComponent.QueryInterface(&mpegComponent);

			long PID = -1;
			hr = mpegComponent->get_PID(&PID);

			long progPID = -1;
			mpegComponent->get_ProgramNumber(&progPID);

			long pcrPID = -1;
			mpegComponent->get_PCRPID(&pcrPID);

			CComPtr <IComponentType> comType;
			mpegComponent->get_Type(&comType);

			ComponentCategory cat;
			comType->get_Category(&cat);

			info->programID = progPID;

			StringCchPrintf(buff, 256, "progPID = %d : PID = %d : pcrPID = %d : CAT = %d\n", progPID, PID, pcrPID, cat);
			log(buff); 

			pComponent.Release();
        }
    }

	return hr;
}

HRESULT
CBDAFilterGraph::EnableFileWriter(int writerIndex, BOOL enabled)
{
	HRESULT hr = S_OK;

	if(writerIndex < 0 || writerIndex >= (int)dumpFilters.size())
	{
		log("win32-DLL-BDA : FileWriter index (%d) not in valid range of (%d)\n", writerIndex, dumpFilters.size());
		return -1;
	}

    CComPtr <IBaseFilter> writer = dumpFilters.at(writerIndex);

	CComPtr <IFileWriter> pIFileWriter;
	HRESULT qhr = writer.QueryInterface(&pIFileWriter);
	if(FAILED (qhr) || pIFileWriter == NULL)
	{
		log("win32-DLL-BDA : Cannot find FileWriter Interface on Dump Filter\n");
	}

	if(enabled)
	{
		log("win32-DLL-BDA : Enabling FileWriter Filter (%d)\n", writerIndex);
		pIFileWriter->WriteEnabled(TRUE);
	}
	else
	{
		log("win32-DLL-BDA : Disabling FileWriter Filter (%d)\n", writerIndex);
		pIFileWriter->WriteEnabled(FALSE);
	}

	return hr;
}

// StopGraph calls MediaCtrl - Stop
HRESULT
CBDAFilterGraph::StopGraph()
{

	// check to see if the graph is already stopped
    if(m_fGraphRunning == false)
        return S_OK;

	HRESULT hr = S_OK;

    ASSERT (m_pIMediaControl);
    // pause before stopping
    hr = m_pIMediaControl->Pause();

    // stop the graph
    hr = m_pIMediaControl->Stop();

    m_fGraphRunning = (FAILED (hr))?true:false;
    return hr;
}


HRESULT
CBDAFilterGraph::ChangeChannel()
{
	return ChangeChannel(m_lFrequency, m_lBandwidth);
}


HRESULT
CBDAFilterGraph::ChangeChannel(
        LONG lFreq,
        LONG lBand 
        )
{
    HRESULT hr = S_OK;
    m_lFrequency = lFreq;
    m_lBandwidth = lBand;

	log("win32-DLL-BDA : Changing Frequency %d bandwidth %d.\n", m_lFrequency, m_lBandwidth);

    if (!m_pNetworkProvider)
    {
        log("win32-DLL-BDA : The FilterGraph is not yet built.\n");
        return E_FAIL;
    }


    // create tune request
    CComPtr <IDVBTuneRequest> pTuneRequest;
    hr = CreateDVBTTuneRequest(
                            lFreq, 
                            lBand, 
                            &pTuneRequest
                            );

    if(SUCCEEDED(hr))
    {
        hr = m_pITuner->put_TuneRequest(pTuneRequest);
        if (FAILED (hr))
            log("win32-DLL-BDA : Cannot submit tune request\n");
    }
    else
    {
        log("win32-DLL-BDA : Cannot Submit Channel Request\n");
		return hr;
    }

    return hr;
}

HRESULT CBDAFilterGraph::queryTunerDevices(std::vector <std::string> *tuners)
{
    HRESULT hr = S_OK;
    CComPtr <IMoniker> pIMoniker;
    CComPtr <IEnumMoniker> pIEnumMoniker;

    if (!m_pICreateDevEnum)
    {
        hr = m_pICreateDevEnum.CoCreateInstance(CLSID_SystemDeviceEnum);
        if (FAILED (hr))
        {
            log("win32-DLL-BDA : queryTunerDevices(): Cannot CoCreate ICreateDevEnum\n");
            return hr;
        }
    }

    // obtain the enumerator
    hr = m_pICreateDevEnum->CreateClassEnumerator(KSCATEGORY_BDA_NETWORK_TUNER, &pIEnumMoniker, 0);
    // the call can return S_FALSE if no moniker exists, so explicitly check S_OK
    if (FAILED (hr))
    {
        log("win32-DLL-BDA : queryTunerDevices(): Cannot CreateClassEnumerator\n");
        return hr;
    }
    if (S_OK != hr)  // Class not found
    {
        log("win32-DLL-BDA : queryTunerDevices(): Class not found, CreateClassEnumerator returned S_FALSE");
        return E_UNEXPECTED;
    }


    while(pIEnumMoniker->Next(1, &pIMoniker, 0) == S_OK)
    {

		// obtain filter's friendly name
		CComPtr <IPropertyBag>  pBag;
		hr = pIMoniker->BindToStorage(
									NULL, 
									NULL, 
									IID_IPropertyBag,
									reinterpret_cast<void**>(&pBag)
									);

		if(FAILED(hr))
		{
			OutputDebugString (TEXT("queryTunerDevices(): Cannot BindToStorage"));
			return hr;
		}

		CComVariant varBSTR;
		hr = pBag->Read(L"FriendlyName", &varBSTR, NULL);
		if(FAILED(hr))
		{
			OutputDebugString (TEXT("queryTunerDevices(): IPropertyBag->Read method failed"));
			pIMoniker = NULL;
			continue;
		}

		char buff[512];
		StringCchPrintf(buff, 512, "%S", varBSTR.bstrVal);

		tuners->push_back(buff);

		//log("win32-DLL-BDA : queryTunerDevices(): %S\n", varBSTR.bstrVal);

		pIMoniker = NULL;
	}

	return hr;
}


#ifdef REGISTER_FILTERGRAPH

// Adds a DirectShow filter graph to the Running Object Table,
// allowing GraphEdit to "spy" on a remote filter graph.
HRESULT CBDAFilterGraph::AddGraphToRot(
        IUnknown *pUnkGraph, 
        DWORD *pdwRegister
        ) 
{
    CComPtr <IMoniker>              pMoniker;
    CComPtr <IRunningObjectTable>   pROT;
    WCHAR wsz[128];
    HRESULT hr;

    if (FAILED(GetRunningObjectTable(0, &pROT)))
        return E_FAIL;

	StringCchPrintfW(wsz, 128, L"FilterGraph %08x pid %08x\0", (DWORD_PTR) pUnkGraph, 
              GetCurrentProcessId());
    //wsprintfW(wsz, L"FilterGraph %08x pid %08x\0", (DWORD_PTR) pUnkGraph, 
    //          GetCurrentProcessId());

    hr = CreateItemMoniker(L"!", wsz, &pMoniker);
    if (SUCCEEDED(hr)) 
        hr = pROT->Register(ROTFLAGS_REGISTRATIONKEEPSALIVE, pUnkGraph, 
                            pMoniker, pdwRegister);
        
    return hr;
}

// Removes a filter graph from the Running Object Table
void CBDAFilterGraph::RemoveGraphFromRot(
        DWORD pdwRegister
        )
{
    CComPtr <IRunningObjectTable> pROT;

    if (SUCCEEDED(GetRunningObjectTable(0, &pROT))) 
        pROT->Revoke(pdwRegister);

}

#endif

HRESULT
CBDAFilterGraph::LoadPsiFilter()
{
	log("win32-DLL-BDA : Deleting pins from demux & loading/connecting PSI filter.\n");

    HRESULT hr = S_OK;
	
	// Find the Sections & tables Pin & delete the rest.
	wchar_t pinName[128];
    PIN_DIRECTION  direction;
	CComPtr <IPin> pIPin = NULL;
	CComPtr <IPin> pDownstreamPin = NULL;
	AM_MEDIA_TYPE *type;
	
	// Get an instance of the Demux control interface
	CComPtr <IMpeg2Demultiplexer> muxInterface;
	hr = m_pDemux.QueryInterface (&muxInterface);
	
	// Enumerate the Demux pins
    CComPtr <IEnumPins> pIEnumPins;
    hr = m_pDemux->EnumPins (&pIEnumPins);
	
    if (FAILED (hr))
    {
        log("win32-DLL-BDA : Cannot get enumpins on Demux \n");
        return hr;
    }
	
    while(pIEnumPins->Next(1, &pIPin, 0) == S_OK)
    {
        hr = pIPin->QueryDirection(&direction);
		
        if(direction == PINDIR_OUTPUT)
        {
            pIPin->ConnectedTo(&pDownstreamPin);
			
            if(pDownstreamPin == NULL)
            {
				PIN_INFO pinInfo;
				if (FAILED(pIPin->QueryPinInfo(&pinInfo)))
				{
					log("win32-DLL-BDA : Cannot Get Demux Output Pin Info\n");
					return hr;
				}
				
				CComPtr <IEnumMediaTypes> ppEnum;
				if (SUCCEEDED (pIPin->EnumMediaTypes(&ppEnum)))
				{
					while(ppEnum->Next(1, &type, 0) == S_OK)
					{
						//Save if Name of Pin for Sections & tables render
						if (type->majortype == KSDATAFORMAT_TYPE_MPEG2_SECTIONS && 
							type->subtype == MEDIASUBTYPE_MPEG2DATA)
						{
							StringCchCopyW(pinName, 128, pinInfo.achName);
						}
						else
						{
							muxInterface->DeleteOutputPin(pinInfo.achName);
						}
						DeleteMediaType(type);
					}
				}
				ppEnum = NULL;
            }
            pDownstreamPin = NULL;
        }
        pIPin = NULL;
    }
	
	// MPEG2 Sections and Tables Loading
    if (FAILED(LoadFilter (
		KSCATEGORY_BDA_TRANSPORT_INFORMATION,
		&m_pMpegSections,
		m_pDemux, 
		TRUE, -1
		)))
    {
		log("win32-DLL-BDA : Rendering MPEG-2 Sections and Tables Filter \n");
		CComPtr <IPin> pIPin;
		CComPtr <IPin> pIPin2;
		
		if (FAILED(m_pDemux->FindPin(pinName, &pIPin)))
		{
			log("win32-DLL-BDA : Cannot find Demux MPEG-2 Sections and Tables Pin \n");
			return hr;
		}
		
		if (FAILED(m_pFilterGraph->Render(pIPin)))
		{
			log("win32-DLL-BDA : Cannot load MPEG-2 Sections and Tables\n");
			return hr;
		}
		
		if (FAILED(pIPin->ConnectedTo(&pIPin2)))
		{
			log("win32-DLL-BDA : Cannot Get MPEG-2 Sections and Tables Input Pin\n");
			return hr;
		}
		
		PIN_INFO pinInfo;
		if (FAILED(pIPin2->QueryPinInfo(&pinInfo)))
		{
			log("win32-DLL-BDA : Cannot Get MPEG-2 Sections and Tables Input Pin Info\n");
			return hr;
		}
		
		m_pMpegSections = pinInfo.pFilter;
    }
	
	log("win32-DLL-BDA : PSI loaded and added with result hr=0x%x\n", hr);

	return hr;
	
}



